/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "GuitoneDriver.h"

#include "DatabaseDialogManager.h"
#include "WorkspaceDialogManager.h"
#include "ServerDialogManager.h"
#include "Message.h"

#include <QGenericArgument>
#include <QMetaMethod>

GuitoneDriver::GuitoneDriver(int & argc, char** argv)
  : GuitoneCore(argc, argv), dialogManager(0)
{
    connect(
        this, SIGNAL(monotoneBinaryPathSet()),
        this, SLOT(processCommands()),
        Qt::QueuedConnection
    );
}

GuitoneDriver::~GuitoneDriver()
{
    if (dialogManager) delete dialogManager;
}

void GuitoneDriver::processCommands()
{
    QStringList args = arguments();
    // we're not interested in the program's name and "exec"
    args.removeFirst();
    args.removeFirst();

    if (args.size() < 2)
    {
        help();
        return;
    }

    QString path = args.takeFirst();

    if (path == "-")
    {
        dialogManager = new DialogManager(NULL);
    }
    else
    {
        MonotoneHandlePtr handle;
        try
        {
            handle = MonotoneHandle::create(path);
        }
        catch (GuitoneException & e)
        {
            msg(
                QString("Path / URI '%1' could not be loaded: %2")
                    .arg(path).arg(e.what())
            );
            QCoreApplication::exit(1);
            return;
        }

        APP->manager()->getThread(handle)->waitForStarted();

        if (handle->getType() == MonotoneHandle::workspace_handle)
        {
            dialogManager = new WorkspaceDialogManager(NULL);
            qobject_cast<WorkspaceDialogManager *>(dialogManager)->init(handle);
        }
        else if (handle->getType() == MonotoneHandle::database_handle)
        {
            dialogManager = new DatabaseDialogManager(NULL);
            qobject_cast<DatabaseDialogManager *>(dialogManager)->init(handle);
        }
        else if (handle->getType() == MonotoneHandle::server_handle)
        {
            dialogManager = new ServerDialogManager(NULL);
            qobject_cast<ServerDialogManager *>(dialogManager)->init(handle);
        }
        else
            I(false);
    }

    // if all dialogs have been closed, quit the event loop and exit
    connect(
        dialogManager, SIGNAL(allDialogsClosed()),
        APP, SLOT(quit())
    );

    // connect to all signals of the dialog manager
    // FIXME: unfortunately our "general purpose" slots have no idea which
    // signal they're responding to, since Qt has no API for that, obviously
    // they store this information. It would be too cool if there would be
    // a QObject::senderSignal() similar to QObject::sender() in the future...
    const QMetaObject * metaObject = dialogManager->metaObject();
    for (int i=0; i<metaObject->methodCount(); i++)
    {
        QMetaMethod m = metaObject->method(i);
        if (m.methodType() != QMetaMethod::Signal) continue;

        QByteArray sig = m.signature();
        // ignore these signals from QObject
        if (sig == "destroyed()" || sig == "destroyed(QObject*)")
            continue;

        D(QString("connecting to signal '%1'").arg(sig.constData()));

        sig.prepend("2"); // see qobjectdefs.h
        if (m.parameterTypes().size() == 0)
        {
            connect(
                dialogManager, sig,
                this, SLOT(signalWithNoArguments())
            );
        }
        else
        if (m.parameterTypes().size() == 1 &&
            m.parameterTypes().at(0) == "QString")
        {
            connect(
                dialogManager, sig,
                this, SLOT(signalWithOneStringArgument(const QString &))
            );
        }
        else
        {
            W(QString("could not connect to signal '%1'")
              .arg(sig.constData()));
        }
    }

    QString reqMethod = args.takeFirst();

    const QMetaObject * obj = dialogManager->metaObject();
    QList<QMetaMethod> methodCandidates;

    for (int i=0; i<obj->methodCount(); i++)
    {
        QMetaMethod method = obj->method(i);
        QByteArray sig(method.signature());
        sig = sig.left(sig.indexOf('('));
        if (sig == reqMethod.toLatin1())
        {
            methodCandidates.push_back(method);
        }
    }

    if (methodCandidates.size() == 0)
    {
        msg(QString("method signature '%1' not found").arg(reqMethod));
        QCoreApplication::exit(1);
        return;
    }

    // FIXME: for now we only compare the number of arguments of a function
    // but not their argument type - which of course would make a valid
    // overload as well. since this is not a problem in the current API
    // we leave this task for later...
    const QMetaMethod * chosenMethod = 0;
    foreach (const QMetaMethod & m, methodCandidates)
    {
        if (m.parameterTypes().size() == args.size())
        {
            chosenMethod = &m;
        }
    }

    if (!chosenMethod)
    {
        msg(QString("found %1 overloads for '%2', but none takes %3 arguments")
                .arg(methodCandidates.size()).arg(reqMethod).arg(args.size()));
        QCoreApplication::exit(1);
        return;
    }

    if (args.size() > 10)
    {
        msg("cannot handle more than 10 function arguments");
        QCoreApplication::exit(1);
        return;
    }

    QGenericArgument genArgs[10];
    QStringList listElements;

    for (int i=0; i<chosenMethod->parameterTypes().size(); i++)
    {
        const char * typeName = chosenMethod->parameterTypes().at(i).constData();
        QVariant::Type type = QVariant::nameToType(typeName);

        // we currently only use / support strings and string lists
        if (type == QVariant::String)
        {
            genArgs[i] = Q_ARG(QString, args.at(i));
        }
        else if (type == QVariant::StringList)
        {
            QString arg = args.at(i);
            QRegExp rx("[^\\\\],");
            int pos = 0;
            while ((pos = rx.indexIn(arg)) != -1)
            {
                listElements << arg.left(pos + 1);
                arg = arg.mid(pos + 2);
            }
            listElements << arg;

            for (int j=0; j<listElements.size(); j++)
                listElements[j] = listElements[j].replace("\\,", ",");

            genArgs[i] = Q_ARG(QStringList, listElements);
        }
        else if (type == QVariant::Invalid)
        {
            I(false);
        }
        else
        {
            msg(QString("unrecognized type '%1'").arg(typeName));
            QCoreApplication::exit(1);
            return;
        }
    }

    const char * methodName = static_cast<const char *>(reqMethod.toLatin1().data());
    bool ok = QMetaObject::invokeMethod(
        dialogManager, methodName, Qt::QueuedConnection,
        genArgs[0], genArgs[1], genArgs[2], genArgs[3],
        genArgs[4], genArgs[5], genArgs[6], genArgs[7],
        genArgs[8], genArgs[9]
    );

    D(QString("execution successful: %1").arg(ok));
}

void GuitoneDriver::msg(const QString & message)
{
    Message msg(NULL, "guitone driver interface", message);
    msg.exec();
}

void GuitoneDriver::help()
{
    QList<const QMetaObject *> objs;
    objs.append((new DialogManager(NULL))->metaObject());
    objs.append((new DatabaseDialogManager(NULL))->metaObject());
    objs.append((new WorkspaceDialogManager(NULL))->metaObject());
    objs.append((new ServerDialogManager(NULL))->metaObject());

    QStringList output;
    output << "Usage: guitone exec <path-or-uri> <function> [<args> ...]\n\n";
    output << "Available functions:\n\n";

    foreach (const QMetaObject * obj, objs)
    {
        QMetaObject const * parent = obj->superClass();
        output << obj->className();
        if (strcmp(parent->className(), "QObject") != 0)
            output << " (inherits " << parent->className() << ")";
        output << ":\n";

        for (int i=obj->methodOffset(); i < obj->methodCount(); i++)
        {
            QMetaMethod m = obj->method(i);
            if (m.access() != QMetaMethod::Public ||
                m.methodType() != QMetaMethod::Slot)
                continue;

            QByteArray sig(m.signature());
            sig = sig.left(sig.indexOf('('));
            QList<QByteArray> names = m.parameterNames();
            QList<QByteArray> types = m.parameterTypes();

            output << "\t" << sig;
            for (int i=0; i<types.size(); i++)
            {
                output << " <" << types.at(i);
                if (names.at(i).size() > 0)
                    output << ":" << names.at(i);
                output << ">";
            }
            output << "\n";
        }
        output << "\n";
    }
    output << "The QString type is a simple, optionally quoted string.\n";
    output << "\n";
    output << "The QStringList type has to be given as a single argument. "
              "Elements are separated with a comma ','; any comma inside "
              "a string element has to be escaped with a backslash '\\,'.\n";
    output << "\n";
    output << "The MonotoneHandle::type is an integral enum type which can "
              "be either '2' to denote a workspace, '3' to denote a database "
              "or '4' to denote a server. The internal empty handle '1' "
              "should not be used.\n";

    msg(output.join(""));
}

void GuitoneDriver::signalWithNoArguments()
{
    L("signalled");
}

void GuitoneDriver::signalWithOneStringArgument(const QString & arg)
{
    L(QString("signalled, value: %1").arg(arg));
}

