/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

// raise the file descriptor limit from 1042 (default)
#ifdef Q_WS_MAC
#define FD_SETSIZE 4096
#endif

#include "DebugLog.h"
#include "GuitoneCore.h"
#include "GuitoneStandalone.h"
#include "GuitoneDriver.h"

#include "CertList.h"

#include <QMap>
#include <QLocale>
#include <QTranslator>
#include <QLibraryInfo>

#include <iostream>

void guitoneMsgHandler(QtMsgType type, const char * msg)
{
    static QMap<QtMsgType, DebugLog::Type> typeMap;
    if (typeMap.size() == 0)
    {
        typeMap.insert(QtDebugMsg, DebugLog::Debug);
        typeMap.insert(QtWarningMsg, DebugLog::Warn);
        typeMap.insert(QtCriticalMsg, DebugLog::Critical);
        typeMap.insert(QtFatalMsg, DebugLog::Fatal);
    }

    if (typeMap.contains(type))
        DebugLog::log(typeMap.value(type), QString::fromUtf8(msg));
    else
        DebugLog::log(DebugLog::Critical, QString::fromUtf8(msg));

    if (type == QtFatalMsg)
        abort();
}

int main(int argc, char** argv)
{
    // install our own message handler to catch debug messages, warnings, etc.
    qInstallMsgHandler(guitoneMsgHandler);

    try
    {
        qRegisterMetaType<TickerMap>("TickerMap");
        qRegisterMetaType<CertList>("CertList");
        qRegisterMetaType<MonotoneTaskPtr>("MonotoneTaskPtr");
        qRegisterMetaType<MonotoneHandlePtr>("MonotoneHandlePtr");
        qRegisterMetaType<QProcess::ProcessError>("QProcess::ProcessError");

        QString locale = QLocale::system().name();

        // load the default Qt translations
        QTranslator qtTranslator;
        QString transPath = QLibraryInfo::location(QLibraryInfo::TranslationsPath);
        if (!qtTranslator.load("qt_" + locale, transPath) && !locale.startsWith("en"))
        {
            W(QString("Couldn't load qt translation for %1 - using default")
              .arg(locale));
        }

        // try to load the application's translation
        QTranslator appTranslator;
        QString transFileName("guitone_" + locale);

        if (!appTranslator.load(transFileName, ":/i18n") && !locale.startsWith("en"))
        {
            W(QString("Couldn't load app translation for %1 - using default")
              .arg(locale));
        }

        GuitoneCore * guitone;
        if (argc > 1 && strcmp(argv[1], "exec") == 0)
        {
            guitone = new GuitoneDriver(argc, argv);
        }
        else
        {
            guitone = new GuitoneStandalone(argc, argv);
        }

#ifdef Q_WS_MAC
        guitone->addLibraryPath(
            QDir::cleanPath(guitone->applicationDirPath() + "/../PlugIns")
        );
#endif

        guitone->installTranslator(&appTranslator);
        guitone->installTranslator(&qtTranslator);

        int retValue = guitone->exec();
        delete guitone;
        return retValue;
    }
    catch (GuitoneException & e)
    {
        F(QString("unhandled guitone exception: %1").arg(e.what()));
        return 1;
    }
    catch (std::exception & e)
    {
        F(QString("unhandled std::exception: %1").arg(e.what()));
        return 1;
    }
    catch (...)
    {
        F("unhandled unknown exception");
        return 1;
    }

    I(false);
}

