/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Annotate.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QSet>

Annotation::Annotation(Annotate * ann, QString rev, QString c, bool s)
  : model(ann), revision(rev), code(c), show_info(s)
{
    annotationFont.setPointSize(10);

    codeFont.setStyleHint(QFont::Courier);
    codeFont.setFamily("Courier");
}

QVariant Annotation::data(int col, int role) const
{
    if (role == Qt::DisplayRole)
    {
        if (col == 0)
        {
            if (!show_info)
                return QVariant();

            CertList certs = model->getCachedCertList(revision);
            if (certs.revision.isEmpty())
                return QVariant(revision);

            QList<Cert> authorCerts = certs.findByName("author");
            QList<Cert> dateCerts = certs.findByName("date");

            if (authorCerts.size() == 0 || dateCerts.size() == 0)
                return QVariant(revision);

            QString author = authorCerts.at(0).value;
            if (author.size() > 8)
                author = author.left(6) + "..";

            QString date = dateCerts.at(0).dateTimeValue().date().toString(
                Qt::DefaultLocaleShortDate
            );

            return QVariant(QObject::tr("%1 by %2 on %3")
                .arg(revision.left(8)).arg(author).arg(date)
            );
        }

        if (col == 1)
            return QVariant(code);
    }

    if (role == Qt::FontRole)
    {
        if (col == 0)
            return QVariant(annotationFont);

        if (col == 1)
            return QVariant(codeFont);
    }

    if (role == Qt::UserRole)
    {
        return QVariant(revision);
    }

    return QVariant();
}

Annotate::Annotate(QObject * parent)
    : QAbstractItemModel(parent), proc(0), certsModel(0)
{
    proc = new MonotoneProcess();
    connect(
        proc, SIGNAL(finished(int, QProcess::ExitStatus)),
        this, SLOT(processFinished(int, QProcess::ExitStatus))
    );
}

Annotate::~Annotate()
{
    if (proc) delete proc;
    if (certsModel) delete certsModel;
}

void Annotate::readAnnotation(const MonotoneHandlePtr & handle,
                              const QString & rev,
                              const QString & file)
{
    // FIXME if there is an annotate available in automate
    I(handle->getType() != MonotoneHandle::server_handle);

    brushMap.clear();
    inverseBrushMap.clear();
    annotationLines.clear();
    reset();

    if (certsModel)
        delete certsModel;

    certsModel = new Certs(this, handle);
    connect(
        certsModel, SIGNAL(certsRead(const CertList &)),
        this, SLOT(certsRead(const CertList &))
    );

    monotoneHandle = handle;
    baseRevision = rev;

    QStringList params;
    params << "annotate" << "--revs-only" << file;
    if (!rev.isEmpty())
    {
        params << "-r" << rev;
    }

    proc->setMonotoneHandle(handle);
    proc->start(params);
}

void Annotate::processFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QString output(proc->getBufferedOutput());
    if (exitCode != 0 || exitStatus != QProcess::NormalExit)
    {
        C(QString("Could not read annotation: %1").arg(output));
        return;
    }

    QStringList lines = output.split('\n', QString::SkipEmptyParts);

    QRegExp rx("([0-9a-f]{40}):");

    QSet<QString> revisions;
    QString lastRevision;

    foreach (QString line, lines)
    {
        if (line.indexOf(rx) != 0)
        {
            W(QString("Could not parse annotation line: %1").arg(line));
            continue;
        }

        QString code = line.mid(rx.matchedLength() + 1);
        QString revision = rx.cap(1);
        bool showAnnotation = revision != lastRevision;
        lastRevision = revision;

        annotationLines.append(
            Annotation(this, revision, code, showAnnotation)
        );
        revisions.insert(revision);
    }

    QStringList sortedRevisions =
        MonotoneUtil::topsortRevisions(monotoneHandle, revisions.toList());

    QColor startColor = QColor(Settings::getString("AnnotationColorOldestRevision"));
    QColor endColor   = QColor(Settings::getString("AnnotationColorNewestRevision"));

    float rStep = (endColor.redF() - startColor.redF()) / sortedRevisions.size();
    float gStep = (endColor.greenF() - startColor.greenF()) / sortedRevisions.size();
    float bStep = (endColor.blueF() - startColor.blueF()) / sortedRevisions.size();
    float r = startColor.redF(), g = startColor.greenF(), b = startColor.blueF();

    foreach (QString rev, sortedRevisions)
    {
        brushMap.insert(rev, QBrush(QColor::fromRgbF(r, g, b)));
        inverseBrushMap.insert(rev, QBrush(QColor::fromRgbF(1.0 - r, 1.0 - g, 1.0 - b)));

        r += rStep; g += gStep; b += bStep;

        if (!certMap.contains(rev))
        {
            certsToRead.push(rev);
        }
    }

    highlightBrush = QBrush(QColor(
        Settings::getString("AnnotationColorHighlight")
    ));

    reset();
    emit annotationRead();

    readNextCerts();
}

void Annotate::readNextCerts()
{
    if (certsToRead.size() == 0)
        return;
    certsModel->readCerts(certsToRead.pop());
}

void Annotate::certsRead(const CertList & certs)
{
    certMap[certs.revision] = certs;
    for (int i=0; i<annotationLines.size(); ++i)
    {
        Annotation ann = annotationLines.at(i);
        if (ann.revision == certs.revision && ann.show_info)
        {
            emit dataChanged(
                index(i, 0, QModelIndex()),
                index(i, 0, QModelIndex())
            );
        }
    }

    if (certs.revision == selectedRevision)
        emit currentCertList(certs);

    readNextCerts();
}

CertList Annotate::getCachedCertList(const QString & rev) const
{
    if (!certMap.contains(rev))
        return CertList();
    return certMap.value(rev);
}

int Annotate::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant Annotate::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    int row = index.row(), col = index.column();

    if (row >= annotationLines.size())
        return QVariant();

    Annotation ann = annotationLines.at(row);

    if (role == Qt::BackgroundRole && col == 0)
    {
        return QVariant(brushMap.value(ann.revision));
    }

    if (role == Qt::ForegroundRole && col == 0)
    {
        return QVariant(inverseBrushMap.value(ann.revision));
    }

    if (role == Qt::BackgroundRole && col == 1 && ann.revision == selectedRevision)
    {
        return QVariant(highlightBrush);
    }

    return ann.data(col, role);
}

Qt::ItemFlags Annotate::flags(const QModelIndex & index) const
{
    if (index.isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return 0;
}

QVariant Annotate::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_UNUSED(section);
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Annotation"));
            case 1: return QVariant(tr("Code"));
        }
        return QVariant();
    }
    return QVariant();
}

int Annotate::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return annotationLines.size();
}

QModelIndex Annotate::index(int row, int column, const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    return createIndex(row, column, 0);
}

QModelIndex Annotate::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

void Annotate::setSelectedRevision(const QString & rev)
{
    if (selectedRevision != rev)
    {
        selectedRevision = rev;
        reset();

        if (certMap.contains(rev))
            emit currentCertList(certMap.value(rev));
    }
}

QString Annotate::getSelectedRevision() const
{
    return selectedRevision;
}

QString Annotate::getBaseRevision() const
{
    return baseRevision;
}

