/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "AutomateCommand.h"
#include "GuitoneCore.h"

AutomateCommand::AutomateCommand(AutomateCommandHelper * helper)
{
    // ensure that we only delete stuff later on we actually created ourselves
    if (helper == 0)
    {
        cmdHelperCreated = true;
        cmdHelper = new AutomateCommandHelper(this);
    }
    else
    {
        cmdHelperCreated = false;
        cmdHelper = helper;
    }
}

AutomateCommand::~AutomateCommand()
{
    if (cmdHelperCreated)
    {
        delete cmdHelper;
    }
}

void AutomateCommand::enqueueTask(const MonotoneHandlePtr & handle,
                                  const MonotoneTaskPtr & task)
{
    QMutexLocker locker(&lock);

    MonotoneThreadPtr thread = APP->manager()->getThread(handle);
    int threadNumber = thread->getThreadNumber();

    // we want to avoid multiple thread connections
    if (!connectedThreads.contains(threadNumber))
    {
        QObject::connect(
            thread.data(), SIGNAL(taskAborted(const MonotoneTaskPtr &)),
            cmdHelper, SLOT(taskAborted(const MonotoneTaskPtr &))
        );

        QObject::connect(
            thread.data(), SIGNAL(taskFinished(const MonotoneTaskPtr &)),
            cmdHelper, SLOT(taskFinished(const MonotoneTaskPtr &))
        );

        QObject::connect(
            thread.data(), SIGNAL(tickersChanged(const TickerMap &)),
            cmdHelper, SLOT(tickersChanged(const TickerMap &))
        );

        connectedThreads.append(threadNumber);
    }

    if (!queuedCommands.contains(threadNumber))
    {
        queuedCommands.insert(threadNumber, QList<int>());
    }

    queuedCommands[threadNumber].append(
        thread->enqueueTask(task)
    );

}

bool AutomateCommand::hasQueuedCommands() const
{
    int count = 0;

    QMapIterator<int, QList<int> > it(queuedCommands);
    while (it.hasNext())
    {
        it.next();
        count += it.value().size();
    }

    return count > 0;
}

void AutomateCommand::abortThreads()
{
    QMutexLocker locker(&lock);

    foreach (int threadNumber, connectedThreads)
    {
        // FIXME: this plainly aborts all threads for which we have
        // commands enqueued, but other models may have added their own
        // commands in the same thread queue and are therefor also aborted!
        //
        // We should find a clean and nice way to only abort the currently
        // running task and to restart the mtn process proceeding with the
        // left items in the task queue...
        try
        {
            APP->manager()->stopThread(threadNumber);
        }
        catch (GuitoneException & e)
        {
            W(QString("could not stop thread %1: %2")
                .arg(threadNumber).arg(e.what()));
        }
    }

    connectedThreads.clear();
    queuedCommands.clear();
}

void AutomateCommand::tickerUpdate(const TickerMap & tickers)
{
    Q_UNUSED(tickers);
}

void AutomateCommand::processTaskAbortion(const MonotoneTaskPtr & task)
{
    // the task was aborted explicitely, no need to be verbose here
    if (task->isAborted()) return;

    // do sth more reasonable here in a subclass,
    // i.e. try to query the data again
    C(QString("Task %1 (%2) aborted (%3): %4")
        .arg(task->getCommandNumber())
        .arg(QString::fromUtf8(task->getEncodedInput()))
        .arg(task->getReturnCode())
        .arg(task->getDecodedOutput())
    );
}

AutomateCommandHelper::AutomateCommandHelper(AutomateCommand * cmd)
  : command(cmd)
{}

void AutomateCommandHelper::taskAborted(const MonotoneTaskPtr & task)
{
    int threadNumber = task->getThreadNumber();
    int cmdNumber = task->getCommandNumber();

    bool threadFound = command->queuedCommands.contains(threadNumber);
    int commandIdx = !threadFound ? -1 :
        command->queuedCommands[threadNumber].indexOf(cmdNumber);

    if (commandIdx == -1)
        return;

    command->queuedCommands[threadNumber].removeAt(commandIdx);
    if (command->queuedCommands[threadNumber].size() == 0)
    {
        command->queuedCommands.remove(threadNumber);
    }

    command->processTaskAbortion(task);
}

void AutomateCommandHelper::taskFinished(const MonotoneTaskPtr & task)
{
    int threadNumber = task->getThreadNumber();
    int cmdNumber = task->getCommandNumber();

    bool threadFound = command->queuedCommands.contains(threadNumber);
    int commandIdx = !threadFound ? -1 :
        command->queuedCommands[threadNumber].indexOf(cmdNumber);

    if (commandIdx == -1)
        return;

    command->queuedCommands[threadNumber].removeAt(commandIdx);
    command->processTaskResult(task);
}

void AutomateCommandHelper::tickersChanged(const TickerMap & tickers)
{
    command->tickerUpdate(tickers);
}

