/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Branches.h"
#include "BasicIOParser.h"
#include "Settings.h"
#include "CertList.h"

/*
 * We use two hacks here:
 * 1) we mark suspended branches by setting the data behind Qt::UserRole
 *    in the specific branch item to true
 * 2) we set the full branch name in tree and list mode as data behind
 *    Qt::ToolTipRole
 */
Branches::Branches(QObject * parent, const MonotoneHandlePtr & handle)
  : QStandardItemModel(parent), AutomateCommand(0),
    builder(0), monotoneHandle(handle)
{
    branchesAsTree = false;
    suspendedBranches = false;
}

Branches::~Branches()
{
    if (builder)
        delete builder;
}

void Branches::readBranches()
{
    branches.clear();
    activeHeadsMap.clear();
    reset();

    suspendedBranchesBrush = QBrush(QColor(
        Settings::getString("ChangesetBrowserSuspendedBranches")
    ));

    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "branches",
        QStringList() << "ignore-suspend-certs" << ""
    ));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void Branches::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    QString cmd = task->getArguments().at(0);
    if (cmd == "branches")
    {
        branches = task->getDecodedOutput()
            .split("\n", QString::SkipEmptyParts);

        rebuildBranchModel();

        emit branchesRead();

        setHorizontalHeaderItem(0, new QStandardItem(
            tr("Branches (determining suspended branches)")
        ));

        foreach (const QString & branch, branches)
        {
            MonotoneTaskPtr task(new MonotoneTask(
                QStringList() << "select" << "h:" + branch,
                QStringList() << "ignore-suspend-certs" << ""
            ));
            AutomateCommand::enqueueTask(monotoneHandle, task);
        }
    }
    else if (cmd == "select")
    {
        QString branch = task->getArguments().at(1).mid(2);
        QStringList heads = task->getDecodedOutput()
            .split("\n", QString::SkipEmptyParts);

        foreach (const QString & head, heads)
        {
            activeHeadsMap[branch].insert(head);

            MonotoneTaskPtr task(new MonotoneTask(
                QStringList() << "certs" << head
            ));
            AutomateCommand::enqueueTask(monotoneHandle, task);
        }
    }
    else if (cmd == "certs")
    {
        QString head = task->getArguments().at(1);
        BasicIOParser parser(task->getDecodedOutput());
        I(parser.parse());

        CertList certs;
        certs.revision = head;
        certs.fill(parser.getStanzas());

        QStringList suspendedBranches = certs.findByNameValues("suspend");
        foreach (const QString & suspendedBranch, suspendedBranches)
        {
            activeHeadsMap[suspendedBranch].remove(head);
        }

        if (!hasQueuedCommands())
        {
            markSuspendedBranches(invisibleRootItem());

            setHorizontalHeaderItem(0, new QStandardItem(tr("Branches")));

            reset();
        }
    }
}

Qt::ItemFlags Branches::flags(const QModelIndex & index) const
{
    if (!index.isValid())
        return 0;

    // tree parents group branches, but are not valid branches itself
    // so we don't want them to be selectable
    QFlags<Qt::ItemFlag> flags = Qt::ItemIsEnabled;
    if (rowCount(index) == 0)
    {
        flags |= Qt::ItemIsSelectable;
    }

    return flags;
}

void Branches::markBranchSuspended(const QString & branch)
{
    if (activeHeadsMap.contains(branch))
    {
        activeHeadsMap[branch].clear();
        markSuspendedBranches(invisibleRootItem());
        reset();
    }
}

void Branches::rebuildBranchModel()
{
    branchItemMap.clear();
    clear();

    if (branchesAsTree)
    {
        builder = new TreeBuilder(branchItemMap, invisibleRootItem(), this);
        builder->addList(branches);
    }
    else
    {
        foreach (const QString & branch, branches)
        {
            QStandardItem * item = new QStandardItem();
            item->setData(branch, Qt::DisplayRole);
            item->setData(branch, Qt::ToolTipRole);
            appendRow(item);
            branchItemMap.insert(branch, item);
        }
    }

    reset();
}

void Branches::displayBranchesAsTree(bool tree)
{
    branchesAsTree = tree;
    rebuildBranchModel();
    markSuspendedBranches(invisibleRootItem());
    setHorizontalHeaderItem(0, new QStandardItem(tr("Branches")));
    reset();
}

bool Branches::markSuspendedBranches(QStandardItem * item)
{
    bool suspended = false;
    if (item->hasChildren())
    {
        bool allSuspended = true;
        for (int i=0; i<item->rowCount(); ++i)
        {
            allSuspended &= markSuspendedBranches(item->child(i));
        }
        suspended = allSuspended;
    }
    else
    {
        QString branch = item->data(Qt::ToolTipRole).toString();
        if (activeHeadsMap.contains(branch) &&
            activeHeadsMap.value(branch).size() == 0)
        {
            suspended = true;
        }
    }

    item->setData(
        suspended ? suspendedBranchesBrush : QVariant(),
        Qt::ForegroundRole
    );
    item->setData(suspended, Qt::UserRole);

    return suspended;
}

