/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CertList.h"

QDateTime Cert::dateTimeValue() const
{
    return QDateTime::fromString(value, "yyyy-MM-ddThh:mm:ss");
}

QString Cert::plainValue() const
{
    return QString(value).replace(QRegExp("\\s+"), " ");
}

QString Cert::htmlValue() const
{
    QString val(value);
    QStringList lines = val.replace("<", "&lt;").split(QRegExp("(\\r\\n|\\n|\\n\\r)"));
    QString htmlValue;
    bool inList = false;
    foreach (const QString & line, lines)
    {
        if (line.isEmpty())
        {
            if (inList)
            {
                htmlValue += "</li>\n</ul>\n";
                inList = false;
                continue;
            }

            htmlValue += "<br/>\n";
            continue;
        }
        bool listStart = (0 == line.indexOf(QRegExp("^[*-]")));
        if (listStart)
        {
            if (inList)
            {
                htmlValue += "</li>\n";
            }
            else
            {
                inList = true;
                htmlValue += "<ul>\n";
            }

            htmlValue += "<li>" + line.mid(1).trimmed();
            continue;
        }
        if (inList)
        {
            htmlValue += " " + line.trimmed();
            continue;
        }
        htmlValue += line.trimmed() + "<br/>\n";
    }
    if (inList)
    {
        htmlValue += "</li>\n</ul>\n";
    }
    return htmlValue
        .replace(QRegExp("(\\b)([_])(\\w+)\\2"), "\\1<u>\\2\\3\\2</u>")
        .replace(QRegExp("(\\b)([*])(\\w+)\\2"), "\\1<b>\\2\\3\\2</b>")
        .replace(QRegExp("(https?://[^() ]+)"), "<a href=\"\\1\">\\1</a>")
        .replace(QRegExp("\\t"), "&nbsp;&nbsp;&nbsp;&nbsp;");
}

void CertList::add(const Cert & c)
{
    certs.push_back(c);
}

int CertList::size() const
{
    return certs.size();
}

Cert CertList::at(int pos) const
{
    return certs.at(pos);
}

void CertList::clear()
{
    certs.clear();
    revision = QString();
}

QStringList CertList::signers() const
{
    QStringList signers;
    foreach (const Cert & cert, certs)
    {
        if (!signers.contains(cert.key))
            signers.push_back(cert.key);
    }
    return signers;
}

QList<Cert> CertList::findBySigner(const QString & key) const
{
    QList<Cert> foundCerts;
    foreach (const Cert & cert, certs)
    {
        if (key.isEmpty() || key == cert.key)
            foundCerts.push_back(cert);
    }
    return foundCerts;
}

QList<Cert> CertList::findByName(const QString & name,
                                 const QString & key) const
{
    QList<Cert> foundCerts;
    foreach (const Cert & cert, findBySigner(key))
    {
        if (cert.name != name)
            continue;
        foundCerts.push_back(cert);
    }
    return foundCerts;
}

QStringList CertList::findByNameValues(const QString & name,
                                       const QString & key) const
{
    QStringList values;
    QList<Cert> certs = findByName(name, key);
    foreach (const Cert & cert, certs)
    {
        values.push_back(cert.value);
    }
    return values;
}

void CertList::fill(const StanzaList & stanzas)
{
    foreach (const Stanza & st, stanzas)
    {
        Cert cert;
        foreach (const StanzaEntry & entry, st)
        {
            if (entry.sym == "key")
            {
                I(entry.vals.size() == 1);
                cert.key = entry.vals.at(0);
                continue;
            }
            if (entry.sym == "name")
            {
                I(entry.vals.size() == 1);
                cert.name = entry.vals.at(0);
                continue;
            }
            if (entry.sym == "value")
            {
                I(entry.vals.size() == 1);
                cert.value = entry.vals.at(0).trimmed();
                continue;
            }
            if (entry.sym == "signature")
            {
                I(entry.vals.size() == 1);
                QString sig = entry.vals.at(0);
                if (sig == "ok")
                    cert.signature = Cert::Ok;
                else if (sig == "bad")
                    cert.signature = Cert::Bad;
                else if (sig == "unknown")
                    cert.signature = Cert::Unknown;
                else
                    W(QString("Unknown cert signature state '%1'.").arg(sig));
                continue;
            }
            if (entry.sym == "trust")
            {
                I(entry.vals.size() == 1);
                QString trust = entry.vals.at(0);
                if (trust == "trusted")
                    cert.trust = Cert::Trusted;
                else if (trust == "untrusted")
                    cert.trust = Cert::Untrusted;
                else
                    W(QString("Unknown cert trust '%1'.").arg(trust));
                continue;
            }
        }
        if (cert.key.isEmpty())
            continue;
        add(cert);
    }
}

