/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "ContentDiff.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QFont>
#include <QBrush>

ContentDiff::ContentDiff(QObject * parent)
    : QAbstractItemModel(parent), AutomateCommand(0), diffParser(0)
{}

ContentDiff::~ContentDiff()
{
    cleanup();
}

void ContentDiff::cleanup()
{
    if (diffParser) delete diffParser;
    diffParser = 0;
    qDeleteAll(lines);
    lines.clear();
}

void ContentDiff::readDiff(const MonotoneHandlePtr & handle,
                           const QString & fileName,
                           const QString & base,
                           const QString & target,
                           const QString & enc)
{
    if (enc.isEmpty())
        encoding = "UTF-8";
    else
        encoding = enc;

    QStringList cmd;
    cmd << "content_diff";

    if (!fileName.isEmpty())
    {
        cmd << fileName;
    }

    QStringList revs;
    if (!base.isEmpty())
    {
        revs << base;
    }
    if (!target.isEmpty())
    {
        revs << target;
    }

    // sort base and target topologically in case the user accidently chosed
    // the child as base and the parent as target (and would therefor get a
    // reverse diff)
    if (revs.size() == 2 && Settings::getBool("FixUnwantedReverseDiffs"))
    {
        revs = MonotoneUtil::topsortRevisions(handle, revs);
    }

    QStringList opts;
    foreach (const QString & r, revs)
    {
        opts << "r" << r;
    }

    MonotoneTaskPtr task(new MonotoneTask(cmd, opts));
    AutomateCommand::enqueueTask(handle, task);
}

void ContentDiff::switchOutputEncoding(const QString & enc)
{
    if (!lastTask->isFinished())
    {
        D("task not yet finished, skip encoding switch");
        return;
    }

    encoding = enc;

    cleanup();
    loadOutputIntoModel(lastTask->getOutput(), enc);

    emit encodingChanged(enc);
}

void ContentDiff::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    cleanup();

    lastTask = task;
    loadOutputIntoModel(task->getOutput(), encoding);

    emit diffRead();
}

void ContentDiff::loadOutputIntoModel(const QByteArray & output, const QString & codec)
{
    I(!diffParser);
    diffParser = new DiffParser(output, codec);

    // flatten the data for the current view
    FileDiffs fileDiffs = diffParser->getAllDiffs();

    QMapIterator<QString, Diff *> i(fileDiffs);
    while (i.hasNext())
    {
        i.next();
        QString fileName = i.key();
        Diff * diff = i.value();

        ListLine * fileLine = new ListLine(fileName, diff->hunks.size());
        fileLine->isNewFile = diff->is_new_file;
        lines.append(fileLine);

        if (diff->is_binary)
        {
            fileLine->isBinary = true;
            continue;
        }

        for (int j=0, k=diff->hunks.size(); j<k; j++)
        {
            DiffHunk * hunk = diff->hunks.at(j);

            // The oldLineNumber is only raised for kept and old lines
            // in the following loop.
            // This way people should be able to navigate in the original
            // file and see the the changes even better.
            int oldLineNumber = hunk->leftStart;

            for (int m=0, n=hunk->lines.size(); m<n; m++)
            {
                DiffLine * line = hunk->lines.at(m);
                ListLine * hunkLine = new ListLine(ListLine::HunkLine);
                fileLine->addChild(hunkLine);

                switch (line->state)
                {
                    case DiffLine::Unchanged:
                        oldLineNumber++;
                        hunkLine->lineMarker = QString::number(oldLineNumber);
                        break;
                    case DiffLine::UnchangedMissingNewline:
                        oldLineNumber++;
                        hunkLine->lineMarker = QString("%1 (%2)").arg(oldLineNumber).arg(tr("missing \\n"));
                        break;
                    case DiffLine::Added:
                        hunkLine->lineMarker = QString("+");
                        break;
                    case DiffLine::AddedMissingNewline:
                        hunkLine->lineMarker = QString("+ (%1)").arg(tr("missing \\n"));
                        break;
                    case DiffLine::Removed:
                        oldLineNumber++;
                        hunkLine->lineMarker = QString("-");
                        break;
                    case DiffLine::RemovedMissingNewline:
                        oldLineNumber++;
                        hunkLine->lineMarker = QString("- (%1)").arg(tr("missing \\n"));
                        break;
                    default: I(false);
                }
                hunkLine->lineContent = line->content;
            }

            if (j < k-1)
            {
                ListLine * sep = new ListLine(ListLine::HunkSep);
                fileLine->addChild(sep);
            }
        }
    }

    reset();
}

int ContentDiff::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant ContentDiff::data(const QModelIndex & index, int role) const
{
    if (!index.isValid()) return QVariant();

    int col = index.column();
    ListLine * line = static_cast<ListLine *>(index.internalPointer());

    if (role == Qt::DisplayRole)
    {
        switch (col)
        {
            case 0: return QVariant(line->firstCol());
            case 1: return QVariant(line->secondCol());
        }

        I(false);
    }

    if (role == Qt::FontRole && col == 1)
    {
        QFont font;

        if (line->type == ListLine::FileLine)
        {
            font.setBold(true);
        }

        if (line->type == ListLine::HunkLine)
        {
            font.setStyleHint(QFont::Courier);
            font.setFamily("Courier");
        }

        return QVariant(font);
    }

    if (role == Qt::TextAlignmentRole && col == 0)
    {
        return QVariant(Qt::AlignRight);
    }

    if (role == Qt::BackgroundRole)
    {
        if (line->type != ListLine::HunkLine) return QVariant();

        if (line->lineMarker.startsWith("+"))
        {
            QColor added(Settings::getString("DiffColorAddedLines"));
            return QVariant(QBrush(added));
        }

        if (line->lineMarker.startsWith("-"))
        {
            QColor removed(Settings::getString("DiffColorRemovedLines"));
            return QVariant(QBrush(removed));
        }

        // unchanged lines
        return QVariant();
    }

    return QVariant();
}

Qt::ItemFlags ContentDiff::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant ContentDiff::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole || orientation != Qt::Horizontal)
        return QVariant();
    if (section == 0) return QVariant(tr("Line"));
    return QVariant(tr("File/Content"));
}

int ContentDiff::rowCount(const QModelIndex & parent) const
{
    if (!parent.isValid()) return lines.size();
    ListLine * line = static_cast<ListLine *>(parent.internalPointer());
    return line->lines.size();
}

QModelIndex ContentDiff::index(int row, int column, const QModelIndex & parent) const
{
    if (!parent.isValid())
    {
        if (row < 0 || row >= lines.size())
            return QModelIndex();
        ListLine * line = lines.at(row);
        return createIndex(row, column, line);
    }

    ListLine * parentLine = static_cast<ListLine*>(parent.internalPointer());
    if (row < 0 || row >= parentLine->lines.size())
        return QModelIndex();
    ListLine * line = parentLine->lines.at(row);
    return createIndex(row, column, line);
}

QModelIndex ContentDiff::parent(const QModelIndex & index) const
{
    if (!index.isValid()) return QModelIndex();
    ListLine * line = static_cast<ListLine*>(index.internalPointer());
    if (!line->parent) return QModelIndex();
    return createIndex(lines.indexOf(line->parent) , 0, line->parent);
}

