/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#ifndef CONTENT_DIFF_H
#define CONTENT_DIFF_H

#include "AutomateCommand.h"
#include "DiffParser.h"

#include <QString>
#include <QAbstractItemModel>

struct ListLine
{
    enum Type { FileLine, HunkSep, HunkLine };

    Type type;
    QString fileName;
    int hunkCount;
    bool isBinary;
    bool isNewFile;
    QString lineMarker;
    QString lineContent;

    ListLine * parent;
    QList<ListLine*> lines;

    ListLine(Type t) : type(t), parent(0) {}
    ListLine(const QString & file, int count)
        : type(FileLine), fileName(file), hunkCount(count),
          isBinary(false), parent(0) {}

    ~ListLine()
    {
        qDeleteAll(lines);
        lines.clear();
    }

    QString firstCol() const
    {
        switch (type)
        {
            case HunkSep: return QString("...");
            case FileLine: return QString();
            case HunkLine: return lineMarker;
            default: I(false);
        }
    }

    QString secondCol() const
    {
        switch (type)
        {
            case HunkSep: return QString();
            case FileLine: return isBinary
                ? QObject::tr("%1 (binary)").arg(fileName)
                : QObject::tr("%1 (%2 hunks)").arg(fileName).arg(hunkCount);
            case HunkLine: return lineContent;
            default: I(false);
        }
    }

    bool isFileLineAndDiffable()
    {
        return type == FileLine && !isBinary && !isNewFile;
    }

    void addChild(ListLine * l) { lines.append(l); l->parent = this; }
};

typedef QList<ListLine*> ListLines;

class ContentDiff : public QAbstractItemModel, public AutomateCommand
{
    Q_OBJECT
public:
    ContentDiff(QObject *);
    virtual ~ContentDiff();

    // needed Qt Model methods
    QVariant data(const QModelIndex &, int) const;
    Qt::ItemFlags flags(const QModelIndex &) const;
    QVariant headerData(int, Qt::Orientation, int) const;
    QModelIndex index(int, int, const QModelIndex &) const;
    QModelIndex parent(const QModelIndex &) const;
    int rowCount(const QModelIndex &) const;
    int columnCount(const QModelIndex &) const;

    inline Diff * getDiff(QString fileName) const
    {
        return diffParser->getDiff(fileName);
    }

    inline FileDiffs getAllDiffs() const
    {
        return diffParser->getAllDiffs();
    }

    inline QString getBase() const
    {
        return lastTask->getOptions().size() > 1 ?
            QString(lastTask->getOptions().at(1)) :
            "";
    }

    inline QString getTarget() const
    {
        return lastTask->getOptions().size() > 3 ?
            QString(lastTask->getOptions().at(3)) :
            "";
    }

public slots:
    void readDiff(const MonotoneHandlePtr &, const QString &, const QString &,
                  const QString &, const QString & encoding = QString());
    void switchOutputEncoding(const QString &);

signals:
    void diffRead();
    void encodingChanged(const QString &);

private:
    void processTaskResult(const MonotoneTaskPtr &);
    void loadOutputIntoModel(const QByteArray &, const QString &);
    void cleanup();

    DiffParser * diffParser;
    ListLines lines;
    MonotoneTaskPtr lastTask;
    QString encoding;
};

#endif
