/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GetAttributes.h"
#include "MonotoneUtil.h"
#include "BasicIOParser.h"

#include <QIcon>
#include <QFont>

GetAttributes::GetAttributes(QObject * parent)
    : QAbstractItemModel(parent), AutomateCommand(0), itemPath()
{}

GetAttributes::~GetAttributes() {}

void GetAttributes::revert()
{
    attributes.clear();
    itemPath = QString();
    reset();
}
void GetAttributes::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
}

void GetAttributes::readAttributes(const QString & path)
{
    I(!monotoneHandle.isNull());

    // clear current attributes list
    attributes.clear();
    // reset the view
    reset();

    // empty paths are not allowed, if we get one, we certainly
    // mean the workspace' root directory, which we can address
    // with "." as well
    if (path.isEmpty() && itemPath.isEmpty())
        itemPath = ".";
    else if (!path.isEmpty())
        itemPath = path;

    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_attributes" << itemPath));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void GetAttributes::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);

        Attribute attr;

        bool isItem = false;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            // we're now only interested in stanzas starting with a "attr" entry
            if (j == 0 && entry.sym != "attr") break;

            isItem = true;

            if (entry.sym == "attr")
            {
                I(entry.vals.size() == 2);
                attr.key = entry.vals.at(0);
                attr.value = entry.vals.at(1);
                continue;
            }

            if (entry.sym == "state")
            {
                I(entry.vals.size() == 1);
                QString state = entry.vals.at(0);

                if (state == "added")          attr.state = Attribute::Added;
                else if (state == "dropped")   attr.state = Attribute::Dropped;
                else if (state == "changed")   attr.state = Attribute::Changed;
                else if (state == "unchanged") attr.state = Attribute::Unchanged;
                else
                    W(QString("Unknown attribute state '%1'.").arg(state));

                continue;
            }
            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        // check if we really processed an item entry
        if (!isItem) continue;
        attributes.append(attr);
    }

    // reset the view
    reset();

    // signal that we've finished (whoever listens to that)
    emit attributesRead();
}

int GetAttributes::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 3;
}

QVariant GetAttributes::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    int row = index.row(), col = index.column();
    if (row >= attributes.size()) return QVariant();

    Attribute attr = attributes.at(row);

    if (role == Qt::DisplayRole)
    {
        switch (col)
        {
            case 1: return QVariant(attr.key.length() == 0 ? tr("empty") : attr.key);
            case 2: return QVariant(attr.value.length() == 0 ? tr("empty") : attr.value);
            default: return QVariant();
        }
    }

    if (role == Qt::ToolTipRole)
    {
        switch (attr.state)
        {
            case Attribute::Added:     return QVariant(tr("added"));
            case Attribute::Dropped:   return QVariant(tr("dropped"));
            case Attribute::Changed:   return QVariant(tr("changed"));
            case Attribute::Unchanged: return QVariant(tr("unchanged"));
            default: return QVariant();
        }
    }

    if (role == Qt::DecorationRole && col == 0)
    {
        switch (attr.state)
        {
            case Attribute::Added:     return QVariant(QIcon(":/icons/green_dot.png"));
            case Attribute::Dropped:   return QVariant(QIcon(":/icons/red_dot.png"));
            case Attribute::Unchanged: return QVariant(QIcon(":/icons/yellow_dot.png"));
            case Attribute::Changed:   return QVariant(QIcon(":/icons/blue_dot.png"));
            default: return QVariant();
        }
    }

    if (role == Qt::FontRole)
    {
        QFont font;
        font.setItalic(true);

        if (col == 1 && attr.key.length() == 0)
            return QVariant(font);

        if (col == 2 && attr.value.length() == 0)
            return QVariant(font);

        return QVariant();
    }

    if (role == Qt::UserRole)
    {
        switch (col)
        {
            case 1: return QVariant(attr.key);
            case 2: return QVariant(attr.value);
            default: return QVariant();
        }
    }

    return QVariant();
}

Qt::ItemFlags GetAttributes::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant GetAttributes::headerData(int section, Qt::Orientation orientation, int role) const
{
    I(orientation == Qt::Horizontal);

    if (role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(QChar(9662)); // black arrow
            case 1: return QVariant(tr("Key"));
            case 2: return QVariant(tr("Value"));
            default: return QVariant();
        }
    }

    return QVariant();
}

int GetAttributes::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return attributes.size();
}

QModelIndex GetAttributes::index(int row, int column, const QModelIndex & parent) const
{
    return hasIndex(row, column, parent) ? createIndex(row, column, 0) : QModelIndex();
}

QModelIndex GetAttributes::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

bool GetAttributes::setAttribute(const QString & key, const QString & value)
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "set_attribute" << itemPath << key << value
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    return task->getReturnCode() == 0;
}

bool GetAttributes::dropAttribute(const QString & key)
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "drop_attribute" << itemPath << key
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    return task->getReturnCode() == 0;
}

bool GetAttributes::dropAllAttributes()
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "drop_attribute" << itemPath
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    return task->getReturnCode() == 0;
}

bool GetAttributes::hasDroppableAttributes() const
{
    foreach(const Attribute & attr, attributes)
    {
        if (attr.state != Attribute::Dropped)
        {
            return true;
        }
    }
    return false;
}

