/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GetDatabaseVariables.h"
#include "BasicIOParser.h"
#include "MonotoneUtil.h"

GetDatabaseVariables::GetDatabaseVariables(QObject * parent)
  : QAbstractItemModel(parent), AutomateCommand(0)
{}

GetDatabaseVariables::~GetDatabaseVariables() {}

void GetDatabaseVariables::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
}

void GetDatabaseVariables::readVariables()
{
    I(!monotoneHandle.isNull());

    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_db_variables"));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void GetDatabaseVariables::processTaskResult(const MonotoneTaskPtr & task)
{
    QString output = task->getOutput();
    if (task->getReturnCode() != 0)
    {
        if (output.indexOf("No variables found") != -1)
        {
            L("no database variables found");
            reset();
            return;
        }

        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    reset();
    variables.clear();

    BasicIOParser parser(output);
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);

        QString domain;
        QList<QPair<QString, QString> > entries;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            if (entry.sym == "domain")
            {
                I(entry.vals.size() == 1);
                domain = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "entry")
            {
                I(entry.vals.size() == 2);
                entries.push_back(
                    QPair<QString, QString>(entry.vals.at(0), entry.vals.at(1))
                );
                continue;
            }

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        I(!domain.isEmpty());
        I(entries.size() > 0);
        variables.insert(domain, entries);
    }

    reset();
    emit variablesRead();
}

int GetDatabaseVariables::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant GetDatabaseVariables::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    if (role != Qt::DisplayRole)
    {
        return QVariant();
    }

    int row = index.row();

    QStringList keys = variables.keys();

    if (!index.parent().isValid())
    {
        I(row < keys.size());
        return QVariant(keys.at(row));
    }

    int parentRow = index.internalId();
    I(parentRow < keys.size());
    QString domain = keys.at(parentRow);
    QPair<QString, QString> entry = variables.value(domain).at(row);

    switch (index.column())
    {
        case 0: return QVariant(entry.first);
        case 1: return QVariant(entry.second);
    }

    return QVariant();
}

Qt::ItemFlags GetDatabaseVariables::flags(const QModelIndex & index) const
{
    if (index.isValid() && index.parent().isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return Qt::ItemIsEnabled;
}

QVariant GetDatabaseVariables::headerData(int section,
                                          Qt::Orientation orientation,
                                          int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Domain / Key"));
            case 1: return QVariant(tr("Value"));
        }
    }
    return QVariant();
}

int GetDatabaseVariables::rowCount(const QModelIndex & parent) const
{
    QStringList keys = variables.keys();

    if (parent.isValid())
    {
        if (parent.internalId() == -1)
        {
            int row = parent.row();
            I(row < keys.size());
            return variables.value(keys.at(row)).size();
        }
        // an entry node
        return 0;
    }

    return keys.size();
}

QModelIndex GetDatabaseVariables::index(int row, int column,
                                        const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    if (parent.isValid())
    {
        QStringList keys = variables.keys();
        int parentRow = parent.row();
        I(parentRow < keys.size());
        return createIndex(row, column, parentRow);
    }
    return createIndex(row, column, -1);
}

QModelIndex GetDatabaseVariables::parent(const QModelIndex & index) const
{
    if (!index.isValid() || index.internalId() == -1)
    {
        return QModelIndex();
    }

    int row = index.internalId();
    return createIndex(row, 0, -1);
}

bool GetDatabaseVariables::setVariable(const QString & domain,
                                       const QString & key,
                                       const QString & value)
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "set_db_variable" << domain << key << value
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    return task->getReturnCode() == 0;
}

bool GetDatabaseVariables::dropVariable(const QString & domain,
                                        const QString & key)
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "drop_db_variables" << domain << key
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    return task->getReturnCode() == 0;
}

