/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GetFile.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QBrush>
#include <QFont>

GetFile::GetFile(QObject * parent)
  : QAbstractItemModel(parent), AutomateCommand(0)
{}

GetFile::~GetFile() {}

void GetFile::readFileByName(const MonotoneHandlePtr & handle, const QString & fileName,
                             const QString & rev, const QString & encoding)
{
    QStringList cmd;
    cmd << "get_file_of" << fileName;

    QStringList opts;
    opts << "r" << rev;

    readFile(handle, cmd, opts, encoding);
}

void GetFile::readFileById(const MonotoneHandlePtr & handle, const QString & fileID,
                           const QString & encoding)
{
    QStringList cmd;
    cmd << "get_file" << fileID;

    readFile(handle, cmd, QStringList(), encoding);
}

void GetFile::readFile(const MonotoneHandlePtr & handle, const QStringList & cmd,
                       const QStringList & opts, const QString & encoding)
{
    reset();
    MonotoneTaskPtr task(new MonotoneTask(cmd, opts));

    if (!encoding.isEmpty())
    {
        task->setOutputEncoding(encoding);
    }

    AutomateCommand::enqueueTask(handle, task);
}

void GetFile::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }
    lastTask = task;

    loadOutputIntoModel(task->getDecodedOutput());

    emit fileRead();
}

void GetFile::switchOutputEncoding(const QString & codec)
{
    if (!lastTask->isFinished())
    {
        D("task not yet finished, skip encoding switch");
        return;
    }

    loadOutputIntoModel(lastTask->getDecodedOutput(codec));

    emit encodingChanged(codec);
}

void GetFile::loadOutputIntoModel(const QString & output)
{
    QRegExp rx("\\0x00");
    if (rx.indexIn(output) != -1)
    {
        // FIXME: signal this anywhere...
        W("File is binary");
        return;
    }

    Content oldContents = fileContents;
    fileContents.clear();

    QStringList lines(output.split(QRegExp("\\n|\\r\\n")));
    for (int i=0, s=lines.size(); i<s; i++)
    {
        fileContents.push_back(
            new ContentLine(lines.at(i), DiffLine::Unchanged, i+1, i+1)
        );
    }

    reset();
    qDeleteAll(oldContents);
}

void GetFile::applyDiff(Diff * diff)
{
    int addedLines = 0;

    for (int i=0, s=diff->hunks.size(); i<s; i++)
    {
        DiffHunk * hunk = diff->hunks.at(i);
        int lineStart = hunk->leftStart + addedLines;

        if (lineStart == 0)
        {
            lineStart = 1;
        }

        int leftLineCount = hunk->leftStart;
        int rightLineCount = hunk->rightStart;

        for (int j=0, t=hunk->lines.size(); j<t; j++)
        {
            DiffLine * line = hunk->lines.at(j);

            switch (line->state)
            {
                case  DiffLine::Unchanged:
                    fileContents.value(lineStart - 1)->leftLineCount = leftLineCount;
                    fileContents.value(lineStart - 1)->rightLineCount = rightLineCount;
                    leftLineCount++;
                    rightLineCount++;
                    break;

                case DiffLine::UnchangedMissingNewline:
                case DiffLine::Removed:
                case DiffLine::RemovedMissingNewline:
                    fileContents.value(lineStart - 1)->state = line->state;
                    fileContents.value(lineStart - 1)->leftLineCount = leftLineCount;
                    fileContents.value(lineStart - 1)->rightLineCount = 0;
                    leftLineCount++;
                    break;

                case DiffLine::Added:
                case DiffLine::AddedMissingNewline:
                    fileContents.insert(
                        lineStart - 1,
                        new ContentLine(line->content, line->state, 0, rightLineCount)
                    );
                    rightLineCount++;
                    addedLines++;
                    break;

                default: I(false);
            }
            lineStart++;
        }

        for (int i=lineStart-1; i<fileContents.size(); i++)
        {
            fileContents.value(i)->rightLineCount = rightLineCount++;
        }
    }

    reset();
}

QModelIndex GetFile::getNextGroup(const QModelIndex & base, bool recurse)
{
    int startRow = 0;
    bool startedInChange = false;
    if (base.isValid() && base.row() < fileContents.size())
    {
        startRow = base.row();
        ContentLine * line = fileContents.at(startRow);
        startedInChange = line->state != DiffLine::Unchanged &&
                          line->state != DiffLine::UnchangedMissingNewline;
    }

    for (int i=startRow, j=fileContents.size(); i<j; i++)
    {
        ContentLine * line = fileContents.at(i);
        bool inChange = line->state != DiffLine::Unchanged &&
                        line->state != DiffLine::UnchangedMissingNewline;

        if (( startedInChange &&  inChange) ||
            (!startedInChange && !inChange))
            continue;

        if (startedInChange && !inChange)
        {
            // if we started in a change, we obviously want to go to the
            // next change which should follow after at least one
            // unchanged line
            startedInChange = false;
            continue;
        }

        return index(i, 0, QModelIndex());
    }

    // if we already recursed, stop here
    if (recurse) return QModelIndex();
    // try to get the first group
    return getNextGroup(QModelIndex(), true);
}

QModelIndex GetFile::getPrevGroup(const QModelIndex & base, bool recurse)
{
    int startRow = fileContents.size() - 1;
    bool startedInChange = false;
    if (base.isValid() && base.row() < fileContents.size())
    {
        startRow = base.row();
        ContentLine * line = fileContents.at(startRow);
        startedInChange = line->state != DiffLine::Unchanged &&
                          line->state != DiffLine::UnchangedMissingNewline;
    }

    bool foundLastLine = false;
    for (int i=startRow; i>=0; i--)
    {
        ContentLine * line = fileContents.at(i);
        bool inChange = line->state != DiffLine::Unchanged &&
                        line->state != DiffLine::UnchangedMissingNewline;

        if (( startedInChange &&  inChange) ||
            (!startedInChange && !inChange))
            continue;

        if (startedInChange && !inChange)
        {
            // if we already found the last line, then this is the sign
            // that we've just left the hunk we want to scroll to next,
            // so we're returning the previously processed line as scrol
            // target
            if (foundLastLine)
            {
                return index(i + 1, 0, QModelIndex());
            }

            // if we started in a change, we obviously want to go to the
            // prev change which should be separated with at least one
            // unchanged line
            startedInChange = false;
            continue;
        }

        // now we've seem to have found the last line of the previous
        // hunk, we just need to find the first line and return that
        // and we just use the same machinery for this ... we start in a
        // change and go up until we're no longer in this change...
        foundLastLine = true;
        startedInChange = true;
    }

    // if we've already recursed, stop here
    if (recurse) return QModelIndex();
    // try to get the last group
    return getPrevGroup(QModelIndex(), true);
}

int GetFile::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 3;
}

QVariant GetFile::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    int row = index.row();
    if (row >= fileContents.size()) return QVariant();

    ContentLine * line = fileContents.at(row);

    if (role == Qt::DisplayRole)
    {
        switch (index.column())
        {
            case 0: return QVariant(line->leftLineCount == 0 ?
                                    "-" : QString::number(line->leftLineCount));
            case 1: return QVariant(line->rightLineCount == 0 ?
                                    "-" : QString::number(line->rightLineCount));
            case 2: return QVariant(line->content);
            default: return QVariant();
        }
    }
    else
    if (role == Qt::BackgroundRole)
    {
        QColor added(Settings::getString("DiffColorAddedLines"));
        QColor removed(Settings::getString("DiffColorRemovedLines"));

        switch (line->state)
        {
            case DiffLine::Added:
            case DiffLine::AddedMissingNewline:
                return QVariant(QBrush(added));
            case DiffLine::Removed:
            case DiffLine::RemovedMissingNewline:
                return QVariant(QBrush(removed));
            default: return QVariant();
        }
    }
    else
    if (role == Qt::ToolTipRole)
    {
        // FIXME: do people want to see this more prominently?
        switch (line->state)
        {
            case DiffLine::AddedMissingNewline:
            case DiffLine::RemovedMissingNewline:
            case DiffLine::UnchangedMissingNewline:
                return QVariant(tr("No newline at the end of this line"));
            default: return QVariant();
        }
    }
    else
    if (role == Qt::FontRole && index.column() > 1)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }
    else
    if (role == Qt::TextAlignmentRole && index.column() < 2)
    {
        return QVariant(Qt::AlignRight);
    }
    return QVariant();
}

Qt::ItemFlags GetFile::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled;
}

QVariant GetFile::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_UNUSED(orientation);

    if (role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Left Line No"));
            case 1: return QVariant(tr("Right Line No"));
            case 2: return QVariant(tr("Content"));
        }
    }

    return QVariant();
}

int GetFile::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return fileContents.size();
}

QModelIndex GetFile::index(int row, int column, const QModelIndex& parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    ContentLine * line = fileContents.at(row);
    return createIndex(row, column, line);
}

QModelIndex GetFile::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

