/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GetRevision.h"
#include "vocab.h"
#include "MonotoneUtil.h"

#include <QFont>
#include <QBrush>
#include <QTextStream>
#include <QSet>

GetRevision::GetRevision(QObject * parent)
    : QAbstractItemModel(parent), AutomateCommand(0)
{}

GetRevision::~GetRevision() {}

void GetRevision::readExistingRevision(const MonotoneHandlePtr & handle,
                                       const QString & rev)
{
    revision.clear();
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "get_revision" << rev
    ));
    AutomateCommand::enqueueTask(handle, task);
}

void GetRevision::readWorkspaceRevision(const MonotoneHandlePtr & handle,
                                        const QStringList & paths)
{
    I(handle->getType() == MonotoneHandle::workspace_handle);
    revision.clear();

    QStringList args;
    args << "get_current_revision";
    if (paths.size() > 0)
    {
        args << paths;
    }

    MonotoneTaskPtr task(new MonotoneTask(args));
    AutomateCommand::enqueueTask(handle, task);
}

void GetRevision::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        emit readFailed(MonotoneUtil::stripMtnPrefix(task->getDecodedOutput()));
        return;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    StanzaList list = parser.getStanzas();
    QString currentRevision;

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);
        Change change;
        bool found_change = false;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            if (i == 0 && j == 0)
            {
                if (entry.sym != "format_version")
                {
                    W("First entry in first stanza should be used format version.");
                } else
                if (entry.vals.at(0) != "1")
                {
                    W("Format version is unequal to '1' - newer interface version?");
                }
                break;
            }

            if (j == 0 && entry.sym == "new_manifest")
            {
                I(entry.vals.size() == 1);
                revision.new_manifest = entry.vals.at(0);
                break;
            }

            if (entry.sym == "old_revision")
            {
                I(entry.vals.size() == 1);
                currentRevision = entry.vals.at(0);
                QList<Change> changelist;
                revision.changesAgainstParent.insert(currentRevision, changelist);
                break;
            }

            found_change = true;

            if (entry.sym == "delete")
            {
                I(entry.vals.size() == 1);
                change.type = Delete;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "rename")
            {
                I(entry.vals.size() == 1);
                change.type = Rename;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "add_dir")
            {
                I(entry.vals.size() == 1);
                change.type = AddDir;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "add_file")
            {
                I(entry.vals.size() == 1);
                change.type = AddFile;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "patch")
            {
                I(entry.vals.size() == 1);
                change.type = Patch;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "clear")
            {
                I(entry.vals.size() == 1);
                change.type = AttrClear;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "set")
            {
                I(entry.vals.size() == 1);
                change.type = AttrSet;
                change.stanza = stanza;
                break;
            }

            found_change = false;

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        // check if we really processed an item entry
        if (!found_change) continue;
        I(revision.changesAgainstParent.contains(currentRevision));
        revision.changesAgainstParent[currentRevision].append(change);
    }

    // set the first as the default parent against which changes are displayed
    currentParentRevision = revision.changesAgainstParent.keys().at(0);

    // reset the view
    reset();

    // signal that we've finished (whoever listens to that)
    emit revisionRead();
}

int GetRevision::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant GetRevision::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
        return QVariant();

    if (!revision.changesAgainstParent.contains(currentParentRevision))
        return QVariant();

    Change change =
        revision.changesAgainstParent[currentParentRevision].at(index.row());

    if (role == Qt::DisplayRole)
    {
        switch (index.column())
        {
            case 0: return QVariant(change.getTypeString());
            case 1: return QVariant(change.getDisplayData());
            default: I(false);
        }
    }
    else
    if (role == Qt::BackgroundRole)
    {
        if (index.column() == 0)
        {
            return QVariant(QBrush(change.getTypeColor()));
        }
        return QVariant();
    }
    else
    if (role == Qt::FontRole && index.column() == 0)
    {
        QFont font;
        font.setPointSize(9);
        return QVariant(font);
    }
    else
    if (role == Qt::TextAlignmentRole && index.column() == 0)
    {
        return QVariant(Qt::AlignCenter);
    }
    return QVariant();
}

Qt::ItemFlags GetRevision::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant GetRevision::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_UNUSED(orientation);
    if (role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Type"));
            case 1: return QVariant(tr("Change"));
        }
    }

    return QVariant();
}

int GetRevision::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (!revision.changesAgainstParent.contains(currentParentRevision))
        return 0;

    return revision.changesAgainstParent[currentParentRevision].size();
}

QModelIndex GetRevision::index(int row, int column, const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (!revision.changesAgainstParent.contains(currentParentRevision))
        return QModelIndex();

    if (row >= revision.changesAgainstParent[currentParentRevision].size())
        return QModelIndex();

    return createIndex(row, column);
}

QModelIndex GetRevision::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

QStringList GetRevision::getParentRevisions() const
{
    QStringList parents = revision.changesAgainstParent.keys();
    if (parents.size() == 1 && parents.at(0) == "")
        return QStringList();
    return parents;
}

void GetRevision::showChangesAgainstParent(const QString & parent)
{
    QStringList parents = getParentRevisions();
    if (parents.size() == 0)
    {
        currentParentRevision = QString("");
    }
    else
    {
        if (!parents.contains(parent))
            return;
        currentParentRevision = parent;
    }
    reset();
}

QMap<QString, QPair<QString, QString> > GetRevision::getChangedFiles() const
{
    QMap<QString, QPair<QString, QString> > fileMap;

    if (revision.changesAgainstParent.contains(currentParentRevision))
    {
        QList<Change> changes =
            revision.changesAgainstParent.value(currentParentRevision);
        foreach (const Change & ch, changes)
        {
            Stanza st = ch.stanza;
            QString path;
            QPair<QString, QString> hashes;
            bool isPatch = false;
            bool isAdd = false;

            foreach (const StanzaEntry & en, st)
            {
                if (en.sym == "add_file")
                {
                    path = en.vals.at(0);
                    isAdd = true;
                    continue;
                }

                if (isAdd && en.sym == "content")
                {
                    I(en.vals.size() == 1);
                    hashes.first = en.vals.at(0);
                    fileMap.insert(path, hashes);
                    break;
                }

                if (en.sym == "patch")
                {
                    path = en.vals.at(0);
                    isPatch = true;
                    continue;
                }

                if (isPatch && en.sym == "from")
                {
                    I(en.vals.size() == 1);
                    hashes.second = en.vals.at(0);
                    continue;
                }

                if (isPatch && en.sym == "to")
                {
                    I(en.vals.size() == 1);
                    hashes.first = en.vals.at(0);
                    fileMap.insert(path, hashes);
                    break;
                }
            }
        }
    }

    return fileMap;
}

QString GetRevision::getRevisionText() const
{
    return revision.print();
}

QString GetRevision::getCurrentParentRevision() const
{
    return currentParentRevision;
}

QStringList GetRevision::getPathsForCompletion() const
{
    QSet<QString> paths;
    QMapIterator<QString, QList<Change> > it(revision.changesAgainstParent);

    while (it.hasNext())
    {
        it.next();
        foreach (const Change & ch, it.value())
        {
            QStringList fullPaths;
            switch (ch.type)
            {
                case Rename:
                    fullPaths.push_back(ch.stanza.at(1).vals.at(0));
                case AddDir:
                case AddFile:
                case Patch:
                case Delete:
                    fullPaths.push_back(ch.stanza.at(0).vals.at(0));
                    break;
                default:
                    // ignore attribute changes
                    break;
            }

            foreach (const QString & path, fullPaths)
            {
                if (path.isEmpty()) continue;
                paths.insert(path);
            }
        }
    }

    QStringList pathList = paths.toList();
    qSort(pathList.begin(), pathList.end());

    return pathList;
}

