/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#ifndef GET_REVISION_H
#define GET_REVISION_H

#include "AutomateCommand.h"
#include "BasicIOParser.h"
#include "BasicIOWriter.h"

#include <QAbstractItemModel>

enum ChangeType { Delete, Rename, AddDir, AddFile, Patch, AttrSet, AttrClear };

struct Change {
    ChangeType type;
    Stanza stanza;

    inline QString getTypeString()
    {
        switch (type)
        {
            case Delete:    return QObject::tr("Deleted");
            case Rename:    return QObject::tr("Renamed");
            case AddDir:    return QObject::tr("Dir added");
            case AddFile:   return QObject::tr("File added");
            case Patch:     return QObject::tr("Patched");
            case AttrSet:   return QObject::tr("Attr set");
            case AttrClear: return QObject::tr("Attr cleared");
        }
        return QString();
    }

    inline Qt::GlobalColor getTypeColor()
    {
        switch (type)
        {
            case Delete:    return Qt::red;
            case Rename:    return Qt::gray;
            case AddDir:    return Qt::darkGreen;
            case AddFile:   return Qt::green;
            case Patch:     return Qt::yellow;
            case AttrSet:   return Qt::cyan;
            case AttrClear: return Qt::magenta;
        }
        return Qt::transparent;
    }

    // the workspace root is an empty string - make it more recognizable
    inline QString getPathName(const QString path)
    {
        if (path.isEmpty())
        {
            return QObject::tr("<workspace root>");
        }
        return path;
    }

    inline QString getDisplayData()
    {

        switch (type)
        {
            case AddDir:
            case AddFile:
            case Patch:
            case Delete:
                return getPathName(stanza.at(0).vals.at(0));
            case Rename:
                return QObject::tr("%1 to\n%2")
                    .arg(getPathName(stanza.at(0).vals.at(0)))
                    .arg(stanza.at(1).vals.at(0));
            case AttrClear:
                return QObject::tr("'%1' from\n%2")
                    .arg(stanza.at(1).vals.at(0))
                    .arg(getPathName(stanza.at(0).vals.at(0)));
            case AttrSet:
                return QObject::tr("'%1' to\n'%2' for\n%3")
                    .arg(stanza.at(1).vals.at(0))
                    .arg(stanza.at(2).vals.at(0))
                    .arg(getPathName(stanza.at(0).vals.at(0)));
        }
        return QString();
    }
};

struct Revision
{
    QString new_manifest;
    QMap<QString, QList<Change > > changesAgainstParent;

    inline void clear()
    {
        new_manifest.clear();
        changesAgainstParent.clear();
    }

    inline QString print() const
    {
        StanzaList list;

        StanzaEntry format("format_version", QStringList() << "1");
        Stanza formst;
        formst.append(format);
        list.append(formst);

        StanzaEntry manifest("new_manifest", "[" + new_manifest + "]");
        Stanza manst;
        manst.append(manifest),
        list.append(manst);

        QMapIterator<QString, QList<Change> > it(changesAgainstParent);
        while (it.hasNext())
        {
            it.next();

            StanzaEntry revision("old_revision", "[" + it.key() + "]");
            Stanza revst;
            revst.append(revision),
            list.append(revst);

            foreach (const Change & ch, it.value())
            {
                Stanza st(ch.stanza);
                if (st[0].sym == "add_file")
                {
                    st[1].vals[0] = "[" + st[1].vals[0] + "]";
                }
                if (st[0].sym == "patch")
                {
                    st[1].vals[0] = "[" + st[1].vals[0] + "]";
                    st[2].vals[0] = "[" + st[2].vals[0] + "]";
                }
                list.append(st);
            }
        }

        BasicIOWriter writer(list);
        return writer.write();
    }
};

class GetRevision : public QAbstractItemModel, public AutomateCommand
{
    Q_OBJECT
public:
    GetRevision(QObject *);
    virtual ~GetRevision();
    QStringList getParentRevisions() const;

    // needed Qt Model methods
    QVariant data(const QModelIndex &, int) const;
    Qt::ItemFlags flags(const QModelIndex &) const;
    QVariant headerData(int, Qt::Orientation, int) const;
    QModelIndex index(int, int, const QModelIndex &) const;
    QModelIndex parent(const QModelIndex &) const;
    int rowCount(const QModelIndex &) const;
    int columnCount(const QModelIndex &) const;
    QMap<QString, QPair<QString, QString> > getChangedFiles() const;
    QString getRevisionText() const;
    QString getCurrentParentRevision() const;
    QStringList getPathsForCompletion() const;

public slots:
    void readExistingRevision(const MonotoneHandlePtr &, const QString &);
    void readWorkspaceRevision(const MonotoneHandlePtr &, const QStringList &);
    void showChangesAgainstParent(const QString &);

signals:
    void revisionRead();
    void readFailed(const QString &);

private:
    void processTaskResult(const MonotoneTaskPtr &);

    Revision revision;
    QString currentParentRevision;
    MonotoneHandlePtr monotoneHandle;
};

#endif

