/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Inventory.h"
#include "MonotoneUtil.h"
#include "BasicIOParser.h"
#include "Settings.h"

#include <QApplication>
#include <QSet>

const int Inventory::QueryLevel = 2;

Inventory::Inventory(QObject * parent)
  : QObject(parent), AutomateCommand(0)
{
    rootItem = new ModelItem("__root__");
    rootItem->setParent(rootItem);
}

Inventory::~Inventory()
{
    delete rootItem;
}

void Inventory::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
}

void Inventory::read(const QString & path, bool fullQuery)
{
    I(!monotoneHandle.isNull());

    QStringList cmd = QStringList() << "inventory";

    QString query(path);
    // remove the trailing slash from the query if there is any.
    // there is usually one for automated queries from the filesystem
    // watcher. we explicitely mark directories for it with a trailing
    // slash to be able to distinguish them from normal files for the
    // lru cache.
    if (query.right(1) == "/")
        query.chop(1);

    if (!query.isEmpty())
        cmd << query;

    QStringList opts;

    // if requested, just read one directory level ahead
    // (useful for big directory trees)
    if (!fullQuery && Settings::getBool("ReadWorkspaceIncrementally"))
    {
        opts << "depth" << QString::number(QueryLevel);
        opts << "no-corresponding-renames" << QString();
    }

    MonotoneTaskPtr task(new MonotoneTask(cmd, opts));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

/*!
    @fixme
    The whole item adding procedure could eventually make problems
    if inventory is executed with a path which was renamed. The
    inventory output then contains another stanza of a path which
    was not yet read in, so we have no plan where to tack this on.
    A possible resolution would be to enqueue a new inventory
    request with the baseDir right here, delete the node and
    continue with the next one
*/
void Inventory::processTaskResult(const MonotoneTaskPtr & task)
{
    QString queriedPath = "";
    if (task->getArguments().size() > 1)
    {
        queriedPath = task->getArguments().at(1);
    }

    int queriedDepth = -1;
    if (task->getOptions().size() > 1)
    {
        queriedDepth = task->getOptions().at(1).toInt();
        if (!queriedPath.isEmpty())
        {
            queriedDepth += queriedPath.split("/", QString::SkipEmptyParts).size();
        }
    }

    QString output = task->getDecodedOutput();

    // internal errors
    if (task->getReturnCode() == 1)
    {
        C(QString("Command returned with an error: %1").arg(output));
        return;
    }

    // user errors
    if (task->getReturnCode() == 2)
    {
        // if the path we've silently queried became unknown (i.e. was removed),
        // remove it from the model. Note however that automate inventory does
        // not tell us if the node exists on the filesystem - in this case we
        // want to _keep_ the node because otherwise people won't be able to
        // add unknown paths later on...
        if (task->outOfBandMessagesMatch(
                MonotoneTask::Warning,
                QRegExp("restriction includes unknown path")
            ) && itemMap.contains(queriedPath))
        {
            QString workspace = monotoneHandle->getData();
            if (QFile::exists(workspace + "/" + itemMap[queriedPath]->getPath()))
            {
                D(QString("keeping existing unknown node %1").arg(queriedPath));
                // clear the previous state of the node and its children
                itemMap[queriedPath]->setStatusRecursive(InventoryItem::Unknown);
                emit dataChanged(itemMap[queriedPath]);
                return;
            }

            D(QString("clean out removed node %1").arg(queriedPath));
            InventoryItem * oldItem = itemMap.value(queriedPath);
            ModelItem * oldParent = oldItem->parent();
            int row = oldItem->row();
            removeRowsRecursive(oldParent, row, row);
            return;
        }

        // everything else _should_ be a problem with the workspace format.
        // note that we do not explicitely check for this, we throw it just
        // back to the user and hopefully log something useful
        emit invalidWorkspaceFormat();
        return;
    }

    BasicIOParser parser(output);
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();

    ModelItem * parentItem = rootItem;
    if (itemMap.contains(queriedPath))
    {
        // if we just queried one item we replace it directly underknees
        // the appropriate parent record, if not, we re-use the original
        // parent (i.e. skip new data we get from inventory here) and just
        // include the data _underknees_ this node
        if (stanzas.size() == 1)
        {
            parentItem = itemMap.value(queriedPath)->parent();
        }
        else
        {
            stanzas.removeFirst();
            parentItem = itemMap.value(queriedPath);
        }
    }

    //
    // For the lazy population code to work properly, Qt needs to get
    // all items added within their very own beginInsertRows and
    // endInsertRows. To prevent that we need to call those two for each
    // single item, we do not yet set the parent / child dependencies just
    // now, but record them temporarily in parentChildRelations.
    // If we've processed all stanzas, we rework these relations into
    // actual relations and notify Qt about that in insertRowsRecursive
    //
    QMap<ModelItem *, QList<ModelItem *> > parentChildRelations;
    QMap<QString, ModelItem *> tempItemMap;
    foreach (const Stanza & st, stanzas)
    {
        InventoryItem * item = new InventoryItem(st);

        QString path = item->getPath();
        int itemDepth = 0;
        if (!path.isEmpty())
        {
            itemDepth = path.split("/", QString::SkipEmptyParts).size();
        }
        else
        {
            QString displayPath = monotoneHandle->getData();
            QString homePath = QDir::homePath();
            if (displayPath.indexOf(homePath) == 0)
            {
                displayPath = QDir(homePath).relativeFilePath(displayPath);
                displayPath.prepend("~/");
            }
            item->setLabel(displayPath);
        }

        // add a cdup item for this item if we either queried everything
        // already or the current directory level corresponds to the queried one
        if (item->isDirectory() && (queriedDepth == -1 || queriedDepth > itemDepth))
        {
            PseudoItem * cdUp = new PseudoItem("..", PseudoItem::CdUp);
            parentChildRelations[item].append(cdUp);
        }

        QString baseDir = item->getBaseDirectory();

        ModelItem * parent = parentItem;
        if (tempItemMap.contains(baseDir))
        {
            parent = tempItemMap.value(baseDir);
        }

        parentChildRelations[parent].append(item);
        tempItemMap.insert(path, item);
    }

    insertRowsRecursive(parentItem, parentChildRelations);

    emit inventoryRead(queriedPath);
}

void Inventory::insertRowsRecursive(ModelItem * parentItem, const QMap<ModelItem *, QList<ModelItem *> > & parentChildRelations)
{
    I(parentItem);
    I(parentChildRelations.contains(parentItem));
    const QList<ModelItem *> & children = parentChildRelations.value(parentItem);
    I(children.size() > 0);

    //
    // remove existing new nodes, just to ensure we don't add duplicates
    //
    foreach (ModelItem * item, children)
    {
        InventoryItem * invitem = qobject_cast<InventoryItem*>(item);
        if (!invitem) continue;

        QString path = invitem->getPath();
        if (itemMap.contains(path))
        {
            InventoryItem * oldItem = itemMap.value(path);
            ModelItem * oldParent = oldItem->parent();
            int row = oldItem->row();
            removeRowsRecursive(oldParent, row, row);
        }
    }

    //
    // now additionally remove all those nodes which are no longer part of
    // the tree, but still attached to the old parent item
    //
    InventoryItem * newInventoryParent = qobject_cast<InventoryItem *>(parentItem);
    if (newInventoryParent)
    {
        InventoryItem * oldInventoryParent = itemMap.value(newInventoryParent->getPath());
        if (oldInventoryParent)
        {
            foreach (ModelItem * item, oldInventoryParent->getChildren())
            {
                InventoryItem * invitem = qobject_cast<InventoryItem*>(item);
                if (!invitem) continue;

                QString path = invitem->getPath();
                if (itemMap.contains(path))
                {
                    InventoryItem * oldItem = itemMap.value(path);
                    ModelItem * oldParent = oldItem->parent();
                    int row = oldItem->row();
                    removeRowsRecursive(oldParent, row, row);
                }
            }
        }
    }

    //
    // now do the actual row insertion while notifying Qt properly about that
    //
    int parentRowCount = parentItem->childCount();

    emit beginInsertRows(parentItem, parentRowCount, parentRowCount + children.size() - 1);

    foreach (ModelItem * item, children)
    {
        InventoryItem * invitem = qobject_cast<InventoryItem*>(item);
        if (invitem)
        {
            QString path = invitem->getPath();
            I(!itemMap.contains(path));
            itemMap.insert(path, invitem);
        }
        parentItem->appendChild(item);
    }

    emit endInsertRows();

    //
    // finally check if the just added items have children we need
    // to insert recursively
    //
    foreach (ModelItem * item, children)
    {
        if (parentChildRelations.contains(item))
            insertRowsRecursive(item, parentChildRelations);
    }
}

void Inventory::removeRowsRecursive(ModelItem * parent, int startRow, int endRow)
{
    I(parent);
    I(startRow >= 0);
    I(endRow >= startRow && endRow < parent->childCount());

    // we need to ensure that we recursively delete all childrens before we
    // actually delete the requested items for parent, otherwise Qt will
    // throw too many signals around
    QList<ModelItem *> itemsToDelete;
    for (int i=startRow; i<=endRow; i++)
    {
        ModelItem * child = parent->child(i);
        I(child);
        int childCount = child->childCount();
        if (childCount > 0)
        {
            removeRowsRecursive(child, 0, childCount - 1);
        }

        itemsToDelete.append(child);
    }

    emit beginRemoveRows(parent, startRow, endRow);

    foreach (ModelItem * item, itemsToDelete)
    {
        I(item);
        I(item->childCount() == 0);

        InventoryItem * invitem = qobject_cast<InventoryItem *>(item);
        if (invitem)
        {
            itemMap.remove(invitem->getPath());
        }

        item->parent()->removeChild(item);
        delete item;
    }

    emit endRemoveRows();
}

QMap<FileEntry, FileEntryList> Inventory::findUnaccountedRenames() const
{
    QList<InventoryItem *> missingItems;
    QList<InventoryItem *> unknownItems;

    foreach (InventoryItem * item, itemMap.values())
    {
        if (item->hasStatus(InventoryItem::Missing))
        {
            missingItems.append(item);
        }

        if (item->hasStatus(InventoryItem::Unknown))
        {
            unknownItems.append(item);
        }
    }

    // TODO: progess bar here!
    QString parentRev = MonotoneUtil::getBaseWorkspaceRevision(monotoneHandle);
    FileEntryList parentList =  MonotoneUtil::getRevisionManifest(
        monotoneHandle, parentRev
    );

    QMap<FileEntry, FileEntryList> unaccountedRenames;
    QMap<QString, QString> fileIds;

    foreach (InventoryItem * missingItem, missingItems)
    {
        FileEntryList candidates;
        FileEntry entry;

        // this is a new entry not recorded in the base roster
        if (missingItem->hasStatus(InventoryItem::Added) ||
            missingItem->hasStatus(InventoryItem::RenameTarget))
        {
            entry.path = missingItem->getPath();
            entry.is_dir = missingItem->isDirectory();
        }
        else
        {
            bool found = false;
            foreach (entry, parentList)
            {
                if (entry.path == missingItem->getPath())
                {
                    found = true;
                    break;
                }
            }
            I(found);
        }

        foreach (InventoryItem * unknownItem, unknownItems)
        {
            QString unknownPath = unknownItem->getPath();

            // calculate the file id of the unknown file
            if (!unknownItem->isDirectory() &&
                !fileIds.contains(unknownPath))
            {
                QString fileid = MonotoneUtil::getFileId(
                    monotoneHandle, unknownPath
                );
                // file was not readable, etc.
                if (fileid.isEmpty()) continue;
                fileIds.insert(unknownPath, fileid);
            }

            // at first do a simple file name check for items of the same type
            if (missingItem->isDirectory() == unknownItem->isDirectory())
            {
                // equal names / different paths or different names / equal paths
                if (missingItem->getFilename() == unknownItem->getFilename() ||
                    missingItem->getBaseDirectory() == unknownItem->getBaseDirectory())
                {
                    FileEntry can(unknownPath, unknownItem->isDirectory());
                    if (fileIds.contains(unknownPath))
                    {
                        can.fileid = fileIds.value(unknownPath);
                    }
                    candidates.append(can);
                    continue;
                }
            }

            // we can't do anything for directories from here on
            if (missingItem->isDirectory() || unknownItem->isDirectory())
                continue;

            // we now rely on the fact that we have a fileid
            I(fileIds.contains(unknownPath));

            if (fileIds.value(unknownPath) == entry.fileid)
            {
                candidates.append(FileEntry(unknownPath, false, entry.fileid));
            }
        }

        // only add those missing items to the map which are having any
        // candidates, i.e. are likely to be renamed
        if (candidates.size() > 0)
        {
            unaccountedRenames.insert(entry, candidates);
        }
    }

    return unaccountedRenames;
}

