/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "InventoryModel.h"
#include "MonotoneProcess.h"

InventoryModel::InventoryModel(QObject * parent)
  : QAbstractItemModel(parent), inventory()
{
    inventory = new Inventory(this);

    connect(
        inventory, SIGNAL(invalidWorkspaceFormat()),
        this, SIGNAL(invalidWorkspaceFormat())
    );

    connect(
        inventory, SIGNAL(beginInsertRows(ModelItem *, int, int)),
        this, SLOT(triggerBeginInsertRows(ModelItem *, int, int))
    );

    connect(
        inventory, SIGNAL(endInsertRows()),
        this, SLOT(triggerEndInsertRows())
    );

    connect(
        inventory, SIGNAL(beginRemoveRows(ModelItem *, int, int)),
        this, SLOT(triggerBeginRemoveRows(ModelItem *, int, int))
    );

    connect(
        inventory, SIGNAL(endRemoveRows()),
        this, SLOT(triggerEndRemoveRows())
    );

    connect(
        inventory, SIGNAL(dataChanged(ModelItem *)),
        this, SLOT(triggerDataChanged(ModelItem *))
    );

    connect(
        inventory, SIGNAL(inventoryRead(const QString &)),
        this, SIGNAL(layoutChanged())
    );
}

InventoryModel::~InventoryModel()
{
    I(nodeInsertions.size() == 0);
    delete inventory;
}

void InventoryModel::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
    inventory->setMonotoneHandle(handle);
}

void InventoryModel::refresh(const QString & path)
{
    inventory->read(path, false);
}

bool InventoryModel::canFetchMore(const QModelIndex & parent) const
{
    // special handling for the root item
    if (!parent.isValid())
        return true;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return false;
    }

    if (invitem->isExpanded(Inventory::QueryLevel))
    {
        return false;
    }

    if (invitem->isAboutToBeExpanded())
    {
        return false;
    }

    return true;
}

void InventoryModel::fetchMore(const QModelIndex & parent)
{
    if (!parent.isValid())
        return;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return;
    }

    // again, if this item is already expanded, don't expand it again
    if (invitem->isExpanded(Inventory::QueryLevel) ||
        invitem->isAboutToBeExpanded()) return;

    invitem->setAboutToBeExpanded();

    inventory->read(invitem->getPath());
}

QModelIndex InventoryModel::indexFromItem(ModelItem * item, int col) const
{
    if (item->isRoot())
    {
        return QModelIndex();
    }
    return createIndex(item->row(), col, reinterpret_cast<void *>(item));
}

QModelIndex InventoryModel::index(int row, int column, const QModelIndex & parent) const
{
    if (column < 0)
    {
        return QModelIndex();
    }

    ModelItem * parentItem;

    if (!parent.isValid())
    {
        parentItem = inventory->rootItem;
    }
    else
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
        I(parentItem);
    }

    ModelItem * childItem = parentItem->child(row);

    if (childItem)
    {
        return createIndex(row, column, childItem);
    }

    return QModelIndex();
}

int InventoryModel::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant InventoryModel::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    I(item);
    return item->data(index.column(), role);
}

bool InventoryModel::setData(const QModelIndex & index, const QVariant & data, int role = Qt::EditRole)
{
    if (role != Qt::EditRole || index.column() != 0)
        return false;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    InventoryItem * invitem = qobject_cast<InventoryItem *>(item);
    if (!invitem)
        return false;

    QString newFilename = data.toString();
    if (newFilename.indexOf(QRegExp("[/\\\?:|*\"<>]"))  != -1)
    {
        emit workspaceCommandError(tr(
            "The new file name must not contain the "
            "following characters: / \\ ? : | * \" < >"
        ));
        return false;
    }

    QString oldFilename = invitem->getFilename();
    if (oldFilename == newFilename)
        return true;

    QString baseDirectory = invitem->getBaseDirectory();
    if (!baseDirectory.isEmpty())
    {
        baseDirectory.append("/");
    }

    renamePath(baseDirectory + oldFilename, baseDirectory + newFilename);
    return true;
}

Qt::ItemFlags InventoryModel::flags(const QModelIndex & index) const
{
    if (!index.isValid())
        return 0;

    QFlags<Qt::ItemFlag> flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

    if (index.column() == 0)
    {
        ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
        I(item);

        InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);
        if (invitem && invitem->isNewNode())
        {
            if (invitem->getPath() != "")
            {
                flags |= Qt::ItemIsEditable;
                flags |= Qt::ItemIsDragEnabled;
            }

            if (invitem->isDirectory() && index.column() == 0)
            {
                flags |= Qt::ItemIsDropEnabled;
            }
        }
    }

    return flags;
}

QStringList InventoryModel::mimeTypes() const
{
    return QStringList() << "application/x-inventory-items";
}

QMimeData * InventoryModel::mimeData(const QModelIndexList & indexes) const
{
    QMimeData * mimeData = new QMimeData();
    QByteArray encodedData;

    QDataStream stream(&encodedData, QIODevice::WriteOnly);

    foreach (QModelIndex index, indexes)
    {
        if (!index.isValid())
            continue;

        ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
        I(item);
        InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);
        stream << invitem->getPath();
    }

    mimeData->setData("application/x-inventory-items", encodedData);
    return mimeData;
}

bool InventoryModel::dropMimeData(const QMimeData * data, Qt::DropAction action,
                                  int row, int column, const QModelIndex & parent)
{
    Q_UNUSED(row);

    if (action == Qt::IgnoreAction)
        return true;

    if (!data->hasFormat("application/x-inventory-items"))
        return false;

    if (column > 0)
        return false;

    if (!parent.isValid())
        return false;

    ModelItem * item = static_cast<ModelItem *>(parent.internalPointer());
    I(item);
    InventoryItem * invitem = qobject_cast<InventoryItem *>(item);
    if (!invitem)
        return false;

    QString targetPath = invitem->getPath();
    if (!targetPath.isEmpty())
        targetPath.append("/");

    QByteArray encodedData = data->data("application/x-inventory-items");
    QDataStream stream(&encodedData, QIODevice::ReadOnly);

    while (!stream.atEnd())
    {
        QString oldPath, newPath;
        stream >> oldPath;

        int pos = oldPath.lastIndexOf('/');
        QString filename = pos == -1
            ? oldPath
            : oldPath.right(oldPath.length() - pos - 1);
        newPath = targetPath + filename;

        renamePath(oldPath, newPath);
    }

    return true;
}

// the root item needs to store the header for each field
QVariant InventoryModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        return inventory->rootItem->data(section, role);
    }

    return QVariant();
}

QModelIndex InventoryModel::parent(const QModelIndex & index) const
{
    if (!index.isValid())
    {
        return QModelIndex();
    }

    ModelItem * childItem = static_cast<ModelItem *>(index.internalPointer());
    I(childItem);
    ModelItem * parentItem = childItem->parent();

    if (parentItem == inventory->rootItem)
    {
        return QModelIndex();
    }

    I(parentItem);
    return createIndex(parentItem->row(), 0, parentItem);
}

int InventoryModel::rowCount(const QModelIndex & parent) const
{
    if (parent.column() > 0)
    {
        return 0;
    }

    ModelItem * parentItem = inventory->rootItem;

    if (parent.isValid())
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
    }

    I(parentItem);
    return parentItem->childCount();
}

void InventoryModel::triggerBeginInsertRows(ModelItem * parent, int start, int end)
{
    beginInsertRows(indexFromItem(parent, 0), start, end);
    nodeInsertions.push(new NodeInsertion(parent, start, end));
}

void InventoryModel::triggerEndInsertRows()
{
    NodeInsertion * insertion = nodeInsertions.pop();

    QStringList paths;
    for (int i=insertion->start; i<=insertion->end; i++)
    {
        ModelItem * child = insertion->parent->child(i);
        InventoryItem * invItem = qobject_cast<InventoryItem *>(child);
        if (!invItem) continue;
        // do not monitor non-existing files
        if (invItem->getFSType() == InventoryItem::None) continue;
        // a little optimization: we don't care about / monitor ignored files
        if (invItem->hasStatus(InventoryItem::Ignored)) continue;

        QString path = invItem->getPath();
        if (invItem->isDirectory()) path += "/";
        paths.push_back(path);
    }

    delete insertion;
    emit pathsInserted(paths);

    endInsertRows();
}

void InventoryModel::triggerBeginRemoveRows(ModelItem * parent, int start, int end)
{
    beginRemoveRows(indexFromItem(parent, 0), start, end);

    QStringList paths;
    for (int i=start; i<=end; i++)
    {
        ModelItem * child = parent->child(i);
        InventoryItem * invItem = qobject_cast<InventoryItem *>(child);
        if (!invItem) continue;

        QString path = invItem->getPath();
        if (invItem->isDirectory()) path += "/";
        paths.push_back(path);
    }

    emit pathsRemoved(paths);
}

void InventoryModel::triggerEndRemoveRows()
{
    endRemoveRows();
}

void InventoryModel::triggerDataChanged(ModelItem * item)
{
    emit dataChanged(indexFromItem(item, 0), indexFromItem(item, 2));
}

void InventoryModel::runWorkspaceCommand(const QStringList & args)
{
    MonotoneProcess proc;
    proc.setMonotoneHandle(monotoneHandle);
    proc.start(args);
    proc.waitForFinished();

    if (!proc.successful())
    {
        emit workspaceCommandError(proc.getBufferedOutput());
    }
}

void InventoryModel::addPaths(const QStringList & paths)
{
    QStringList args;
    args << "add" << "--recursive";
    args += paths;

    runWorkspaceCommand(args);
}

void InventoryModel::dropPaths(const QStringList & paths)
{
    QStringList args;
    args << "drop" << "--recursive";
    args += paths;

    runWorkspaceCommand(args);
}

void InventoryModel::revertPaths(const QStringList & paths)
{
    QStringList args;
    args << "revert";
    args += paths;

    runWorkspaceCommand(args);
}

void InventoryModel::renamePath(const QString & oldPath, const QString & newPath)
{
    // some sanity (see http://savannah.nongnu.org/bugs/?29484)
    if (oldPath == newPath ||
        newPath.indexOf(oldPath + "/") == 0)
    {
        W(QString("ignoring invalid rename from %1 to %2")
          .arg(oldPath).arg(newPath));
        return;
    }

    QStringList args;
    args << "rename" << oldPath << newPath;

    QFileInfo fileInfo(monotoneHandle->getData() + "/" + oldPath);
    if (!fileInfo.exists())
    {
        args << "--bookkeep-only";
    }

    runWorkspaceCommand(args);
}

void InventoryModel::newFile(const QString & path)
{
    QFile file(monotoneHandle->getData() + "/" + path);
    if (!file.open(QIODevice::ReadWrite))
    {
        emit workspaceCommandError(tr("Could not create file %1").arg(path));
        return;
    }
    file.close();

    addPaths(QStringList() << path);
}

void InventoryModel::newDirectory(const QString & path)
{
    QDir workspace(monotoneHandle->getData());
    if (!workspace.mkpath(path))
    {
        emit workspaceCommandError(tr("Could not create directory %1").arg(path));
        return;
    }
    addPaths(QStringList() << path);
}

void InventoryModel::ignorePaths(const QStringList & paths)
{
    QString path(monotoneHandle->getData());
    path.append(QDir::separator()).append(".mtn-ignore");

    QFile file(path);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Append))
    {
        emit workspaceCommandError(
            tr("could not open .mtn-ignore for writing")
        );
        return;
    }

    foreach (QString path, paths)
    {
        file.write(QRegExp::escape(path).toUtf8() + "\n");
    }
    file.close();
}

void InventoryModel::unignorePaths(const QStringList & paths)
{
    QString path(monotoneHandle->getData());
    path.append(QDir::separator()).append(".mtn-ignore");

    QFile file(path);
    if (!file.open(QIODevice::ReadWrite))
    {
        emit workspaceCommandError(
           tr("could not open .mtn-ignore for reading and writing")
        );
        return;
    }

    QString data = QString::fromUtf8(file.readAll().data());
    foreach (QString path, paths)
    {
        data.replace(QRegExp(QRegExp::escape(path).toUtf8() + "\n"), "");
    }

    file.resize(0);
    file.write(data.toUtf8());
    file.close();
}

