/***************************************************************************
 *   Copyright (C) 2006 by Ingo Maindorfer								   *
 *   ingo@liquidcooling.de												   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "InventoryProxyModel.h"
#include "InventoryItem.h"
#include "vocab.h"

InventoryProxyModel::InventoryProxyModel(QObject * parent, bool isFolderTree)
  : QSortFilterProxyModel(parent), folderTree(isFolderTree),
    sortColumn(0), hideIgnored(false), viewOption(All),
    sortOrder(Qt::AscendingOrder)
{
    // FIXME: maybe this should be configurable?
    filter.setCaseSensitivity(Qt::CaseSensitive);
    filter.setPatternSyntax(QRegExp::Wildcard);
}

InventoryProxyModel::~InventoryProxyModel() {}

bool InventoryProxyModel::filterAcceptsRow(int sourceRow, const QModelIndex & sourceParent) const
{
    QModelIndex sourceIndex = sourceModel()->index(sourceRow, 0, sourceParent);

    ModelItem * item = static_cast<ModelItem *>(sourceIndex.internalPointer());

    // always return the root item
    if (item->isRoot())
        return true;

    PseudoItem * psitem = qobject_cast<PseudoItem *>(item);
    if (psitem)
    {
        // only display this item if this is not the folder tree view
        // and the parent of the item's parent is invalid (i.e. there is no
        // way to cdUp anymore)
        if (psitem->isCdUp())
            return !folderTree && sourceParent.parent().isValid();

        I(false);
    }

    InventoryItem * invitem = qobject_cast<InventoryItem *>(item);
    I(invitem);

    // hide ignored files (not recursively)
    if (invitem->hasStatus(InventoryItem::Ignored) && hideIgnored)
    {
        return false;
    }

    bool acceptRow = true;

    // check the view options
    int state        = invitem->getStatusRecursive();
    bool cntchanged  = state & InventoryItem::ContentsChanged;
    bool attrchanged = state & InventoryItem::ContentsChanged;
    bool added       = state & InventoryItem::Added;
    bool dropped     = state & InventoryItem::Dropped;
    bool renamed     = state & InventoryItem::RenameSource ||
                       state & InventoryItem::RenameTarget;
    bool unknown     = state & InventoryItem::Unknown;
    bool missing     = state & InventoryItem::Missing;
    bool ignored     = state & InventoryItem::Ignored;
    bool invalid     = state & InventoryItem::Invalid;
    bool changed     = cntchanged || attrchanged || added || dropped || renamed;

    acceptRow &= viewOption != Changed           || changed;
    acceptRow &= viewOption != ContentsChanged   || cntchanged;
    acceptRow &= viewOption != AttributesChanged || attrchanged;
    acceptRow &= viewOption != Added             || added;
    acceptRow &= viewOption != Dropped           || dropped;
    acceptRow &= viewOption != Renamed           || renamed;
    acceptRow &= viewOption != Unknown           || unknown;
    acceptRow &= viewOption != Missing           || missing;
    acceptRow &= viewOption != Ignored           || ignored;
    acceptRow &= viewOption != Invalid           || invalid;

    // check if we should only display folders
    acceptRow &= !folderTree || invitem->isDirectory();

    acceptRow &= filter.isEmpty() || itemMatchesFilter(invitem);

    return acceptRow;
}

bool InventoryProxyModel::lessThan(const QModelIndex & left, const QModelIndex & right) const
{
    ModelItem * itemLeft = static_cast<ModelItem *>(left.internalPointer());
    ModelItem * itemRight = static_cast<ModelItem *>(right.internalPointer());

    InventoryItem * invitemLeft = qobject_cast<InventoryItem*>(itemLeft);
    InventoryItem * invitemRight = qobject_cast<InventoryItem*>(itemRight);

    // place pseudo items always right on top, in no particular order
    if (!invitemLeft || !invitemRight)
        return false;

    // we shall sort by the name of the item
    if (sortColumn == 0)
    {
        if (invitemLeft->isDirectory() && invitemRight->isDirectory())
        {
            return invitemLeft->getFilename() < invitemRight->getFilename();
        }

        if (!invitemLeft->isDirectory() && !invitemRight->isDirectory())
        {
            return invitemLeft->getFilename() < invitemRight->getFilename();
        }

        if (invitemLeft->isDirectory() && !invitemRight->isDirectory())
        {
            return true;
        }

        // last case: !invitemLeft->isDirectory() && invitemRight->isDirectory()
        return false;
    }

    // we shall sort only by status
    if (sortColumn == 1)
    {
        return invitemLeft->getStatusString() < invitemRight->getStatusString();
    }

    // anything else
    return false;
}

void InventoryProxyModel::sort(int column, Qt::SortOrder order)
{
    sortOrder = order;
    sortColumn = column;
    QSortFilterProxyModel::sort(column, order);
}

void InventoryProxyModel::setHideIgnoredFiles(bool hide)
{
    if (hide == hideIgnored)
        return;

    hideIgnored = hide;
    invalidate();
}

void InventoryProxyModel::setViewOption(int opt)
{
    ViewOption newOpt = static_cast<ViewOption>(opt);
    if (viewOption == newOpt)
        return;

    viewOption = newOpt;
    invalidate();
}

void InventoryProxyModel::filterItemsByName(const QString & pattern)
{
    filter.setPattern(pattern);
    invalidate();
}

// FIXME: we're basically doing this twice for each proxy model, while
// a single proxy model which does the filtering for both (sub) proxy
// models should be enough. Tweak that some time.
bool InventoryProxyModel::itemMatchesFilter(const InventoryItem * item) const
{
    if (filter.indexIn(item->getLabel()) >= 0)
    {
        return true;
    }

    // we always have to filter _all_ items of the view even if we only
    // display a subset of them in a list-alike way, since the parents
    // of a specific matched node must match as well in order to keep
    // the node in
    foreach (const ModelItem * child, item->getChildren())
    {
        const InventoryItem * invitem =
        qobject_cast<const InventoryItem *>(child);

        if (invitem && itemMatchesFilter(invitem))
            return true;
    }

    return false;
}

