/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Manifest.h"
#include "BasicIOParser.h"
#include "IconProvider.h"

Manifest::Manifest(QObject * parent, const MonotoneHandlePtr & handle)
  : QAbstractItemModel(parent), AutomateCommand(0), monotoneHandle(handle), root(0)
{}

Manifest::~Manifest()
{
    if (root) delete root;
}

void Manifest::readManifest(const QString & rev)
{
    // reset the view
    reset();

    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_manifest_of" << rev));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void Manifest::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    QMap<QString, ManifestEntry *> directoryMap;

    if (root)
    {
        delete root;
        root = 0;
    }

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);
        ManifestEntry * mEntry = NULL;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            // skip the format version
            if (j == 0)
            {
                if (entry.sym == "format_version") break;
                mEntry = new ManifestEntry();
            }

            I(mEntry);

            if (entry.sym == "dir")
            {
                I(entry.vals.size() == 1);
                mEntry->path = entry.vals.at(0);
                mEntry->is_directory = true;

                // check if this is the root entry
                if (mEntry->path == "")
                {
                    I(!root);
                    root = mEntry;
                }

                directoryMap.insert(mEntry->path, mEntry);
            }

            if (entry.sym == "file")
            {
                I(entry.vals.size() == 1);
                mEntry->path = entry.vals.at(0);
                mEntry->is_directory = false;
            }

            if (entry.sym == "dir" || entry.sym == "file")
            {
                if (mEntry->path != "")
                {
                    int pos = mEntry->path.lastIndexOf("/");
                    QString basePath = "";
                    if (pos > -1)
                    {
                        basePath = mEntry->path.left(pos);
                    }
                    I(directoryMap.contains(basePath));
                    ManifestEntry * parent = directoryMap.value(basePath);
                    mEntry->parent = parent;
                    parent->children.append(mEntry);
                }
                continue;
            }

            if (entry.sym == "content")
            {
                I(entry.vals.size() == 1);
                mEntry->hash = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "attr")
            {
                I(entry.vals.size() == 2);
                mEntry->attributes.insert(entry.vals.at(0), entry.vals.at(1));
                continue;
            }

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }
    }

    I(root);

    // reset any connected view(s)
    reset();
    emit manifestRead();
}

int Manifest::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant Manifest::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    ManifestEntry * entry = static_cast<ManifestEntry *>(index.internalPointer());

    if (role == Qt::DisplayRole)
    {
        switch (index.column())
        {
            case 0: return QVariant(entry->filename());
            case 1: return QVariant(entry->hash);
        }
    }

    if (role == Qt::DecorationRole && index.column() == 0)
    {
        IconProvider * provider = IconProvider::singleton();
        if (entry->is_directory) return provider->getPlainFolderIcon();
        return provider->getPlainFileIcon();
    }

    return QVariant();
}

Qt::ItemFlags Manifest::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant Manifest::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("File"));
            case 1: return QVariant(tr("Hash"));
        }
    }
    return QVariant();
}

int Manifest::rowCount(const QModelIndex & parent) const
{
    ManifestEntry * parentEntry = root;

    if (parent.isValid())
    {
        parentEntry = static_cast<ManifestEntry *>(parent.internalPointer());
    }

    if (!parentEntry) return 0;

    return parentEntry->children.size();
}

QModelIndex Manifest::index(int row, int column, const QModelIndex & parent) const
{
    ManifestEntry * parentEntry;

    if (!parent.isValid())
    {
        parentEntry = root;
    }
    else
    {
        parentEntry = static_cast<ManifestEntry *>(parent.internalPointer());
    }

    if (!parentEntry) return QModelIndex();

    ManifestEntry * entry = parentEntry->child(row);

    if (entry)
    {
        return createIndex(row, column, entry);
    }

    return QModelIndex();
}

QModelIndex Manifest::parent(const QModelIndex & index) const
{
    if (!index.isValid())
    {
        return QModelIndex();
    }

    ManifestEntry * child = static_cast<ManifestEntry *>(index.internalPointer());
    ManifestEntry * parent = child->parent;

    if (parent == root)
    {
        return QModelIndex();
    }

    return createIndex(parent->row(), 0, parent);
}

