/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Tags.h"
#include "BasicIOParser.h"

Tags::Tags(QObject * parent, const MonotoneHandlePtr & handle)
  : QAbstractItemModel(parent), AutomateCommand(0), monotoneHandle(handle)
{
    tags = new TagList();
}

Tags::~Tags()
{
    tags->clear();
    delete tags;
}

void Tags::readTags(const QString & branch)
{
    QStringList cmd;
    cmd << "tags";
    if (branch.length() > 0)
    {
        cmd << branch;
    }

    MonotoneTaskPtr task(new MonotoneTask(cmd));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void Tags::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    tags->clear();

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);

        Tag tag;

        bool isItem = false;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            // we're now only interested in stanzas
            // starting with a "tag" entry
            if (j == 0 && entry.sym != "tag") break;

            isItem = true;

            if (entry.sym == "tag")
            {
                I(entry.vals.size() == 1);
                tag.name = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "revision")
            {
                I(entry.vals.size() == 1);
                tag.revision = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "signer")
            {
                I(entry.vals.size() == 1);
                tag.signer = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "branches")
            {
                tag.branches = entry.vals;
                continue;
            }

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        // check if we really processed an item entry
        if (!isItem) continue;
        tags->append(tag);
    }

    reset();
    emit tagsRead();
}

int Tags::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 4;
}

QVariant Tags::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    if (role != Qt::DisplayRole)
    {
        return QVariant();
    }

    int row = index.row();
    if (row >= tags->size()) return QVariant();

    Tag tag = tags->at(row);

    switch (index.column())
    {
        case 0: return QVariant(tag.name);
        case 1: return QVariant(tag.revision);
        case 2: return QVariant(tag.signer);
        case 3: return QVariant(tag.branches.join("\n"));
    }

    return QVariant();
}

Qt::ItemFlags Tags::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant Tags::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Name"));
            case 1: return QVariant(tr("Revision"));
            case 2: return QVariant(tr("Signer"));
            case 3: return QVariant(tr("Branches"));
        }
    }
    return QVariant();
}

int Tags::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return tags->size();
}

QModelIndex Tags::index(int row, int column, const QModelIndex & parent) const
{
    return hasIndex(row, column, parent) ? createIndex(row, column, 0) : QModelIndex();
}

QModelIndex Tags::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

