/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                *
 *   ganwell@fangorn.ch                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Toposort.h"

#include <QStringList>
#include <QAbstractItemModel>

Toposort::Toposort(QObject * parent, const MonotoneHandlePtr & handle)
  : QSortFilterProxyModel(parent), AutomateCommand(0), monotoneHandle(handle),
    sortColumn(0), sortOrder(Qt::AscendingOrder)
{
    ordRevisions = new RevisionMap();
    showRows     = false;
}

Toposort::~Toposort()
{
    if (ordRevisions)
    {
        ordRevisions->clear();
        delete ordRevisions;
    }
}

void Toposort::setSourceModel(QAbstractItemModel * source)
{
    if (sourceModel() != 0)
    {
        disconnect(
            sourceModel(), SIGNAL(modelReset()),
            this, SLOT(sortModel())
        );
    }

    QSortFilterProxyModel::setSourceModel(source);

    connect(
        source, SIGNAL(modelReset()),
        this, SLOT(sortModel())
    );
}

bool Toposort::lessThan(const QModelIndex & left, const QModelIndex & right) const
{
    QString leftRev = sourceModel()->data(left).toString();
    QString rightRev = sourceModel()->data(right).toString();

    if (!ordRevisions->contains(leftRev) || !ordRevisions->contains(leftRev))
    {
        D("One or both revisions not found.");
        return false;
    }

    return ordRevisions->value(leftRev) < ordRevisions->value(rightRev);
}

bool Toposort::filterAcceptsRow(int sourceRow, const QModelIndex & sourceParent) const
{
    Q_UNUSED(sourceRow);
    Q_UNUSED(sourceParent);
    return showRows;
}

void Toposort::sort(int column, Qt::SortOrder order)
{
    sortOrder = order;
    sortColumn = column;
    QSortFilterProxyModel::sort(column, order);
}

void Toposort::sortModel()
{
    showRows = false;
    ordRevisions->clear();

    int row = 0;
    QModelIndex current = sourceModel()->index(row, 0);
    if (!current.isValid())
    {
        D("No items found in source model");
        return;
    }

    QStringList revs;
    do
    {
        QString rev = current.data().toString();
        // skip everything that does not look like a revision id
        if (rev.size() == 40)
        {
            revs.append(rev);
        }
        current = current.sibling(++row, 0);
        if (!current.isValid()) break;
    }
    while (true);

    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "toposort" << revs));
    AutomateCommand::enqueueTask(monotoneHandle, task);
}

void Toposort::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    QStringList sortedRevs(task->getDecodedOutput().split("\n"));
    for (int i=0, j=sortedRevs.size(); i<j; i++)
    {
        ordRevisions->insert(sortedRevs.at(i), i);
    }

    showRows = true;
    filterChanged();
    clear();

    D("Sorting finished");

    // signal that we've finished (whoever listens to that)
    emit sortingFinished();
}

