/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneHandle.h"
#include "MonotoneUtil.h"
#include "BasicIOParser.h"

#include <QDir>
#include <QTcpSocket>

QStringList MonotoneHandle::defaultDatabaseLocations;
QMap<QString, QString> MonotoneHandle::privateKeys;

MonotoneHandle::MonotoneHandle(Type t, const QString & d)
    : type(t), data(d)
{}

QString MonotoneHandle::identify() const
{
    return QString("%1:%2").arg((int)type).arg(data);
}

void MonotoneHandle::validateDatabaseFile(const QString & database)
{
    QFile dbfile(database);
    if (!dbfile.open(QIODevice::ReadOnly))
    {
        throw GuitoneException(
            QObject::tr("Could not open database for reading: %1")
                .arg(dbfile.errorString())
        );
    }

    if (dbfile.read(15) != QByteArray("SQLite format 3"))
    {
        dbfile.close();
        throw GuitoneException(
            QObject::tr("Database file not an SQLite 3 database")
        );
    }

    if (!dbfile.seek(60) || dbfile.read(4) != QByteArray("_MTN"))
    {
        dbfile.close();
        throw GuitoneException(QObject::tr(
            "SQLite database is not a monotone database or a database "
            "of an older, unsupported version of monotone"
        ));
    }
    dbfile.close();
}

void MonotoneHandle::validateWorkspacePath(QString & workspace,
                                           QString & database)
{
    QDir tempDir(workspace);
    I(tempDir.exists());

    bool found = false;
    do
    {
        if (tempDir.cd("_MTN"))
        {
            tempDir.cdUp();
            workspace = tempDir.absolutePath();
            found = true;
            break;
        }
    }
    while (!tempDir.isRoot() && tempDir.cdUp());

    if (!found)
    {
        throw GuitoneException(
            QObject::tr("could not find _MTN directory")
        );
    }

    QFile optionsFile(workspace + "/_MTN/options");
    if (!optionsFile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        throw GuitoneException(
            QObject::tr("could not open _MTN/options for reading")
        );
    }

    QByteArray contents = optionsFile.readAll();
    optionsFile.close();

    if (contents.size() == 0)
    {
        throw GuitoneException(QObject::tr("file _MTN/options is empty"));
    }

    BasicIOParser parser(QString::fromUtf8(contents));
    if (!parser.parse())
    {
        throw GuitoneException(
            QObject::tr("could not parse basic_io from _MTN/options")
        );
    }

    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1);
    Stanza st = stanzas.at(0);

    foreach (const StanzaEntry & entry, st)
    {
        if (entry.sym == "database")
        {
            I(entry.vals.size() == 1);
            database = entry.vals.at(0);
            break;
        }
    }

    if (database.isEmpty())
    {
        throw GuitoneException(
            QObject::tr("could not find database for workspace")
        );
    }
}

void MonotoneHandle::validatePrivateKey(const QString & keyFragment, QString & keyId)
{
    if (privateKeys.size() == 0)
    {
        privateKeys = MonotoneUtil::getPrivateKeyMap(MonotoneHandle::create());
    }

    QMap<QString, QString> possibleKeys;

    QMapIterator<QString, QString> it(privateKeys);
    while (it.hasNext())
    {
        it.next();

        if (it.key().indexOf(keyFragment) != -1 ||
            it.value().indexOf(keyFragment) != -1)
        {
            possibleKeys.insert(it.key(), it.value());
        }
    }

    if (possibleKeys.size() == 0)
    {
        throw GuitoneException(
            QObject::tr("no private key for key fragment '%1' found")
                .arg(keyFragment)
        );
    }

    if (possibleKeys.size() > 1)
    {
        throw GuitoneException(
           QObject::tr("multiple private keys for key fragment '%1' found:\n\t%2")
             .arg(keyFragment)
             .arg(QStringList(possibleKeys.values()).join("\n\t"))
        );
    }

    keyId = possibleKeys.constBegin().key();
}

void MonotoneHandle::validateServerConnection(const QString & host)
{
    QStringList hostAndPort = host.split(":", QString::SkipEmptyParts);
    if (hostAndPort.size() == 1)
    {
        hostAndPort.append("4691");
    }

    QTcpSocket sock;
    sock.connectToHost(hostAndPort.at(0), hostAndPort.at(1).toUInt());
    sock.waitForConnected(1000);

    if (sock.state() == QAbstractSocket::UnconnectedState ||
        !sock.isValid())
    {
        sock.disconnectFromHost();
        if (sock.state() != QAbstractSocket::UnconnectedState)
            sock.waitForDisconnected();

        throw GuitoneException(QObject::tr(
            "Could not connect to '%1':\n%2"
        ).arg(host).arg(sock.errorString()));
    }

    sock.disconnectFromHost();
    if (sock.state() != QAbstractSocket::UnconnectedState)
        sock.waitForDisconnected();
}

void MonotoneHandle::resolveDatabaseAlias(QString & database)
{
    QRegExp rx("^:([^\\/:]+)$");
    if (rx.indexIn(database) != -1)
    {
        QString alias = rx.cap(1);

        if (defaultDatabaseLocations.size() == 0)
        {
            defaultDatabaseLocations =
            MonotoneUtil::getDefaultDatabaseLocations(MonotoneHandle::create());
        }

        foreach (const QString & defaultLocation, defaultDatabaseLocations)
        {
            QFileInfo finfo(defaultLocation + "/" + alias);
            if (finfo.isFile())
            {
                database = finfo.canonicalFilePath();
                break;
            }
            QFileInfo finfoext(defaultLocation + "/" + alias + ".mtn");
            if (finfoext.isFile())
            {
                database = finfoext.canonicalFilePath();
                break;
            }
        }
    }
}

MonotoneHandlePtr MonotoneHandle::create(const QString & pathOrURI)
{
    Type type;
    QString cleanedPathOrURI(pathOrURI.trimmed());

    if (cleanedPathOrURI.isEmpty())
    {
        type = empty_handle;
    }
    else
    {
        QRegExp rx("^mtn://(?:(\\w*)@)?([\\w\\-]+(?:\\.[\\w\\-]+)*(?::\\d+)?)$");
        if (rx.indexIn(cleanedPathOrURI) != -1)
        {
            type = server_handle;
            validateServerConnection(rx.cap(2));

            QString keyFragment = rx.cap(1);
            if (!keyFragment.isEmpty())
            {
                QString keyId;
                validatePrivateKey(rx.cap(1), keyId);
                cleanedPathOrURI = "mtn://" + keyId + "@" + rx.cap(2);
            }
            else
            {
                cleanedPathOrURI = "mtn://" + rx.cap(2);
            }
        }
        else
        {
            QFileInfo finfo(cleanedPathOrURI);
            if (finfo.isDir())
            {
                type = workspace_handle;

                cleanedPathOrURI = finfo.canonicalFilePath();

                QString database;
                validateWorkspacePath(cleanedPathOrURI, database);
                resolveDatabaseAlias(database);
                validateDatabaseFile(database);
            }
            else
            {
                type = database_handle;

                resolveDatabaseAlias(cleanedPathOrURI);
                validateDatabaseFile(cleanedPathOrURI);
            }
        }

        I(!cleanedPathOrURI.isEmpty());
    }

    return MonotoneHandlePtr(new MonotoneHandle(type, cleanedPathOrURI));
}

