/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#ifndef MONOTONE_TASK_H
#define MONOTONE_TASK_H

#include "vocab.h"

class MonotoneTask
{
public:
    typedef enum {Warning = 'w', Info = 'p', Error = 'e' } MessageType;

    MonotoneTask();
    MonotoneTask(const MonotoneTask &);
    MonotoneTask(const QStringList &);
    MonotoneTask(const QStringList &, const QStringList &);
    MonotoneTask(const ByteArrayList &);
    MonotoneTask(const ByteArrayList &, const ByteArrayList &);

    //! sets the command number of this task (increased internally by the thread)
    void setCommandNumber(int num) { commandNumber = num; }

    //! sets the thread number to identify the thread
    void setThreadNumber(int num) { threadNumber = num; }

    //! sets the raw command output
    void setOutput(const QByteArray & out) { output = out; }

    //! sets the return code of the command
    void setReturnCode(int code) { returnCode = code; }

    //! marks that the task has been finished
    void setFinished() { finished = true; }

    //! set the output encoding
    void setOutputEncoding(const QString & e) { outputEncoding = e; }

    //! appends more out-of-band messages for this task
    void appendOutOfBandMessages(const QList<MessagePair> &);

    //! returns true if one or more out-of-band messages match the given regex
    bool outOfBandMessagesMatch(MessageType, const QRegExp &);

    //! returns the last out-of-band message of the given type
    QString getLast(MessageType);

    //! returns the command input stdio-encoded
    QByteArray getEncodedInput() const;

    //! returns the given arguments for the command
    ByteArrayList getArguments() const { return arguments; }

    //! returns the given options for the command
    ByteArrayList getOptions() const { return options; }

    //! returns the raw monotone output
    QByteArray getOutput() const { return output; }

    //! converts the raw byte output into a unicode encoded QString
    QString getDecodedOutput(const QString & enc = QString()) const;

    //! returns the command's return code
    int getReturnCode() const { return returnCode; }

    //! returns the command's number (i.e. sequential number of the running mtn process)
    int getCommandNumber() const { return commandNumber; }

    //! returns the number of the thread on which this task should/has been run
    int getThreadNumber() const { return threadNumber; }

    //! true, if the task has already been completed
    bool isFinished() const { return finished; }

    //! aborts the processing of the task
    void setAborted() { abortTask = true; }

    //! checks if the task should be aborted
    bool isAborted() const { return abortTask; }

private:
    void init(const ByteArrayList &, const ByteArrayList &);
    ByteArrayList stringToByteArrayList(const QStringList &);

    int returnCode;
    int commandNumber;
    int threadNumber;
    bool finished;
    bool abortTask;

    ByteArrayList arguments;
    ByteArrayList options;
    QByteArray output;
    QString outputEncoding;
    QList<MessagePair> oobMessages;
};

#endif

