/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keler                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneThread.h"
#include "StdioParser.h"
#include "BasicIOParser.h"
#include "MonotoneResourceFile.h"
#include "Settings.h"

#include <QMetaType>
#include <QEventLoop>

MonotoneThread::MonotoneThread(int thread, const QString & mtn,
                               const MonotoneHandlePtr & handle)
 : QThread(), doAbort(false), commandNumber(0),
   threadNumber(thread), mtnBinary(mtn), monotoneHandle(handle)
{}

MonotoneThread::~MonotoneThread() {}

//! returns true if the process started successfully,
//! false in case of an abortion (whatever comes first)
bool MonotoneThread::waitForStarted()
{
    QEventLoop * ev = new QEventLoop(this);

    connect(this, SIGNAL(started(int)),
            ev, SLOT(quit()));
    connect(this, SIGNAL(terminated()),
            ev, SLOT(quit()));

    ev->exec();
    delete ev;

    return isRunning();
}

int MonotoneThread::enqueueTask(const MonotoneTaskPtr & task)
{
    QMutexLocker locker(&lock);

    if (doAbort) return -1;

    commandNumber++;
    task->setCommandNumber(commandNumber);
    task->setThreadNumber(threadNumber);
    queue.enqueue(task);

    waitForTasks.wakeAll();
    D(QString("thread %1: new task enqueued: %2")
      .arg(threadNumber).arg(QString(task->getEncodedInput())));
    return commandNumber;
}

void MonotoneThread::abort()
{
    QMutexLocker locker(&lock);
    doAbort = true;
    waitForTasks.wakeAll();
}

void MonotoneThread::cleanup(QProcess * proc)
{
    QMutexLocker locker(&lock);

    if (queue.size() > 0)
    {
        foreach (MonotoneTaskPtr task, queue)
        {
            emit taskAborted(task);
        }
        queue.clear();
    }

    // close the pipes
    proc->close();
    // send SIGTERM
    proc->terminate();
    // block until the process has really been finished
    proc->waitForFinished();
}

void MonotoneThread::run()
{
    // use a local object instance to automatically trigger the
    // destruction of the process if we leave run()
    QObject threadParent;
    QProcess * process = new QProcess(&threadParent);

    QStringList args;
    if (!Settings::getBool("ReadUserRcFiles"))
    {
        args << "--no-standard-rcfiles";
    }
    MonotoneResourceFile rcFile;
    args << "--rcfile" << rcFile.fileName();

    if (monotoneHandle->getType() == MonotoneHandle::database_handle)
    {
        args << "--db" << monotoneHandle->getData();
    }
    else
    // FIXME: this is a workaround since automate stdio always
    // asks for a open database before it proceeds
    if (monotoneHandle->getType() == MonotoneHandle::server_handle ||
        monotoneHandle->getType() == MonotoneHandle::empty_handle)
    {
        args << "--db" << ":memory:";
    }

    args << "automate";

    if (monotoneHandle->getType() == MonotoneHandle::server_handle)
    {
        // remote_stdio outputs a lot of junk we don't want to listen to
        args << "-q" << "-q";

        // remove the mtn:// prefix - monotone up until 0.48 doesn't
        // understand the mtn:// syntax everywhere
        QString host = monotoneHandle->getData().mid(6);

        // FIXME: monotone unfortunately also does not understand our
        // custom '@' syntax - so we need to give the key / auth explicitely
        // with --key
        int pos = host.indexOf("@");
        QString key("");
        if (pos != -1)
        {
            key = host.left(pos);
            host = host.mid(pos + 1);
        }
        args << "--key=" + key
             << "remote_stdio" << host;
    }
    else
    {
        args << "stdio";
    }

    if (monotoneHandle->getType() == MonotoneHandle::workspace_handle)
    {
        process->setWorkingDirectory(monotoneHandle->getData());
    }
    else
    {
        // we need to ensure that the current directory is no workspace
        // directory, so paths won't get interpreted wrongly
        QString tmpPath = QDir::tempPath();
        if (QDir(tmpPath + "/_MTN").exists())
        {
            cleanup(process);
            // this is actually not a QProcess error, but we're misusing
            // this infrastructure for this special error case
            emit aborted(threadNumber, QProcess::UnknownError,
                tr("The temporary directory '%1' contains a monotone "
                   "workspace, which may lead to serious path resolution "
                   "errors on execution. Its recommended you remove "
                   "this workspace before you retry this again!").arg(tmpPath));
            return;
        }
        process->setWorkingDirectory(tmpPath);
    }

    // remove any LANG setting so we get mtn's error output in native English
    // (this is hacky, of course, but there is still no error code
    // implementation in mtn, so we have to do raw string parsing...)
    QStringList env = QProcess::systemEnvironment();
    env.replaceInStrings(QRegExp("^LANG=.*", Qt::CaseInsensitive), "");
    env.push_back("LANG=en_US.UTF-8");
    process->setEnvironment(env);

    L(QString("starting %1 %2").arg(mtnBinary).arg(args.join(" ")));
    process->start(mtnBinary, args);

    if (!process->waitForStarted())
    {
        QByteArray errout = process->readAllStandardError();
        QProcess::ProcessError error = process->error();

        cleanup(process);
        emit aborted(threadNumber, error, QString::fromUtf8(errout));
        return;
    }

    // waitForReadyRead returns false, if there are no data available
    // or the time is up
    // FIXME: find the proper timeout for our use case here or better an
    // alternative way to detect if something went wrong while setting up
    // the process _before_ we send the whole thread to sleep later on
    // waiting silently for his first task...
    process->setReadChannel(QProcess::StandardError);
    if (process->waitForReadyRead(2000))
    {
        QByteArray errout = process->readAllStandardError();
        QProcess::ProcessError error = process->error();

        cleanup(process);
        emit aborted(threadNumber, error, QString::fromUtf8(errout));
        return;
    }
    process->setReadChannel(QProcess::StandardOutput);

    D(QString("thread %1: process started").arg(threadNumber));
    emit started(threadNumber);

    QByteArray buffer;
    QByteArray output;

    bool processingTask = false;
    bool headersProcessed = false;

    while (!doAbort)
    {
        if (!headersProcessed)
        {
            typedef QPair<QByteArray, QByteArray> HeaderPair;
            QList<HeaderPair> headers;

            if (process->bytesAvailable() == 0 &&
                !process->waitForReadyRead(500))
                continue;

            try
            {
                QByteArray output = process->readAllStandardOutput();
                headers = StdioParser::parseAndCheckHeaders(output);
            }
            catch (GuitoneException & error)
            {
                cleanup(process);
                emit aborted(threadNumber, QProcess::UnknownError, error.what());
                return;
            }

            foreach (HeaderPair pair, headers)
            {
                L(QString("stdio headers: %1: %2")
                    .arg(QString(pair.first)).arg(QString(pair.second)));
            }
            headersProcessed = true;
        }

        // send the thread to sleep if there are no tasks to process
        lock.lock();
        if (queue.size() == 0)
            waitForTasks.wait(&lock);
        lock.unlock();

        // check if we got notified to abort the processing
        // while we were sleeping...
        if (doAbort)
            break;

        if (process->state() != QProcess::Running)
        {
            QByteArray errout = process->readAllStandardError();
            QProcess::ProcessError error = process->error();

            cleanup(process);
            emit aborted(threadNumber, error, QString::fromUtf8(errout));
            return;
        }

        MonotoneTaskPtr task = queue.head();

        // check if we should cancel the processing of this task
        if (task->isAborted())
        {
            processingTask = false;
            buffer.clear();
            output.clear();
            queue.dequeue();

            emit taskAborted(task);
            continue;
        }

        if (!processingTask)
        {
            QByteArray in(task->getEncodedInput());

            if (process->write(in) == -1)
            {
                QByteArray err;
                err.append("could not write to process (stderr: ");
                err.append(process->readAllStandardError());
                err.append(")");

                task = queue.dequeue();
                task->setOutput(err);
                task->setReturnCode(-1);

                emit taskAborted(task);
                continue;
            }

            buffer.clear();
            output.clear();
            tickers.clear();
            processingTask = true;

            D(QString("thread %1: task %2 (%3) started")
              .arg(threadNumber)
              .arg(task->getCommandNumber())
              .arg(QString(task->getEncodedInput())));
        }

        while (!process->waitForReadyRead(500) && !doAbort)
        {
            if (process->state() != QProcess::Running)
            {
                QByteArray errout = process->readAllStandardError();
                QProcess::ProcessError error = process->error();

                cleanup(process);
                emit aborted(threadNumber, error, QString::fromUtf8(errout));
                return;
            }
        }

        // even though newer mtn versions should not output anything on stderr
        // while running, we check for that nonetheless and abort the task
        QByteArray err = process->readAllStandardError();
        if (err.size() != 0 || doAbort)
        {
            task->setOutput(err);
            task->setReturnCode(-1);

            processingTask = false;
            buffer.clear();
            output.clear();
            queue.dequeue();

            emit taskAborted(task);
            continue;
        }

        buffer.append(process->readAllStandardOutput());

        int returnCode = -1;

        while (!buffer.isEmpty())
        {
            StdioParser parser(buffer);

            // if the chunk is not yet complete, try again later
            if (!parser.parse())
            {
                D("cannot parse - continueing");
                break;
            }

            TickerMap tickerUpdates = parser.getTickers();
            if (tickerUpdates.size() > 0)
            {
                if (tickers.size() == 0)
                {
                    tickers = tickerUpdates;
                }
                else
                {
                    // update previous ticker values
                    foreach (char ident, tickerUpdates.keys())
                    {
                        if (tickers.contains(ident))
                        {
                            // ticker symbols might not be unique during
                            // program execution and new tickers may share
                            // the symbol of an already completed ticker
                            if (tickers[ident].complete)
                                tickers[ident] = tickerUpdates[ident];
                            else
                                tickers[ident].update(tickerUpdates[ident]);
                        }
                        else
                        {
                            tickers[ident] = tickerUpdates[ident];
                        }
                    }
                }

                emit tickersChanged(tickers);
            }

            buffer = parser.getLeftBytes();
            output.append(parser.getPayload());

            // store and log out-of-band messages
            QList<MessagePair> msgs = parser.getOutOfBandMessages();

            task->appendOutOfBandMessages(msgs);

            foreach (MessagePair pair, msgs)
            {
                QString msg =
                    QString("monotone reports for thread %1, task %2: %3")
                        .arg(threadNumber)
                        .arg(task->getCommandNumber())
                        .arg(QString(pair.second));

                if (pair.first == 'w')
                {
                    DebugLog::warn(msg);
                    continue;
                }

                if (pair.first == 'p')
                {
                    DebugLog::info(msg);
                    continue;
                }

                if (pair.first == 'e')
                {
                    DebugLog::critical(msg);
                    continue;
                }

                I(false);
            }

            if (parser.getChunkType() == 'l')
            {
                returnCode = parser.getErrorCode();
                I(buffer.isEmpty());
                break;
            }
        }

        if (returnCode < 0)
            continue;

        task->setOutput(output);
        task->setReturnCode(returnCode);
        task->setFinished();

        queue.dequeue();

        processingTask = false;

        D(QString("thread %1: task %2 (%3) finished")
            .arg(threadNumber)
            .arg(task->getCommandNumber())
            .arg(QString(task->getEncodedInput())));

        emit taskFinished(task);
    }

    cleanup(process);
    emit stopped(threadNumber);
}

