/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "WorkspaceCommitter.h"
#include "MonotoneUtil.h"
#include "BasicIOParser.h"
#include "BasicIOWriter.h"

#include <QDateTime>

WorkspaceCommitter::WorkspaceCommitter(const MonotoneHandlePtr & handle,
                                       const GetRevision * model)
  : QObject(0), monotoneHandle(handle), revModel(model)
{
    I(monotoneHandle->getType() == MonotoneHandle::workspace_handle);
}

WorkspaceCommitter::~WorkspaceCommitter() {}

bool WorkspaceCommitter::run(const QString & key, const QString & branch,
    const QString & author, const QString & changelog)
{
    //
    // Prepare the data
    //

    QMap<QString, QString> certs;

    // use the signing key as default author cert value
    if (author.isEmpty())
        certs.insert("author", key);
    else
        certs.insert("author", author);

    certs.insert("branch", branch);
    certs.insert("changelog", changelog);
    certs.insert("date",
                 QDateTime::currentDateTime().toUTC().toString(Qt::ISODate));

    QDir workspaceDir(monotoneHandle->getData());
    I(workspaceDir.exists());

    //
    // Check the current revision text for file content changes and
    // process all of them
    //
    QMap<QString, QPair<QString, QString> > changedFiles =
        revModel->getChangedFiles();
    QMapIterator<QString, QPair<QString, QString> > it(changedFiles);

    while (it.hasNext())
    {
        it.next();
        // FIXME: what to do with already added file deltas?
        if (!putFile(it.key(), it.value().first, it.value().second))
            return false;
    }

    //
    // commit revision
    //

    // FIXME: this may cause problems with utf-8 encoded filenames
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "put_revision" << revModel->getRevisionText()
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    if (!task->isFinished()) F("task aborted");

    QString data = task->getDecodedOutput();
    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't commit revision: %1").arg(data));
        return false;
    }

    data.chop(1);
    revisionId = data;

    // attach certificates
    QMapIterator<QString, QString> i(certs);
    while (i.hasNext())
    {
        i.next();
        // FIXME: what to do with already added certs?
        if (!putCert(revisionId, key, i.key(), i.value()))
            return false;
    }

    D(QString("Successfully committed revision %1").arg(revisionId));

    // update _MTN/revision and _MTN/options
    return writeRevision(revisionId) && writeOptions(certs);
}

bool WorkspaceCommitter::writeRevision(const QString & revisionId)
{
    QDir workspaceDir(monotoneHandle->getData());
    I(workspaceDir.cd("_MTN"));

    QFile file(workspaceDir.filePath("revision"));
    if (!file.open(QIODevice::WriteOnly))
    {
        C("Can't open revision for writing");
        return false;
    }

    StanzaList stanzas;

    Stanza format;
    format.append(StanzaEntry("format_version", QStringList() << "1"));
    stanzas.append(format);

    Stanza manifest;
    manifest.append(StanzaEntry("new_manifest",
                                "[0000000000000000000000000000000000000000]"));
    stanzas.append(manifest);

    Stanza oldrev;
    oldrev.append(StanzaEntry("old_revision", "[" + revisionId + "]"));
    stanzas.append(oldrev);

    BasicIOWriter writer(stanzas);
    QString rev = writer.write();
    I(file.write(rev.toLatin1()));
    file.close();

    return true;
}

bool WorkspaceCommitter::writeOptions(const QMap<QString, QString> & certs)
{
    QDir workspaceDir(monotoneHandle->getData());
    I(workspaceDir.cd("_MTN"));

    QFile file(workspaceDir.filePath("options"));
    if (!file.open(QIODevice::ReadWrite))
    {
        C("Can't open options for reading and writing");
        return false;
    }

    // read and parse the basic_io stanzas and set the correct branch name
    BasicIOParser parser(QString::fromUtf8(file.readAll()));
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1);

    for (int i=0, j=stanzas[0].size(); i<j; i++)
    {
        if (stanzas[0][i].sym == "branch")
        {
            I(stanzas[0][i].vals.size() == 1);
            stanzas[0][i].vals[0] = certs["branch"];
        }
    }

    BasicIOWriter writer(stanzas);
    QString output = writer.write();
    I(file.resize(0));
    I(file.write(output.toLatin1()));
    file.close();

    return true;
}

bool WorkspaceCommitter::putFile(const QString & fileName,
                                 const QString & fileId,
                                 const QString & baseId)
{
    QDir workspaceDir(monotoneHandle->getData());
    QFile file(workspaceDir.filePath(fileName));
    if (!file.exists())
    {
        // FIXME: we need to bubble this up to the user, this typically
        // happens if there are missing files in the workspace
        C(QString("File %1 does not exist in workpace.").arg(fileName));
        return false;
    }

    if (!file.open(QIODevice::ReadOnly))
    {
        C("Can't open file for reading.");
        return false;
    }

    D(QString("Writing %1 to monotone").arg(fileName));

    ByteArrayList command;
    command << "put_file";
    if (!baseId.isEmpty())
    {
        command << baseId.toAscii();
    }
    command << file.readAll();
    file.close();

    MonotoneTaskPtr task(new MonotoneTask(command));
    MonotoneUtil::runSynchronousTask(monotoneHandle, task);
    if (!task->isFinished()) F("task aborted");

    QString data = task->getDecodedOutput();
    if (task->getReturnCode() > 0)
    {
        C(QString("Cannot execute put_file: %1").arg(data));
        return false;
    }

    // chop the newline
    data.chop(1);

    // sanity check: has the committed file the same fileid as get_revision
    // reported?
    if (data.compare(fileId) != 0)
    {
        C(QString("Wrong fileid for %1: expected %2, got %3")
            .arg(fileName).arg(fileId).arg(data));
        return false;
    }

    return true;
}

bool WorkspaceCommitter::putCert(const QString & rev,
                                 const QString & commitKey,
                                 const QString & key,
                                 const QString & value)
{
    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "cert" << rev << key << value,
        QStringList() << "key" << commitKey
    ));

    MonotoneUtil::runSynchronousTask(monotoneHandle, task);
    if (!task->isFinished()) F("task aborted");

    QString data = task->getDecodedOutput();
    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't attach cert %1: %2").arg(key).arg(data));
        return false;
    }
    return true;
}

