/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "WorkspaceCreator.h"
#include "BasicIOWriter.h"
#include "MonotoneUtil.h"

#include <QInputDialog>

WorkspaceCreator::WorkspaceCreator(const MonotoneHandlePtr & monotoneHandle,
                                   const QString & rev, const QDir & dir)
  : FileExporter(monotoneHandle, rev, dir)
{}

bool WorkspaceCreator::run()
{
    // at first determine the branch name(s) of the revision and let the
    // user select one branch of a list of branches
    CertList certs = MonotoneUtil::getCerts(monotoneHandle, revision);
    QStringList branches = certs.findByNameValues("branch");

    if (branches.size() == 0)
    {
        // TODO: shall we ask the user here for a new branch name?
        C(QString("Revision '%1' has no branch cert attached").arg(revision));
        return false;
    }

    QString selectedBranch = branches.at(0);

    if (branches.size() > 1)
    {
        bool ok;
        QString branch = QInputDialog::getItem(
            NULL, tr("Select a branch"),
            tr("The selected revision has more than one\n"
               "branch cert attached - please select one:"),
            branches, 0, false, &ok);

        if (!ok || branch.isEmpty())
        {
            D("User aborted checkout");
            return false;
        }
        selectedBranch = branch;
    }

    if (!createBookkeepingDirectory(rootDir, monotoneHandle,
                                    selectedBranch, revision))
    {
        return false;
    }

    // finally query and export the manifest contents
    return FileExporter::run(
        MonotoneUtil::getRevisionManifest(monotoneHandle, revision)
    );
}

bool WorkspaceCreator::createBookkeepingDirectory(QDir workspace,
                                                  const MonotoneHandlePtr & handle,
                                                  const QString & branch,
                                                  const QString & revision)
{
    L(QString("creating bookkeeping directory in %1")
        .arg(workspace.absolutePath()));

    if (!workspace.mkdir("_MTN"))
    {
        C("Could not create bookkeeping directory '_MTN'");
        return false;
    }

    I(workspace.cd("_MTN"));

    // _MTN/options
    QFile optionsFile(workspace.filePath("options"));
    if (!optionsFile.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        C("Could not open options file for writing.");
        return true;
    }

    StanzaList stanzas;
    Stanza stanza;
    stanza.append(StanzaEntry("database", QStringList() << handle->getData()));
    stanza.append(StanzaEntry("branch", QStringList() << branch));
    stanzas.append(stanza);

    BasicIOWriter writer(stanzas);

    optionsFile.write(writer.write().toLatin1());
    optionsFile.close();

    // _MTN/revision
    QFile revisionFile(workspace.filePath("revision"));
    if (!revisionFile.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        C("Could not open options file for writing");
        return false;
    }

    stanzas.clear();

    stanza.clear();
    stanza.append(StanzaEntry("format_version", QStringList() << "1"));
    stanzas.append(stanza);

    stanza.clear();
    stanza.append(StanzaEntry("new_manifest", "[0000000000000000000000000000000000000001]"));
    stanzas.append(stanza);

    stanza.clear();
    stanza.append(StanzaEntry("old_revision", "[" + revision + "]"));
    stanzas.append(stanza);

    BasicIOWriter writer2(stanzas);

    revisionFile.write(writer2.write().toLatin1());
    revisionFile.close();

    // _MTN/format
    QFile formatFile(workspace.filePath("format"));
    if (!formatFile.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        C("Could not open format file for writing");
        return false;
    }

    formatFile.write("2\n");
    formatFile.close();

    // _MTN/log (initally empty)
    QFile logFile(workspace.filePath("log"));
    if (!logFile.open(QIODevice::WriteOnly ))
    {
        C("Could not open log file for writing");
        return false;
    }
    logFile.close();

    I(workspace.cdUp());

    return true;
}

