/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DebugLog.h"
#include "Settings.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <QDir>
#include <QDate>
#include <QTime>

DebugLog* DebugLog::instance = 0;

DebugLog* DebugLog::singleton()
{
    if (!instance)
    {
        instance = new DebugLog();
    }
    return instance;
}

DebugLog::DebugLog() : loggedToConsole(false) {}

QStringList DebugLog::intro()
{
    QStringList lines;

    QString sep;
    sep.fill('=', 64);

    lines.append(sep);

    QDate today = QDate::currentDate();
    lines.append(
         QString("              guitone log started  (%1)")
            .arg(today.toString("yyyy-MM-dd"))
    );
    lines.append(
        QString("version: %1, revision: %2")
            .arg(GUITONE_VERSION).arg(GUITONE_REVISION)
    );
    lines.append(sep);

    return lines;
}

DebugLog::~DebugLog()
{
    if (logFile.isOpen()) logFile.close();
}

QByteArray DebugLog::format(Type t, QString msg)
{
    static QMap<int, QString> types;
    if (types.size() == 0)
    {
        types[Fatal]    = "fatal";
        types[Critical] = "critical";
        types[Warn]     = "warning";
        types[Info]     = "info";
        types[Debug]    = "debug";
    }

    // do not use I here since this calls qFatal and results in
    // an endless loop
    assert(t <= types.size());

    QTime now = QTime::currentTime();

    QString logStr = QString("%1: %2: %3\n")
        .arg(now.toString("hh:mm:ss.zzz"))
        .arg(types[t])
        .arg(msg);

    return logStr.toUtf8();
}

void DebugLog::doLog(Type t, QString msg)
{
    bool logToConsole = Settings::getBool("ConsoleLogEnabled");
    bool logToFile    = Settings::getBool("FileLogEnabled");

    if (!logToFile && !logToConsole)
        return;

    int logLevel = Settings::getLogLevel();
    if (logLevel == Fatal && t > Fatal)
        return;
    if (logLevel == Critical && t > Critical)
        return;
    if (logLevel == Warn && t > Warn)
        return;
    if (logLevel == Info && t > Info)
        return;
    if (logLevel == Debug && t > Debug)
        return;

    if (logToFile && !logFile.isOpen())
    {
        logFile.setFileName(logFilePath());

        QFlags<QIODevice::OpenModeFlag> flags =
            QIODevice::WriteOnly | QIODevice::Append |
            QIODevice::Text | QIODevice::Unbuffered;

        if (!logFile.open(flags))
        {
            Settings::setBool("FileLogEnabled", false);

            doLog(Critical, QString(
                "cannot open logfile '%1' for writing, disable file logging "
                "(error was: %2)"
            ).arg(logFilePath()).arg(logFile.errorString()));

            doLog(t, msg);
            return;
        }

        foreach (const QString & line, intro())
        {
            if (!logFile.write(format(Info, line)))
            {
                Settings::setBool("FileLogEnabled", false);

                doLog(Critical, QString(
                    "could not log to file, disabling file log (error: %1)"
                ).arg(logFile.errorString()));
                return;
            }
        }
    }
    else if (!logToFile && logFile.isOpen())
    {
        logFile.close();
    }

    if (!loggedToConsole && logToConsole)
    {
        foreach (const QString & line, intro())
        {
            fprintf(stderr, "%s", format(Info, line).constData());
        }
        loggedToConsole = true;
    }
    else
    if (loggedToConsole && !logToConsole)
    {
        loggedToConsole = false;
    }

    QByteArray formattedMsg = format(t, msg);

    // print the message on console
    if (logToConsole)
    {
        fprintf(stderr, "%s", formattedMsg.constData());
    }

    // print the message to the logfile
    if (logToFile && !logFile.write(formattedMsg))
    {
        Settings::setBool("FileLogEnabled", false);

        doLog(Critical, QString(
            "could not log to file, disabling file log (error: %1)"
        ).arg(logFile.errorString()));
    }
}

QString DebugLog::logFilePath()
{
    QString path(QDir::homePath());
    path.append(QDir::separator()).append("guitone.log");
    return path;
}

