/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DiffParser.h"
#include "vocab.h"

#include <QStringList>
#include <QRegExp>

DiffParser::DiffParser(const QByteArray & input, const QString & enc)
  : codec(0)
{
    // QTextCodec implements codecForName as factory and caches multiple
    // requests, so we don't need to do that ourselves
    codec = QTextCodec::codecForName(enc.toLatin1());
    I(codec);
    parse(input);
}

DiffParser::~DiffParser()
{
    qDeleteAll(fileDiffs);
    fileDiffs.clear();
}

void DiffParser::parse(const QByteArray & input)
{
    if (input.length() == 0)
    {
        W("Nothing to parse");
        return;
    }

    // remove last newline character
    QByteArray in(input);
    in.chop(1);

    // split the output into lines
    ByteArrayList lines = in.split('\n');

    QString curFile;

    Diff * curDiff = 0;
    DiffHunk * curHunk = 0;

    for (int i=0, s=lines.size(); i<s; i++)
    {
        QByteArray line(lines.at(i));
        I(line.length() > 0);
        char curChar = line.at(0);

        //
        // parse file diff separators, setup new DiffFile object
        //
        if (curChar == '=')
        {
            I(i < (s-1));

            // straightly go to the next line
            QByteArray nextLine(lines.at(++i));

            // create a new diff
            curDiff = new Diff();

            // `# path/to/file is binary'
            int isBinaryPos = nextLine.indexOf(" is binary");
            if (isBinaryPos != -1)
            {
                curDiff->is_binary = true;
                // we actually don't get this information from the diff
                curDiff->is_new_file = false;

                I(in.length() > isBinaryPos);
                curFile = nextLine.mid(2, isBinaryPos - 2);
            }
            else
            {
                curDiff->is_binary = false;

                // `--- path/to/file    123456789abcdef...'
                int tabPos = nextLine.indexOf('\t');
                I(tabPos != -1);
                curFile = nextLine.mid(4, tabPos - 4);
                QString fileVerOld = nextLine.mid(tabPos).trimmed();

                nextLine = lines.at(++i);
                tabPos = nextLine.indexOf('\t');
                I(tabPos != -1);
                QString fileVerNew = nextLine.mid(tabPos).trimmed();

                curDiff->is_new_file = fileVerOld == fileVerNew;
            }

            fileDiffs.insert(curFile, curDiff);
            continue;
        }

        I(curDiff);

        //
        // parse hunk lines, setup new hunk object
        //
        if (curChar == '@')
        {
            // this line is far easier to parse with a regex and it should not
            // contain any non-ascii character
            QString str(line);
            I(str.length() == line.length());

            QRegExp rx = QRegExp("^@@\\s+\\-(\\d+)(,(\\d+))?\\s+\\+(\\d+)(,(\\d+))?\\s+@@");
            I(rx.indexIn(str) > -1);
            QStringList list = rx.capturedTexts();
            I(list.size() >= 2);

            // create a new Hunk
            curHunk = new DiffHunk();
            curDiff->hunks.push_back(curHunk);

            // extract line numbers and positions
            curHunk->leftStart = list.at(1).toInt();
            if (list.at(2).length() > 0)
                curHunk->leftCount = list.at(3).toInt();
            else
                curHunk->leftCount = 1;

            curHunk->rightStart = list.at(4).toInt();
            if (list.at(5).length() > 0)
                curHunk->rightCount = list.at(6).toInt();
            else
                curHunk->rightCount = 1;

            continue;
        }

        I(curHunk);

        DiffLine::State state;

        if (curChar == '+') state = DiffLine::Added;
        else if (curChar == '-') state = DiffLine::Removed;
        else if (curChar == ' ') state = DiffLine::Unchanged;
        else if (curChar == '\\')
        {
            // if we have a "no newline at the end of the file" line we
            // also always have an ancestor line which has missing this
            // particular \n character
            I(curHunk->lines.size() > 0);

            DiffLine * lastLine = curHunk->lines.last();
            switch (lastLine->state)
            {
                case DiffLine::Added:
                    lastLine->state = DiffLine::AddedMissingNewline;
                    break;
                case DiffLine::Removed:
                    lastLine->state = DiffLine::RemovedMissingNewline;
                    break;
                case DiffLine::Unchanged:
                    lastLine->state = DiffLine::UnchangedMissingNewline;
                    break;
                default: I(false);
            }

            // we do not create a new diff line for this pseudo line
            continue;
        }
        else I(false);

        DiffLine * diffLine = new DiffLine();
        curHunk->lines.push_back(diffLine);
        diffLine->content = codec->toUnicode(line.mid(1));
        diffLine->state = state;
    }
}

Diff * DiffParser::getDiff(const QString & filename)
{
    if (!fileDiffs.contains(filename)) return 0;
    return fileDiffs.value(filename);
}

