/***************************************************************************
 *   Copyright (C) 2006 by Ingo Maindorfer								   *
 *   ingo@liquidcooling.de												   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "IconProvider.h"
#include "InventoryItem.h"

#include <QStyle>
#include <QApplication>
#include <QPainter>

IconProvider * IconProvider::instance = 0;

IconProvider * IconProvider::singleton()
{
    if (instance == 0)
    {
        instance = new IconProvider();
    }
    return instance;
}

IconProvider::IconProvider()
{
    QStyle * style = QApplication::style();
    int iconSize = style->pixelMetric(QStyle::PM_ListViewIconSize);

    // default icons if no state matches

    fileIcons["default"]   = QIcon(style->standardIcon(QStyle::SP_FileIcon));
    folderIcons["default"] = QIcon(style->standardIcon(QStyle::SP_DirClosedIcon));

    folderIcons["default"].addPixmap(
         style->standardIcon(QStyle::SP_DirOpenIcon).pixmap(iconSize),
         QIcon::Active, QIcon::On
    );

    folderIcons["cdup"] = QIcon(style->standardIcon(QStyle::SP_FileDialogToParent));

    // node overlay states
    QMap<QString, QString> states;
    states["added"]             = QString(":/overlays/added.png");
    states["attr_changed"]      = QString(":/overlays/attributes_changed.png");
    states["cont_attr_changed"] = QString(":/overlays/content_attributes_changed.png");
    states["cont_changed"]      = QString(":/overlays/content_changed.png");
    states["dropped"]           = QString(":/overlays/dropped.png");
    states["ignored"]           = QString(":/overlays/ignored.png");
    states["invalid"]           = QString(":/overlays/invalid.png");
    states["missing"]           = QString(":/overlays/missing.png");
    states["renamed"]           = QString(":/overlays/renamed.png");
    states["unknown"]           = QString(":/overlays/unknown.png");

    QMapIterator<QString, QString> i(states);
    Qt::AspectRatioMode ratio        = Qt::KeepAspectRatio;
    Qt::TransformationMode transform = Qt::SmoothTransformation;

    while (i.hasNext())
    {
        i.next();
        QImage overlay(i.value());
        overlay.createAlphaMask();

        // the basic icons have to be recreated for each state
        // since we later paint on them
        QImage fileImage(
             fileIcons["default"].
                pixmap(iconSize).
                    toImage()
        );
        QImage folderImageClosed(
             folderIcons["default"].
                pixmap(iconSize, QIcon::Normal, QIcon::Off).
                    toImage()
        );
        QImage folderImageOpened(
             folderIcons["default"].
                pixmap(iconSize, QIcon::Normal, QIcon::On).
                    toImage()
        );

        // note: the overlays are scaled to fit the different icon sizes
        // on different platforms / with different stylings, so the overlay
        // always perfectly fits the actual image

        //
        // file icon
        //
        QPainter filePainter(&fileImage);
        if (i.key() == "ignored")
        {
            filePainter.setCompositionMode(QPainter::CompositionMode_SourceAtop);
        }

        filePainter.drawImage(
            0, 0, overlay.scaled(fileImage.size(), ratio, transform)
        );
        filePainter.end();
        fileIcons[i.key()] = QIcon(QPixmap::fromImage(fileImage));

        //
        // folder icons
        //
        QPainter folderPainter1(&folderImageClosed);
        if (i.key() == "ignored")
        {
            folderPainter1.setCompositionMode(QPainter::CompositionMode_SourceAtop);
        }

        folderPainter1.drawImage(
            0, 0, overlay.scaled(folderImageClosed.size(), ratio, transform)
        );
        folderPainter1.end();

        QPainter folderPainter2(&folderImageOpened);
        if (i.key() == "ignored")
        {
            folderPainter2.setCompositionMode(QPainter::CompositionMode_SourceAtop);
        }

        folderPainter2.drawImage(
            0, 0, overlay.scaled(folderImageOpened.size(), ratio, transform)
        );
        folderPainter2.end();

        folderIcons[i.key()] = QIcon(QPixmap::fromImage(folderImageClosed));
        folderIcons[i.key()].addPixmap(QPixmap::fromImage(folderImageOpened),
                                       QIcon::Active, QIcon::On);
    }
}

IconProvider::~IconProvider() {}

QIcon IconProvider::getIcon(const ModelItem * item) const
{
    const PseudoItem * psitem = qobject_cast<const PseudoItem *>(item);
    if (psitem)
    {
        if (psitem->isCdUp())
        {
            return folderIcons.value("cdup");
        }
        // FIXME: do we want to create a special node item?
        return folderIcons.value("default");
    }

    const InventoryItem * invitem = qobject_cast<const InventoryItem *>(item);
    if (invitem)
    {
        // we do not display all possible item state combinations, but
        // give certain ones a precedence over others:
        //
        // 1) invalid state
        // 2) missing state
        // 3) content/attribute changes
        // 4) content changes
        // 5) attribute changes
        // 6) tree changes (adds, drops, renames, new-state preferred!)
        // 7) unknown node
        // 8) ignored node

        QString state = "default";

        if (invitem->hasStatus(InventoryItem::Invalid))
        {
            state = "invalid";
        }
        else
        if (invitem->hasStatus(InventoryItem::Missing))
        {
            state = "missing";
        }
        else
        if (invitem->hasStatus(InventoryItem::AttributesChanged |
                               InventoryItem::ContentsChanged))
        {
            state = "cont_attr_changed";
        }
        else
        if (invitem->hasStatus(InventoryItem::ContentsChanged))
        {
            state = "cont_changed";
        }
        else
        if (invitem->hasStatus(InventoryItem::AttributesChanged))
        {
            state = "attr_changed";
        }
        else
        if (invitem->hasStatus(InventoryItem::RenameTarget))
        {
            state = "renamed";
        }
        else
        if (invitem->hasStatus(InventoryItem::Added))
        {
            state = "added";
        }
        else
        if (invitem->hasStatus(InventoryItem::RenameSource))
        {
            state = "renamed";
        }
        else
        if (invitem->hasStatus(InventoryItem::Dropped))
        {
            state = "dropped";
        }
        else
        if (invitem->hasStatus(InventoryItem::Unknown))
        {
            state = "unknown";
        }
        else
        if (invitem->hasStatus(InventoryItem::Ignored))
        {
            state = "ignored";
        }

        if (invitem->isDirectory())
        {
            return folderIcons.value(state);
        }

        return fileIcons.value(state);
    }

    // default file icon for any other ModelItem
    return fileIcons.value("default");
}

QIcon IconProvider::getPlainFileIcon() const
{
    return getFileIcon("default");
}

QIcon IconProvider::getPlainFolderIcon() const
{
    return getFolderIcon("default");
}

QIcon IconProvider::getFileIcon(const QString & name) const
{
    if (!fileIcons.contains(name))
        return QIcon();
    return fileIcons.value(name);
}

QIcon IconProvider::getFolderIcon(const QString & name) const
{
    if (!folderIcons.contains(name))
        return QIcon();
    return folderIcons.value(name);
}

