/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "vocab.h"
#include "Platform.h"
#include "Settings.h"

#include <QApplication>
#include <QFileInfo>
#include <QProcess>
#include <QDir>

#ifdef Q_WS_WIN
#include <windows.h>
#include <lmcons.h>
#endif

// needed for MS Visual Studio
#ifdef _MSC_VER
#pragma comment(lib, "shell32.lib")
#pragma comment(lib, "Advapi32.lib")
#endif

#ifdef Q_WS_X11
#include <cstdlib>
using namespace std;
#endif

//
// The following code is public domain, with slight modifications, taken from
// http://hrabia.ibib.waw.pl/~winnie/qtopen.html
// It opens a file with the associated program on the specific platform
// Original author: Chris Thompson <chris@hypocrite.org>
//
void Platform::openFile(const QString & filename, bool isSymlink /* = false */)
{
    // if this looks like a file, and not like an URL or something else
    if (filename.indexOf(QRegExp("^\\w+:\\/\\/")) == -1)
    {
        QFileInfo fileInfo(filename);
        if (!fileInfo.exists())
        {
            throw GuitoneException(
                QObject::tr("The file does not exist.")
            );
        }

        if (fileInfo.isSymLink())
        {
            if (isSymlink)
            {
                throw GuitoneException(
                   QObject::tr("No support for nested symlinks.")
                );
            }
            openFile(fileInfo.symLinkTarget(), true);
            return;
        }
    }

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    int retval = -1;

#ifdef Q_WS_X11
    // due to a bug in KDE < 3.5.3 we can't trust the return value kfmclient
    // gives us back (its always 1!)
    if (system("which kfmclient >/dev/null 2>&1") == 0)
    {
        D("Using kfmclient");
        string s("kfmclient exec \"");
        s += filename.toUtf8().data();
        s += "\"";
        // just cross your fingers
        retval = system(s.c_str());
    }
    else
    if (system("which gnome-open >/dev/null 2>&1") == 0)
    {
        D("Using gnome-open");
        string s("gnome-open \"");
        s += filename.toUtf8().data();
        s += "\"";
        retval = system(s.c_str());
    }
    else
    {
        throw GuitoneException(
            QObject::tr("Neither kfmclient nor gnome-open found.")
        );
    }
#endif

#ifdef Q_WS_MACX
    retval = QProcess::execute("open", QStringList() << filename);
    // if a regular open failed, try to open the file in the system's
    // default editor
    if (retval != 0)
    {
        D(QString("try to open %1 in text editor").arg(filename));
        retval = QProcess::execute("open", QStringList() << "-t" << filename);
    }
#endif

#ifdef Q_WS_WIN
    QString winfilename(filename);
    winfilename.replace(QChar('/'), QString("\\"));
    int code = reinterpret_cast<int>(
        ShellExecute(NULL, L"open", winfilename.toStdWString().data(),
                     NULL, NULL, SW_SHOW)
    );

    // if shellexec fails (return value < 32, see
    // http://msdn2.microsoft.com/en-us/library/bb762153.aspx)
    // f.e. because the file's extension is not registered,
    // try to open in with the "Open With" dialog
    if (code > 32)
    {
        retval = 0;
    }
    else
    {
        QStringList arguments;
        arguments << "shell32.dll,OpenAs_RunDLL" << winfilename;
        retval = QProcess::execute("rundll32.exe", arguments);
    }
#endif

    QApplication::restoreOverrideCursor();

    if (retval != 0)
    {
        throw GuitoneException(QObject::tr(
            "Unable to open files on your platform - please "
            "contact the author about this problem."
        ));
    }
}

//
// the following code is loosely based on David Wright's code snippet
// on the Qt interest mailing list
// (http://lists.trolltech.com/qt-interest/2002-12/thread00130-0.html)
//
QString Platform::getUsername()
{
    QString username;

#ifdef Q_WS_WIN
#if defined(UNICODE)
    if (QSysInfo::WindowsVersion & QSysInfo::WV_NT_based)
    {
        TCHAR winUserName[UNLEN + 1]; // UNLEN is defined in LMCONS.H
        DWORD winUserNameSize = sizeof(winUserName);
        GetUserNameW(winUserName, &winUserNameSize);
        username = QString::fromStdWString(winUserName);
    } else
#endif
    {
        char winUserName[UNLEN + 1]; // UNLEN is defined in LMCONS.H
        DWORD winUserNameSize = sizeof(winUserName);
        GetUserNameA(winUserName, &winUserNameSize);
        username = QString::fromLocal8Bit(winUserName);
    }
#endif

#ifdef Q_WS_X11
#define IS_UNIX
#endif

#ifdef Q_WS_MACX
#define IS_UNIX
#endif

#ifdef IS_UNIX
    username = QString::fromAscii(getenv("USER"));
#endif

    D(QString("Found username %1").arg(username));

    return username;
}

QDir Platform::safeMakePath(const QString & path)
{
    QFileInfo fileInfo(path);
    if (fileInfo.exists())
    {
        if (!fileInfo.isDir())
        {
            throw GuitoneException(
                QObject::tr("Path exists, but is not a directory.")
            );
        }

        if (!fileInfo.isWritable())
        {
            throw GuitoneException(
                QObject::tr("Directory exists, but is not writable.")
            );
        }

        QDir dir(fileInfo.absoluteFilePath());
        QDir::Filters filters = QDir::NoDotAndDotDot | QDir::AllEntries |
                                QDir::Hidden | QDir::System;
        if (dir.entryList(filters).size() > 0)
        {
            throw GuitoneException(
                QObject::tr("Directory is not empty.")
            );
        }
        return dir;
    }

    if (path.indexOf(QRegExp("(^|[\\/])\\.\\.([\\/]|$)")) != -1)
    {
        throw GuitoneException(
            QObject::tr("Relative paths are not supported.")
        );
    }

    // default to the application's current directory if we get a
    // relative filename on all platforms
    QString prefixPath(QDir::currentPath());
    QString actualPath(path);

#ifdef Q_WS_WIN
    if (path.left(2) == "\\\\")
    {
        throw GuitoneException(
            QObject::tr("UNC paths are not supported.")
        );
    }

    if (path.mid(1, 2) == ":\\" || path.mid(1, 2) == ":/")
    {
        actualPath = path.mid(3);
        prefixPath = path.left(3).replace('\\', '/').toUpper();
        bool driveFound = false;
        foreach (QFileInfo drive, QDir::drives())
        {
            if (prefixPath == drive.absolutePath())
            {
                driveFound = true;
                break;
            }
        }
        if (!driveFound)
        {
            throw GuitoneException(
                QObject::tr("Drive %1 does not exist.").arg(prefixPath)
            );
        }
    }
#else
    if (path.left(1) == "/")
    {
        prefixPath = "/";
        actualPath = path.mid(1);
    }
#endif

    QDir dir(prefixPath);
    if (!dir.mkpath(actualPath))
    {
        throw GuitoneException(
            QObject::tr("Path could not be created.")
        );
    }

    return dir;
}

void Platform::removeRecursively(const QString & path)
{
    QFileInfo fileInfo(path);
    if (!fileInfo.exists())
    {
        throw GuitoneException(
            QObject::tr("Path '%1' does not exist.").arg(path)
        );
    }

    if (!fileInfo.isWritable())
    {
        throw GuitoneException(
            QObject::tr("Path '%1' is not writable.").arg(path)
        );
    }

#ifdef Q_WS_WIN
    if (path.left(2) == "\\\\")
    {
        throw GuitoneException(
            QObject::tr("UNC paths are not supported.")
        );
    }

    // sanity
    if (path.size() == 3 && (path.mid(1, 2) == ":\\" || path.mid(1, 2) == ":/"))
    {
        throw GuitoneException(
            QObject::tr("Cannot remove drive '%1'.").arg(path)
        );
    }
#else
    if (path.size() == 1 && path.left(1) == "/")
    {
        throw GuitoneException(
            QObject::tr("Cannot remove root directory.")
        );
    }
#endif

    if (!fileInfo.isDir())
    {
        if (!QFile::remove(path))
        {
            throw GuitoneException(
                QObject::tr("Cannot remove file '%1'.").arg(path)
            );
        }
        return;
    }

    QDir dir(path);
    QStringList entries = dir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot |
                                        QDir::Hidden | QDir::System);
    foreach (QString entry, entries)
    {
        Platform::removeRecursively(dir.filePath(entry));
    }

    QString dirName = dir.dirName();
    I(dir.cdUp());

    if (!dir.rmdir(dirName))
    {
        throw GuitoneException(
            QObject::tr("Cannot remove directory '%1'.").arg(path)
        );
    }
}

