/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "StdioParser.h"
#include "vocab.h"

StdioParser::StdioParser(const QByteArray & input) : AbstractParser(input) {}

bool StdioParser::parse()
{
    if (getLeftBytesCount() == 0) return false;

    // chunk format: <cmdNumber>:<chunkType>:<chunkSize>:<payload>
    commandNumber = getNumber();
    I(getNext() == ':');
    chunkType = getNext();
    I(getNext() == ':');
    chunkSize = getNumber();
    I(getNext() == ':');

    int charsLeft = getLeftBytesCount();
    if (chunkSize > charsLeft)
    {
        return false;
    }

    if (chunkType == 'm')
    {
        payload = getNext(chunkSize);
    }
    else
    if (chunkType == 't')
    {
        parseTicks(getNext(chunkSize));
    }
    else
    if (chunkType == 'l')
    {
        errorCode = getNumber();
    }
    else
    {
        oobMessages.push_back(qMakePair(chunkType, getNext(chunkSize)));
    }

    return true;
}

int StdioParser::getNumber()
{
    int number = 0;
    int processedChars = 0;

    while (true)
    {
        char ch = whatsNext();
        if (ch < '0' || ch > '9')
        {
            // ensure that we've read at least one char
            I(processedChars > 0);
            break;
        }
        number *= 10;
        number += (ch - '0');
        advance();
        processedChars++;
    }

    return number;
}

void StdioParser::parseTicks(const QByteArray & ticks)
{
    I(ticks.endsWith(';'));

    QList<QByteArray> tickerList = ticks.left(ticks.size() - 1).split(';');
    I(tickerList.size() > 0);

    foreach (QByteArray tick, tickerList)
    {
        I(tick.size() > 0);
        char ident = tick.at(0);
        tick.remove(0, 1);

        if (!tickers.contains(ident))
        {
            tickers[ident] = Ticker();
        }

        if (tick.size() == 0)
        {
            tickers[ident].complete = true;
            continue;
        }

        if (tick.at(0) == ':')
        {
            tick.remove(0, 1);
            tickers[ident] = Ticker(tick);
            continue;
        }

        if (tick.at(0) == '=')
        {
            tick.remove(0, 1);
            bool ok;
            tickers[ident].total = tick.toUInt(&ok);
            I(ok);
            continue;
        }

        if (tick.at(0) == '#')
        {
            tick.remove(0, 1);
            bool ok;
            tickers[ident].progress = tick.toUInt(&ok);
            I(ok);
            continue;
        }

        I(false);
    }
}

QList<QPair<QByteArray, QByteArray> > StdioParser::parseAndCheckHeaders(const QByteArray & in)
{
    ByteArrayList headerLines = in.trimmed().split('\n');
    I(headerLines.count() > 0);

    QList<QPair<QByteArray, QByteArray> > out;

    foreach (QByteArray line, headerLines)
    {
        int sep = line.indexOf(": ");
        I(sep >= 1);
        QByteArray header = line.left(sep);
        QByteArray value  = line.mid(sep + 2);

        if (header == "format-version" &&
            StdioParser::STDIO_VERSION != value.toInt())
        {
            throw GuitoneException(QString("wrong stdio version, wanted %1, got %2")
                .arg(StdioParser::STDIO_VERSION).arg(value.toInt()));
        }
        out.append(qMakePair(header, value));
    }

    return out;
}

