/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "AnnotateFile.h"
#include "MonotoneUtil.h"

#include <QMessageBox>
#include <QMenu>

AnnotateFile::AnnotateFile(QWidget * parent, const MonotoneHandlePtr & handle)
    : Dialog(parent), monotoneHandle(handle), annotateModel(0)
{
    setupUi(this);
    Dialog::init();

    // clear all remembered revisions when the dialog is closed so
    // we start with a fresh set of items for the next annotation
    connect(
        this, SIGNAL(dialogHidden()),
        baseRevision, SLOT(clear())
    );

    certList->setKeyMap(MonotoneUtil::getPublicKeyMap(monotoneHandle));

    annotateModel = new Annotate(this);
    annotationView->setModel(annotateModel);

    connect(
        annotateModel, SIGNAL(annotationRead()),
        progressWidget, SLOT(hide())
    );

    connect(
        annotateModel, SIGNAL(currentCertList(const CertList &)),
        certList, SLOT(setCertList(const CertList &))
    );

    connect(
        annotationView, SIGNAL(selectedRows(const QModelIndexList &)),
        this, SLOT(rowsSelected(const QModelIndexList &))
    );

    connect(
        annotationView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuRequested(const QModelIndexList &, const QPoint &))
    );

    connect(
        annotationView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(doubleClicked(const QModelIndex &))
    );

    connect(
        selectRevisionButton, SIGNAL(clicked()),
        this, SIGNAL(selectRevision())
    );

    connect(
        baseRevision, SIGNAL(currentIndexChanged(const QString &)),
        this, SLOT(setSelectedRevision(const QString &))
    );

    progressWidget->hide();
}

AnnotateFile::~AnnotateFile()
{
    if (annotateModel) delete annotateModel;
}

void AnnotateFile::read(const QString & rev, const QString & file)
{
    if (rev == annotateModel->getBaseRevision() && file == filePath)
    {
        return;
    }

    filePath = file;

    int pos = baseRevision->findText(rev);
    if (pos == -1)
    {
        baseRevision->insertItem(0, rev);
        pos = 0;
    }
    baseRevision->setCurrentIndex(pos);

    setWindowTitle(tr("Annotation of %1").arg(filePath));
    annotateModel->readAnnotation(monotoneHandle, rev, filePath);
    progressWidget->show();
}

void AnnotateFile::setSelectedRevision(const QString & rev)
{
    if (rev.isEmpty())
        return;

    QString currentBaseRevision = annotateModel->getBaseRevision();
    QString newPath = filePath;

    if (!currentBaseRevision.isEmpty())
    {
        newPath = MonotoneUtil::getCorrespondingPath(
            monotoneHandle, currentBaseRevision, filePath, rev
        );
    }

    if (newPath.isEmpty())
    {
        QMessageBox::information(
             this,
             tr("File does not exist"),
             tr("The file '%1' does not exist in the selected revision '%2'.")
                .arg(filePath).arg(rev),
             QMessageBox::Ok
        );
        return;
    }

    read(rev, newPath);
}

void AnnotateFile::rowsSelected(const QModelIndexList & indexes)
{
    if (indexes.size() == 0)
    {
        return;
    }

    QModelIndex selectedIndex = indexes.at(0);
    QString rev = selectedIndex.data(Qt::UserRole).toString();
    annotateModel->setSelectedRevision(rev);

    // clear the selection so the text under the selected row gets
    // easier readable. Beware that this fires this slot again and
    // if the guard above which checks for no selected indexes is
    // removed, then this leads to infinite recursion!
    annotationView->selectionModel()->clearSelection();
}

void AnnotateFile::contextMenuRequested(const QModelIndexList & indexList,
                                        const QPoint & pos)
{
    // FIXME: since we clear the selection in rowsSelected(), indexList is
    // always empty, so we can't use that. Beware that there could be a race
    // condition where rowsSelected() is executed _after_ contectMenuRequest(),
    // thus the annotate model does not have the correct current revision set!
    Q_UNUSED(indexList);
    QString rev = annotateModel->getSelectedRevision();

    QStringList parents = MonotoneUtil::resolveSelector(
        monotoneHandle, "p:" + rev
    );

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QAction * actSelectRevision = menu.addAction(tr("Select as base revision"));
    QAction * actOpenManifest   = menu.addAction(tr("View files in this revision"));

    QAction * actFileDiff = 0, * actRevisionDiff = 0;
    // FIXME: no support for revisions with multiple parents yet
    if (parents.size() == 1)
    {
        actFileDiff     = menu.addAction(tr("Show file differences"));
        actRevisionDiff = menu.addAction(tr("Show all revision differences"));
    }

    actSelectRevision->setFont(activeFont);

    QAction * act = menu.exec(pos);
    if (act == actSelectRevision)
    {
        setSelectedRevision(rev);
    }
    else if (act == actOpenManifest)
    {
        emit revisionManifest(rev);
    }
    else if (act == actFileDiff)
    {
        I(parents.size() == 1);
        emit fileDiff(filePath, parents.at(0), rev);
    }
    else if (act == actRevisionDiff)
    {
        I(parents.size() == 1);
        emit revisionDiff(QString(), parents.at(0), rev);
    }
}

void AnnotateFile::doubleClicked(const QModelIndex & index)
{
    if (!index.isValid())
        return;

    setSelectedRevision(index.data(Qt::UserRole).toString());
}
