/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "ApplicationUpdate.h"
#include "MonotoneManager.h"
#include "Platform.h"
#include "Settings.h"

#include <QSslCertificate>
#include <QSslSocket>
#include <QDomDocument>
#include <QMessageBox>

ApplicationUpdate::ApplicationUpdate(QWidget * parent)
  : Dialog(parent)
{
    setupUi(this);
    Dialog::init();

    setWindowModality(Qt::ApplicationModal);

    connect(
        visitWebsite, SIGNAL(clicked()),
        this, SLOT(openWebsite())
    );

    connect(
        skipVersion, SIGNAL(clicked()),
        this, SLOT(skipThisVersion())
    );

    httpConnection = new QHttp();

    // import the CAcert root certificate, otherwise the SSL handshake will
    // fail later on
    QFile certFile(":/cacert_root.pem");
    I(certFile.open(QIODevice::ReadOnly));
    QSslCertificate cert(&certFile, QSsl::Pem);
    QSslSocket * sslSocket = new QSslSocket(this);
    sslSocket->addCaCertificate(cert);
    httpConnection->setSocket(sslSocket);

    connect(
        httpConnection, SIGNAL(done(bool)),
        this, SLOT(processHttpResult(bool))
    );
}

ApplicationUpdate::~ApplicationUpdate()
{
    delete httpConnection;
}

/**
 * If the we check for a new application version silently, the user
 * is not nagged with error messages like f.e. that there is no new version
 * available or that the check failed. However, if there is a new version
 * available, the dialog is shown to the user.
 * Now if he clicks on "SkipThisVersion" in the dialog, he is not remembered
 * again on the next application start that there is a new version, BUT
 * only unless he triggers the version check by hand from the "File" menu
 * which calls checkForUpdates without the silent flag set to false. In
 * this case the "Skip this version" button is also made invisible since it
 * has no functionality here.
 */
void ApplicationUpdate::checkForUpdates(bool silent)
{
    if (httpConnection->hasPendingRequests())
    {
        C("HTTP connection has pending requests... aborting");
        return;
    }

    silentCheck = silent;

    QUrl url(APPCAST_URL);
    httpConnection->setHost(url.host(),
                            url.scheme() == "https" ?
                              QHttp::ConnectionModeHttps :
                              QHttp::ConnectionModeHttp);
    httpConnection->get(url.path());

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
}

void ApplicationUpdate::processHttpResult(bool error)
{
    QApplication::restoreOverrideCursor();

    if (error)
    {
        QString errStr = httpConnection->errorString();

        C(QString("unable to fetch update information: %1").arg(errStr));

        if (!silentCheck)
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("Unable to fetch update information:\n%1").arg(errStr),
                QMessageBox::Ok, 0, 0
            );
        }
        return;
    }

    int statusCode = httpConnection->lastResponse().statusCode();
    if (statusCode != 200)
    {
        C(QString("status code was: %1").arg(statusCode));

        if (!silentCheck)
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("Unable to fetch update information (error %1).").arg(statusCode),
                QMessageBox::Ok, 0, 0
            );
        }
        return;
    }

    QDomDocument doc;
    QString err;
    int line, col;

    if (!doc.setContent(httpConnection->readAll(), true, &err, &line, &col))
    {
        W(QString("Cannot parse document: %1 (line %2, col %3)")
          .arg(err).arg(line).arg(col));
        return;
    }

    QDomElement channel = doc.documentElement().firstChild().toElement();
    I(!channel.isNull());

    QDomNodeList items = channel.elementsByTagName("item");

    latestVersion = GUITONE_VERSION;
    QString versionDescription;

    for (int i=0, j=items.size(); i<j; i++)
    {
        QDomElement item = items.item(i).toElement();
        if (item.isNull()) continue;
        QDomNodeList enclosures = item.elementsByTagName("enclosure");

        for (int k=0, l=enclosures.size(); k<l; k++)
        {
            QDomElement enclosure = enclosures.item(k).toElement();
            if (enclosure.isNull()) continue;
            QString version = enclosure.attributeNS(
                "http://www.andymatuschak.org/xml-namespaces/sparkle",
                "version", "not-found"
            );

            // each item can have multiple enclosures, of which at least one
            // should have a version field
            if (version == "not-found") continue;

            if (MonotoneManager::versionCompare(version, latestVersion) > 0)
            {
                latestVersion = version;
                QDomNodeList descs = item.elementsByTagName("description");
                I(descs.size() == 1);
                QDomElement desc = descs.item(0).toElement();
                I(!desc.isNull());
                versionDescription = desc.text();
            }
        }
    }

    if (latestVersion != GUITONE_VERSION)
    {
        QString versionToSkip = Settings::getString("SkipGuitoneVersion");

        if (silentCheck && !versionToSkip.isEmpty() &&
            MonotoneManager::versionCompare(latestVersion, versionToSkip) <= 0)
        {
            L(QString("skipping %1 per user request").arg(latestVersion));
            return;
        }

        textBrowser->setHtml(versionDescription);
        if (!silentCheck)
        {
            skipVersion->setVisible(false);
        }

        versionLabel->setText(versionLabel->text()
                              .arg(latestVersion)
                              .arg(GUITONE_VERSION));
        show();
    }
    else
    {
        if (!silentCheck)
        {
            QMessageBox::information(
                NULL,
                tr("No updates available"),
                tr("Your version of guitone (%1) is already up-to-date.")
                    .arg(GUITONE_VERSION),
                QMessageBox::Ok
            );
        }
    }
}

void ApplicationUpdate::openWebsite()
{
    Platform::openFile("https://guitone.thomaskeller.biz/g/download");
    accept();
}

void ApplicationUpdate::skipThisVersion()
{
    Settings::setString("SkipGuitoneVersion", latestVersion);
    accept();
}

