/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "ChangesetBrowser.h"
#include "Settings.h"
#include "MonotoneUtil.h"

#include <QClipboard>
#include <QMenu>
#include <QMessageBox>
#include <QInputDialog>

ChangesetBrowser::ChangesetBrowser(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle), branchModel(0)
{
    setupUi(this);
    Dialog::init();

    this->setWindowFlags(this->windowFlags() | Qt::WindowMaximizeButtonHint);
    this->setWindowFlags(this->windowFlags() | Qt::WindowMinimizeButtonHint);

    multipleParents->setVisible(false);
    progressWidget->setVisible(false);
    fetchWidget->setVisible(false);

    branchesSplitter->restoreState();
    changesetsSplitter->restoreState();
    changeLogSplitter->restoreState();

    branchModel = new Branches(this, monotoneHandle);

    branchProxyModel = new BranchesProxyModel(this);
    branchProxyModel->setSourceModel(branchModel);

    branchView->setModel(branchProxyModel);

    connect(
        branchViewFilter, SIGNAL(textChanged(const QString &)),
        branchProxyModel, SLOT(filterByName(const QString &))
    );

    connect(
        branchView, SIGNAL(displayBranchesAsTree(bool)),
        branchModel, SLOT(displayBranchesAsTree(bool))
    );

    connect(
        branchView, SIGNAL(showSuspendedBranches(bool)),
        branchProxyModel, SLOT(showSuspendedBranches(bool))
    );

    connect(
        branchModel, SIGNAL(modelReset()),
        branchView, SLOT(expandAll())
    );

    connect(
        branchView, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(branchViewClicked(const QModelIndex &))
    );

    connect(
        branchView, SIGNAL(suspendBranch(const QString &)),
        this, SLOT(suspendBranch(const QString &))
    );

    connect(
        branchView, SIGNAL(reloadBranchList()),
        branchModel, SLOT(readBranches())
    );

    changeLog->setKeyMap(MonotoneUtil::getPublicKeyMap(monotoneHandle));

    branchLogModel = new GetBranchLog(this, monotoneHandle);
    changesetView->setModel(branchLogModel);
    changesetView->setRootIsDecorated(false);

    connect(
        branchLogModel, SIGNAL(readingStopped()),
        this, SLOT(readingStopped())
    );

    connect(
        this, SIGNAL(dialogHidden()),
        branchLogModel, SLOT(stopReading())
    );

    revisionModel = new GetRevision(this);
    revisionView->setModel(revisionModel);

    connect(
        revisionModel, SIGNAL(revisionRead()),
        this, SLOT(updateParentsList())
    );

    connect(
        revisionParents, SIGNAL(currentIndexChanged(const QString &)),
        revisionModel, SLOT(showChangesAgainstParent(const QString &))
    );

    connect(
        changesetView, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(changesetViewClicked(const QModelIndex &))
    );

    connect(
        pushAll, SIGNAL(clicked()),
        this, SLOT(readAll())
    );

    connect(
        pushMore, SIGNAL(clicked()),
        this, SLOT(readMore())
    );

    connect(
        pushStop, SIGNAL(clicked()),
        branchLogModel, SLOT(stopReading())
    );

    connect(
        changesetView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(changesetViewDoubleClicked(const QModelIndex &))
    );

    connect(
        changesetView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(changesetViewContextMenuRequested(const QModelIndexList &, const QPoint &))
    );

    connect(
        revisionView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(revisionViewDoubleClicked(const QModelIndex &))
    );

    connect(
        revisionView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(revisionViewContextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

ChangesetBrowser::~ChangesetBrowser()
{
    branchesSplitter->saveState();
    changesetsSplitter->saveState();
    changeLogSplitter->saveState();

    delete branchLogModel;
    delete revisionModel;
    delete branchModel;
}

void ChangesetBrowser::init()
{
    branchModel->readBranches();
}

void ChangesetBrowser::readAll()
{
    progressWidget->setVisible(true);
    branchLogModel->readAll();
}

void ChangesetBrowser::readMore()
{
    progressWidget->setVisible(true);
    branchLogModel->readMore();
}

void ChangesetBrowser::readingStopped()
{
    progressWidget->setVisible(false);

    // yes, this looks awkward, but there is no other sane
    // way to "reset" the state of a set of exclusive buttons
    pushMore->setAutoExclusive(false);
    pushMore->setChecked(false);
    pushMore->setAutoExclusive(true);

    pushAll->setAutoExclusive(false);
    pushAll->setChecked(false);
    pushAll->setAutoExclusive(true);
}

void ChangesetBrowser::branchViewClicked(const QModelIndex & idx)
{
    // skip clicks on parent nodes in tree mode
    if (branchProxyModel->rowCount(idx) > 0)
        return;

    QString branch = branchProxyModel->data(idx, Qt::ToolTipRole).toString();

    if (branch == branchLogModel->getCurrentBranch())
    {
        return;
    }

    fetchWidget->setVisible(true);
    progressWidget->setVisible(true);
    pushMore->setDown(false);

    branchLogModel->readMore(branch);
}

void ChangesetBrowser::changesetViewClicked(const QModelIndex & idx)
{
    QModelIndex revIdx = branchLogModel->index(idx.row(), 3, QModelIndex());
    currentRevision = branchLogModel->data(revIdx, Qt::DisplayRole).toString();
    changeLog->setCertList(branchLogModel->getCerts(currentRevision));
    revisionModel->readExistingRevision(monotoneHandle, currentRevision);
    revisionParents->clear();
    multipleParents->setVisible(false);
}

void ChangesetBrowser::changesetViewDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;
    QModelIndex revIdx = branchLogModel->index(index.row(), 3, QModelIndex());
    QString rev(revIdx.data().toString());

    QStringList parents = branchLogModel->getParents(rev);
    if (parents.size() == 0)
    {
        emit revisionManifest(rev);
    }
    else
    {
        emit diffRevision(QString(), parents.at(0), rev);
    }
}

void ChangesetBrowser::changesetViewContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QModelIndex revIdx =
        branchLogModel->index(indexList.at(0).row(), 3, QModelIndex());

    QString rev(revIdx.data().toString());

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QList<QAction *> actParentDiff;
    QStringList parents = branchLogModel->getParents(rev);
    if (parents.size() > 0)
    {
        foreach (const QString & parent, parents)
        {
            actParentDiff.append(menu.addAction(
                tr("Show differences against parent %1...")
                    .arg(parent.left(8))
            ));
        }
        menu.addSeparator();
    }

    QAction * actOpenManifest = menu.addAction(tr("View files in this revision"));
    QAction * actCopyRevisionId = menu.addAction(tr("Copy revision ID"));
    menu.addSeparator();
    QAction * actAddCert = menu.addAction(tr("Add tag or other certificate"));

    if (actParentDiff.size() == 0)
    {
        actOpenManifest->setFont(activeFont);
    }
    else
    {
        actParentDiff.at(0)->setFont(activeFont);
    }

    QAction * act = menu.exec(pos);
    if (act == actOpenManifest)
    {
        emit revisionManifest(rev);
    }
    else
    if (act == actCopyRevisionId)
    {
        QClipboard * clipboard = QApplication::clipboard();
        clipboard->setText(rev);
    }
    else
    if (act == actAddCert)
    {
        emit createCert(rev, "tag");
    }
    else
    {
        int pos = actParentDiff.indexOf(act);
        if (pos != -1)
        {
            emit diffRevision(QString(), parents.at(pos), rev);
        }
    }
}

void ChangesetBrowser::updateParentsList()
{
    QStringList parentRevs = revisionModel->getParentRevisions();
    if (parentRevs.size() > 1)
    {
        multipleParents->setVisible(true);
        revisionParents->addItems(revisionModel->getParentRevisions());
    }
}

void ChangesetBrowser::revisionViewDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;

    QModelIndex label, path;
    if (index.column() == 0)
    {
        label = index;
        path = index.sibling(index.row(), 1);
    }
    else
    if (index.column() == 1)
    {
        label = index.sibling(index.row(), 0);
        path = index;
    }
    else
        I(false);

    // not a patched file
    if (label.data().toString() != QObject::tr("Patched"))
        return;

    emit diffFile(path.data().toString(),
                  revisionModel->getCurrentParentRevision(),
                  currentRevision);
}

void ChangesetBrowser::revisionViewContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QModelIndex index = indexList.at(0);

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QModelIndex label, path;
    if (index.column() == 0)
    {
        label = index;
        path = index.sibling(index.row(), 1);
    }
    else
    if (index.column() == 1)
    {
        label = index.sibling(index.row(), 0);
        path = index;
    }
    else
        I(false);

    if (label.data().toString() != QObject::tr("Patched"))
        return;

    QAction * actShowFileDiff = menu.addAction(tr("Show differences"));
    QAction * actAnnotate = menu.addAction(tr("Annotate"));

    // FIXME: change this as soon as there is an automate annotate
    if (monotoneHandle->getType() == MonotoneHandle::server_handle)
    {
        actAnnotate->setEnabled(false);
    }

    actShowFileDiff->setFont(activeFont);

    QAction * act = menu.exec(pos);
    if (act == actShowFileDiff)
    {
         emit diffFile(path.data().toString(),
                       revisionModel->getCurrentParentRevision(),
                       currentRevision);
    }
    else
    if (act == actAnnotate)
    {
        emit annotateFile(currentRevision, path.data().toString());
    }
}

void ChangesetBrowser::closeEvent(QCloseEvent * event)
{
    branchLogModel->stopReading();
    event->accept();
}

void ChangesetBrowser::handleCertCreation(const QString & revision,
                                          const QString & certName,
                                          const QString & certValue,
                                          const QString & key)
{
    Q_UNUSED(key);

    if (certName == "suspend")
    {
        branchModel->markBranchSuspended(certValue);
        branchLogModel->invalidateBranchCache(certValue);
        branchLogModel->invalidateCertCache(revision);
    }
    else
    {
        branchLogModel->invalidateCertCache(revision);
    }
}

void ChangesetBrowser::suspendBranch(const QString & branch)
{
    // since suspend certs are ignored by default already, we should not
    // need to check if the returned revisions have already been suspended
    QStringList revisions =
        MonotoneUtil::resolveSelector(monotoneHandle, "h:" + branch);

    if (revisions.size() == 0)
    {
        QMessageBox::critical(
            this,
            tr("Branch is empty"),
            tr("The branch '%1' is empty - maybe all heads "
               "have already been suspended?").arg(branch),
            QMessageBox::Ok
        );
        return;
    }

    QMap<QString, QString> privateKeys =
        MonotoneUtil::getPrivateKeyMap(monotoneHandle);

    if (privateKeys.size() == 0)
    {
        QMessageBox::critical(
            this,
            tr("No private key found"),
            tr("There are no private keys available with which to sign the "
               "suspend cert(s) for the %n heads. Please go to Database > "
               "Key Management and create a new private key first.", "",
               revisions.size()),
            QMessageBox::Ok
        );
        return;
    }

    QString key;
    if (privateKeys.size() > 1)
    {
        bool ok;
        QString keyname = QInputDialog::getItem(
            this,
            tr("Multiple private keys found"),
            tr("Please select the private key with which you want "
               "to sign the suspend cert(s) for the %n head(s):", "",
               revisions.size()),
            QStringList(privateKeys.values()),
            0, false, &ok
        );
        if (!ok)
            return;

        key = privateKeys.key(keyname);
    }
    else
    {
        key = privateKeys.begin().key();
    }

    foreach (const QString & revision, revisions)
    {
        if (!MonotoneUtil::createCert(monotoneHandle, revision, "suspend", branch, key))
        {
            QMessageBox::critical(
                this,
                tr("Cert creation failed"),
                tr("Could not create certificate. "
                   "Please check the logs for details."),
                QMessageBox::Ok
            );
            return;
        }

        emit certCreated(revision, "suspend", branch, key);
    }
}

