/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CheckoutRevision.h"
#include "MonotoneUtil.h"
#include "Settings.h"
#include "Platform.h"
#include "WorkspaceCreator.h"

#include <QMessageBox>
#include <QFileDialog>

CheckoutRevision::CheckoutRevision(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();

    branchesModel = new Branches(this, monotoneHandle);
    branchesModel->readBranches();
    branchList->setModel(branchesModel);

    tagsModel = new Tags(this, monotoneHandle);
    tagsModel->readTags(QString());
    tagList->setModel(tagsModel);

    selectorCompleter = new QCompleter(
         Settings::getItemList("RecentCheckoutSelectors"),
         this
    );

    selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
    selectedRevision->setCompleter(selectorCompleter);

    localCheckoutPath->setText(QDir::homePath());

    connect(
        selectRevision, SIGNAL(clicked()),
        this, SLOT(slotSelectRevision())
    );

    connect(
        branchList, SIGNAL(activated(int)),
        this, SLOT(selectBranchOption())
    );

    connect(
        tagList, SIGNAL(activated(int)),
        this, SLOT(selectTagOption())
    );

    connect(
        selectedRevision, SIGNAL(textChanged(const QString &)),
        this, SLOT(selectSelectorOption())
    );

    connect(
        browseDirs, SIGNAL(clicked()),
        this, SLOT(openDirBrowser())
    );
}

CheckoutRevision::~CheckoutRevision()
{
    delete branchesModel;
    delete tagsModel;
    delete selectorCompleter;
}

void CheckoutRevision::accept()
{
    QString selector;

    if (checkoutByBranch->isChecked())
    {
        QString selection(branchList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No branch selected"),
                tr("Please select a branch from the list of branches."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "h:" + selection;
    }

    if (checkoutByTag->isChecked())
    {
        QString selection(tagList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No tag selected"),
                tr("Please select a tag from the list of tags (if any are available)."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "t:" + selection;
    }

    if (checkoutByRevision->isChecked())
    {
        QString selection = selectedRevision->text();
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No selector given"),
                tr("Please enter a monotone selector or choose a revision by clicking on 'select revision'."),
                QMessageBox::Ok
             );
             return;
        }
        selector = selection;
        Settings::addItemToList("RecentCheckoutSelectors", selection, 10);
    }

    QStringList revList(MonotoneUtil::resolveSelector(monotoneHandle, selector));

    if (revList.size() == 0)
    {
        QMessageBox::warning(
            this,
            tr("No revisions matched"),
            tr("Your selection didn't match any revisions or was invalid."),
            QMessageBox::Ok
         );
         return;
    }

    if (revList.size() > 1)
    {
        QMessageBox::information(
            this,
            tr("Multiple revisions matched"),
            tr("Your selection matched multiple revisions. Please choose one."),
            QMessageBox::Ok
        );

        emit selectRevisionFromSelector(selector);
        return;
    }

    QDir checkoutDir;
    try
    {
        checkoutDir = Platform::safeMakePath(localCheckoutPath->text());
    }
    catch (GuitoneException & e)
    {
        QMessageBox::critical(
            this,
            tr("Could not create export directory"),
            tr("The directory could not be used / created:\n%1")
                .arg(e.what()),
            QMessageBox::Ok
         );
         return;
    }

    WorkspaceCreator creator(monotoneHandle, revList.at(0), checkoutDir);

    if (!creator.run())
    {
        QMessageBox::critical(
            this,
            tr("Checkout failed"),
            tr("The checkout failed - please report this problem to "
               "the author and attach the guitone.log in your home directory."),
            QMessageBox::Ok
        );
        return;
    }

    emit revisionCheckedOut(MonotoneHandle::create(checkoutDir.absolutePath()));
    done(0);
}

void CheckoutRevision::slotSelectRevision()
{
    emit selectRevisionFromSelector(QString());
}

void CheckoutRevision::setSelectedRevision(const QString & revision)
{
    selectedRevision->setText(revision);
}

void CheckoutRevision::selectBranchOption()
{
    checkoutByBranch->setChecked(true);
}

void CheckoutRevision::selectTagOption()
{
    checkoutByTag->setChecked(true);
}

void CheckoutRevision::selectSelectorOption()
{
    checkoutByRevision->setChecked(true);
}

void CheckoutRevision::openDirBrowser()
{
    QString fn = QFileDialog::getExistingDirectory(
        this, tr("Select a directory for checkout...")
    );

    if (!fn.isEmpty())
    {
        localCheckoutPath->setText(fn);
    }
}

