/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CommitRevision.h"
#include "Settings.h"
#include "MonotoneUtil.h"
#include "WorkspaceCommitter.h"

#include <QHeaderView>
#include <QMessageBox>
#include <QRegExp>
#include <QInputDialog>

CommitRevision::CommitRevision(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();

    revModel = new GetRevision(this);
    changeView->setModel(revModel);

    connect(
        revModel, SIGNAL(revisionRead()),
        this, SLOT(revisionRead())
    );

    connect(
        revModel, SIGNAL(readFailed(const QString &)),
        this, SLOT(readFailed(const QString &))
    );

    connect(
        previousChangelogEntryList, SIGNAL(currentIndexChanged(int)),
        this, SLOT(setChangelogEntryFromList(int))
    );

    connect(
        changesAgainstParent, SIGNAL(currentIndexChanged(const QString &)),
        revModel, SLOT(showChangesAgainstParent(const QString &))
    );

    connect(
        changeView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(changeViewDoubleClicked(const QModelIndex &))
    );
}

CommitRevision::~CommitRevision()
{
    delete revModel;
}

void CommitRevision::readWorkspaceRevision(const QStringList & paths)
{
    revModel->readWorkspaceRevision(monotoneHandle, paths);

    // while we're inserting items we don't want to receive these signals
    // i.e. we don't want to pre-fill the changelog text
    previousChangelogEntryList->blockSignals(true);
    previousChangelogEntryList->clear();

    QStringList entries = Settings::getItemList("ChangelogEntries");
    QRegExp re("\\s+");

    for (int i=0, j=entries.size(); i<j; i++)
    {
        QString entry = entries.at(i);
        QString shortened(entry);
        shortened.replace(re, " ");
        shortened = shortened.left(40);
        previousChangelogEntryList->insertItem(i, shortened, QVariant(entry));
    }

    previousChangelogEntryList->setCurrentIndex(-1);
    previousChangelogEntryList->blockSignals(false);

    QMap<QString, QString> privateKeys =
        MonotoneUtil::getPrivateKeyMap(monotoneHandle);

    if (privateKeys.size() == 0)
    {
        QMessageBox::critical(
            this,
            tr("No private key found"),
            tr("There are no private keys available with which to sign your "
               "commit. Please go to Database > Key Management and create a "
               "new private key first."),
            QMessageBox::Ok
        );
        done(0);
        return;
    }

    for (QMap<QString, QString>::const_iterator it = privateKeys.constBegin();
         it != privateKeys.constEnd(); it++)
    {
        keys->addItem(it.value(), it.key());
    }
}

void CommitRevision::setChangelogEntryFromList(int index)
{
    if (index < 0) return;
    changelogEntry->setPlainText(
        previousChangelogEntryList->itemData(index).toString()
    );
}

void CommitRevision::accept()
{
    // prevent further actions until everything run through
    buttonBox->setEnabled(false);

    Settings::addItemToList("ChangelogEntries", changelogEntry->toPlainText(), 10);

    QString branch = MonotoneUtil::getBranchName(monotoneHandle, QString());

    if (branch.isEmpty() && !checkAltBranch->isChecked())
    {
        QMessageBox::critical(
            this,
            tr("Workspace has no branch set"),
            tr("The workspace has no branch set. Please enter one."),
            QMessageBox::Ok
        );
        checkAltBranch->setChecked(true);
        buttonBox->setEnabled(true);
        return;
    }

    if (checkAltBranch->isChecked())
    {
        QString altBranchStr = altBranch->text().trimmed();
        if (altBranchStr.isEmpty())
        {
            QMessageBox::critical(
                this,
                tr("No branch specified"),
                tr("Please enter the name of an alternative branch."),
                QMessageBox::Ok
            );
            buttonBox->setEnabled(true);
            return;
        }
        branch = altBranchStr;
    }

    QString author;

    if (checkAltAuthor->isChecked())
    {
        QString altAuthorStr = altAuthor->text().trimmed();
        if (altAuthorStr.isEmpty())
        {
            QMessageBox::critical(
                this,
                tr("No author specified"),
                tr("Please enter the name of an alternative author."),
                QMessageBox::Ok
            );
            buttonBox->setEnabled(true);
            return;
        }
        author = altAuthorStr;
    }

    QString selectedKey = keys->itemData(keys->currentIndex()).toString();

    WorkspaceCommitter committer(monotoneHandle, revModel);

    if (!committer.run(selectedKey, branch, author, changelogEntry->toPlainText()))
    {
        QMessageBox::critical(
            this,
            tr("Could not commit revision"),
            tr("Unable to commit the revision - this may be a bug in guitone "
               "or a bug in monotone itself. Please take a closer look at the "
               "logs and optionally send in a bug report."),
            QMessageBox::Ok
        );
        buttonBox->setEnabled(true);
        done(0);
        return;
    }

    buttonBox->setEnabled(true);
    emit revisionCommitted(committer.getRevisionId());
    done(1);
}

void CommitRevision::revisionRead()
{
    changesAgainstParent->clear();

    QStringList parents = revModel->getParentRevisions();
    if (parents.size() > 0)
    {
        changesAgainstParent->addItems(parents);
        changesAgainstParent->setEnabled(true);
    }
    else
    {
        changesAgainstParent->setEnabled(false);
    }

    if (!Settings::getBool("DisableChangelogAutoCompletion"))
    {
        changelogEntry->setCompleterStrings(revModel->getPathsForCompletion());
    }
}

void CommitRevision::readFailed(const QString & error)
{
    QMessageBox::information(
        this,
        tr("Error reading workspace revision"),
        tr("The workspace revision could not be read. monotone returned:\n%1")
            .arg(error),
        QMessageBox::Ok
    );
    reject();
}

void CommitRevision::changeViewDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;

    QModelIndex label, path;
    if (index.column() == 0)
    {
        label = index;
        path = index.sibling(index.row(), 1);
    }
    else
    if (index.column() == 1)
    {
        label = index.sibling(index.row(), 0);
        path = index;
    }
    else
        I(false);

    // not a patched file
    if (label.data().toString() != QObject::tr("Patched"))
        return;

    emit diffFile(path.data().toString());
}

