/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CreateCert.h"
#include "MonotoneUtil.h"

#include <QMessageBox>

CreateCert::CreateCert(QWidget * parent, const MonotoneHandlePtr & handle)
    : Dialog(parent), database(handle)
{
    setupUi(this);
    Dialog::init();

    name->addItem(tr("Tag"), "tag");
    name->addItem(tr("Test result"), "testresult");
    name->addItem(tr("Comment"), "comment");
    name->addItem(tr("Custom..."), "");

    nameTypeMap.insert("tag", "single");
    nameTypeMap.insert("testresult", "boolean");
    nameTypeMap.insert("comment", "multi");

    typeWidgetMap.insert("single", singleLineValue);
    typeWidgetMap.insert("multi", multiLineValue);
    typeWidgetMap.insert("boolean", booleanValue);

    connect(
        name, SIGNAL(currentIndexChanged(int)),
        this, SLOT(nameIndexChanged(int))
    );

    connect(
        btnSelectRevision, SIGNAL(clicked()),
        this, SIGNAL(selectRevision())
    );
}

CreateCert::~CreateCert() {}

void CreateCert::init(const QString & rev, const QString & certName)
{
    signingKey->clear();

    QMap<QString, QString> privateKeys =
        MonotoneUtil::getPrivateKeyMap(database);

    if (privateKeys.size() == 0)
    {
        QMessageBox::critical(
            this,
            tr("No private key found"),
            tr("There are no private keys available with which to sign your "
               "certificate. Please go to Database > Key Management and "
               "create a new private key first."),
            QMessageBox::Ok
        );
        done(0);
        return;
    }

    QMapIterator<QString, QString> it(privateKeys);
    while (it.hasNext())
    {
        it.next();
        signingKey->addItem(it.value(), it.key());
    }

    revision->setText(rev);
    selectCertName(certName);
}

void CreateCert::selectCertName(const QString & certName)
{
    bool knownName = nameTypeMap.contains(certName);
    QString typeValue = knownName ? nameTypeMap.value(certName) : "single";

    QMapIterator<QString, QWidget *> it(typeWidgetMap);
    while (it.hasNext())
    {
        it.next();
        it.value()->setVisible(false);
    }

    I(typeWidgetMap.contains(typeValue));
    typeWidgetMap.value(typeValue)->setVisible(true);

    resize(sizeHint());

    if (!knownName)
    {
        name->lineEdit()->setReadOnly(false);
        name->setCurrentIndex(name->findData(""));
        name->lineEdit()->setText("x-");
    }
    else
    {
        name->lineEdit()->setReadOnly(true);
    }
}

void CreateCert::nameIndexChanged(int pos)
{
    QVariant var = name->itemData(pos);
    QString certName = var.isValid() ? var.toString() : "";
    selectCertName(certName);
}

void CreateCert::setSelectedRevision(const QString & rev)
{
    revision->setText(rev);
}

void CreateCert::accept()
{
    QString certName = name->itemData(name->currentIndex()).toString();
    if (certName.isEmpty())
    {
        certName = name->currentText();
        if (certName.isEmpty())
        {
            QMessageBox::critical(
                this,
                tr("No name for certificate"),
                tr("Please select a name for the certificate or enter one."),
                QMessageBox::Ok
            );
            return;
        }

        if (certName.indexOf(QRegExp("^x(-+[a-zA-Z0-9]+)+$")) != 0)
        {
            QMessageBox::critical(
                this,
                tr("Invalid certificate name"),
                tr("Custom certificates should always start with 'x-' and "
                   "should only contain ASCII letters, numbers and dashes."),
                QMessageBox::Ok
            );
            return;
        }
    }

    QString certValue;
    if (certName == "testresult")
    {
        certValue = booleanValue->isChecked() ? "true" : "false";
    }
    else
    if (certName == "comment")
    {
        certValue = multiLineValue->toPlainText();
    }
    else
    {
        certValue = singleLineValue->text();
    }

    certValue = certValue.trimmed();

    if (certValue.isEmpty())
    {
        QMessageBox::critical(
            this,
            tr("Empty certificate value"),
            tr("Please enter a non-empty certificate value."),
            QMessageBox::Ok
        );
        return;
    }

    QString rev = revision->text();
    QString key = signingKey->itemData(signingKey->currentIndex()).toString();

    if (!MonotoneUtil::createCert(database, rev, certName, certValue, key))
    {
        QMessageBox::critical(
            this,
            tr("Cert creation failed"),
            tr("Could not create certificate. Please check the logs for details."),
            QMessageBox::Ok
        );
        return;
    }

    emit certCreated(rev, certName, certValue, key);
    done(1);
}

