/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DatabaseDialogManager.h"
#include "vocab.h"

DatabaseDialogManager::DatabaseDialogManager(QWidget * parent)
  : DialogManager(parent), changesetBrowser(0),
    checkoutRevision(0), fileDiff(0), fileHistory(0), netsync(0),
    revisionDiff(0), revisionManifest(0), selectRevision(0),
    newProjectSetup(0), annotateFile(0), createCert(0)
{}

DatabaseDialogManager::~DatabaseDialogManager()
{
    if (changesetBrowser)   delete changesetBrowser;
    if (checkoutRevision)   delete checkoutRevision;
    if (fileDiff)           delete fileDiff;
    if (fileHistory)        delete fileHistory;
    if (netsync)            delete netsync;
    if (revisionDiff)       delete revisionDiff;
    if (revisionManifest)   delete revisionManifest;
    if (selectRevision)     delete selectRevision;
    if (newProjectSetup)    delete newProjectSetup;
    if (annotateFile)       delete annotateFile;
    if (createCert)         delete createCert;
}

void DatabaseDialogManager::closeAllDialogs()
{
    if (changesetBrowser)   changesetBrowser->close();
    if (checkoutRevision)   checkoutRevision->close();
    if (fileDiff)           fileDiff->close();
    if (fileHistory)        fileHistory->close();
    if (generateKeypair)    generateKeypair->close();
    if (netsync)            netsync->close();
    if (keyManagement)      keyManagement->close();
    if (revisionDiff)       revisionDiff->close();
    if (revisionManifest)   revisionManifest->close();
    if (selectRevision)     selectRevision->close();
    if (newProjectSetup)    newProjectSetup->close();
    if (annotateFile)       annotateFile->close();
    if (createCert)         createCert->close();

    DialogManager::closeAllDialogs();
}

void DatabaseDialogManager::showChangesetBrowser()
{
    if (!changesetBrowser)
    {
        changesetBrowser = new ChangesetBrowser(parentWidget(), monotoneHandle);

        connect(
            changesetBrowser, SIGNAL(certCreated(const QString &, const QString &,
                                                const QString &, const QString &)),
            this, SIGNAL(certCreated(const QString &, const QString &,
                                     const QString &, const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(revisionManifest(const QString &)),
            this, SLOT(showRevisionManifest(const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(diffRevision(const QString &, const QString &, const QString &)),
            this,  SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(diffFile(const QString &, const QString &, const QString &)),
            this,  SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(annotateFile(const QString &, const QString &)),
            this, SLOT(showAnnotation(const QString &, const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(createCert(const QString &, const QString &)),
            this, SLOT(showCreateCert(const QString &, const QString &))
        );

        connect(
            this, SIGNAL(certCreated(const QString &, const QString &,
                                     const QString &, const QString &)),
            changesetBrowser, SLOT(handleCertCreation(const QString &, const QString &,
                                                      const QString &, const QString &))
        );
    }

    changesetBrowser->init();

    showDialog(changesetBrowser);
}

void DatabaseDialogManager::showCheckoutRevision()
{
    if (!checkoutRevision)
    {
        checkoutRevision = new CheckoutRevision(parentWidget(), monotoneHandle);

        connect(
            checkoutRevision, SIGNAL(selectRevisionFromSelector(const QString &)),
            this, SLOT(showSelectRevision(const QString &))
        );

        // delegate this further
        connect(
            checkoutRevision, SIGNAL(revisionCheckedOut(const MonotoneHandlePtr &)),
            this, SIGNAL(revisionCheckedOut(const MonotoneHandlePtr &))
        );

        connect(
            this, SIGNAL(revisionSelected(const QString &)),
            checkoutRevision, SLOT(setSelectedRevision(const QString &))
        );
    }

    showDialog(checkoutRevision);
}

void DatabaseDialogManager::showFileDiff(const QString & file,
                                         const QString & base,
                                         const QString & target)
{
    if (!fileDiff)
    {
        fileDiff = new FileDiff(parentWidget());
    }

    fileDiff->init(monotoneHandle, file, base, target);
    showDialog(fileDiff);
}

void DatabaseDialogManager::showFileHistory(const QString & file,
                                            const QString & startRevision)
{
    if (!fileHistory)
    {
        fileHistory = new FileHistory(parentWidget(), monotoneHandle);

        connect(
            fileHistory, SIGNAL(fileDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    fileHistory->readHistory(file, startRevision);
    showDialog(fileHistory);
}

void DatabaseDialogManager::showNetsync()
{
    if (!netsync)
    {
        netsync  = new Netsync(parentWidget(), monotoneHandle);
    }

    netsync->init();
    showDialog(netsync);
}

void DatabaseDialogManager::showRevisionDiff(const QString & path,
                                             const QString & base,
                                             const QString & target)
{
    if (!revisionDiff)
    {
        revisionDiff = new RevisionDiff(parentWidget());

        connect(
            revisionDiff, SIGNAL(fileDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    revisionDiff->init(monotoneHandle, path, base, target);
    showDialog(revisionDiff);
}

void DatabaseDialogManager::showRevisionManifest(const QString & revision)
{
    if (!revisionManifest)
    {
        revisionManifest    = new RevisionManifest(parentWidget(), monotoneHandle);

        connect(
            revisionManifest, SIGNAL(fileHistory(const QString &, const QString &)),
            this, SLOT(showFileHistory(const QString &, const QString &))
        );

        connect(
            revisionManifest, SIGNAL(annotateFile(const QString &, const QString &)),
            this, SLOT(showAnnotation(const QString &, const QString &))
        );
    }

    revisionManifest->readManifest(revision);
    showDialog(revisionManifest);
}

void DatabaseDialogManager::showSelectRevision()
{
    showSelectRevision(QString());
}

void DatabaseDialogManager::showSelectRevision(const QString & selector)
{
    if (!selectRevision)
    {
        selectRevision = new SelectRevision(parentWidget(), monotoneHandle);

        connect(
            selectRevision, SIGNAL(revisionDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
        );

        connect(
            selectRevision, SIGNAL(revisionManifest(const QString &)),
            this, SLOT(showRevisionManifest(const QString &))
        );

        connect(
            selectRevision, SIGNAL(revisionSelected(const QString &)),
            this, SIGNAL(revisionSelected(const QString &))
        );
    }

    if (selector.size() > 0)
    {
        selectRevision->queryRevisions(selector);
    }

    showDialog(selectRevision);
}

void DatabaseDialogManager::showNewProjectSetup()
{
    if (!newProjectSetup)
    {
       newProjectSetup = new NewProjectSetup(parentWidget(), monotoneHandle);

       // delegate this further
       connect(
            newProjectSetup, SIGNAL(newProjectCreated(const MonotoneHandlePtr &)),
            this, SIGNAL(newProjectCreated(const MonotoneHandlePtr &))
        );
    }

    newProjectSetup->init();
    showDialog(newProjectSetup);
}

void DatabaseDialogManager::showAnnotation(const QString & rev,
                                           const QString & file)
{
    if (!annotateFile)
    {
        annotateFile = new AnnotateFile(parentWidget(), monotoneHandle);

        connect(
            annotateFile, SIGNAL(selectRevision()),
            this, SLOT(showSelectRevision())
        );

        connect(
            this, SIGNAL(revisionSelected(const QString &)),
            annotateFile, SLOT(setSelectedRevision(const QString &))
        );

        connect(
            annotateFile, SIGNAL(revisionManifest(const QString &)),
            this, SLOT(showRevisionManifest(const QString &))
        );

        connect(
            annotateFile, SIGNAL(revisionDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
        );

        connect(
            annotateFile, SIGNAL(fileDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    annotateFile->read(rev, file);
    showDialog(annotateFile);
}

void DatabaseDialogManager::showCreateCert(const QString & rev,
                                           const QString & certName)
{
    if (!createCert)
    {
        createCert = new CreateCert(parentWidget(), monotoneHandle);

        connect(
            createCert, SIGNAL(selectRevision()),
            this, SLOT(showSelectRevision())
        );

        connect(
            this, SIGNAL(revisionSelected(const QString &)),
            createCert, SLOT(setSelectedRevision(const QString &))
        );

        connect(
            createCert, SIGNAL(certCreated(const QString &, const QString &,
                                           const QString &, const QString &)),
            this, SIGNAL(certCreated(const QString &, const QString &,
                                     const QString &, const QString &))
        );
    }

    createCert->init(rev, certName);
    showDialog(createCert);
}

