/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DialogManager.h"
#include "vocab.h"
#include "FileDialog.h"
#include "ConnectServerDialog.h"

DialogManager::DialogManager(QWidget * parentWidget)
  : QObject(parentWidget), about(0), preferences(0),
    createDatabase(0), keyManagement(0), generateKeypair(0),
    changeKeyPassword(0), sourcePicker(0)
{
    // use an empty handle by default
    monotoneHandle = MonotoneHandle::create();
}

DialogManager::~DialogManager()
{
    if (about)             delete about;
    if (preferences)       delete preferences;
    if (createDatabase)    delete createDatabase;
    if (keyManagement)     delete keyManagement;
    if (generateKeypair)   delete generateKeypair;
    if (changeKeyPassword) delete changeKeyPassword;
    if (sourcePicker)      delete sourcePicker;
}

QWidget * DialogManager::parentWidget() const
{
    return reinterpret_cast<QWidget *>(parent());
}

void DialogManager::init(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
}

void DialogManager::closeAllDialogs()
{
    if (about)           about->close();
    if (preferences)     preferences->close();
    if (createDatabase)  createDatabase->close();
    if (keyManagement)   keyManagement->close();
    if (generateKeypair) generateKeypair->close();
    if (sourcePicker)    sourcePicker->close();
    emit allDialogsClosed();
}

void DialogManager::showAbout()
{
    if (!about)
    {
        about = new About(parentWidget());
    }
    showDialog(about);
}

void DialogManager::showPreferences()
{
    if (!preferences)
    {
        preferences = new Preferences(parentWidget());
    }
    showDialog(preferences);
}

void DialogManager::showCreateDatabase()
{
    if (!createDatabase)
    {
        createDatabase = new CreateDatabase(parentWidget());

        connect(
            createDatabase, SIGNAL(databaseCreated(const MonotoneHandlePtr &)),
            this, SIGNAL(sourceSelected(const MonotoneHandlePtr &))
        );
    }
    showDialog(createDatabase);
}

void DialogManager::showKeyManagement()
{
    if (!keyManagement)
    {
        keyManagement = new KeyManagement(parentWidget(), monotoneHandle);

        connect(
            keyManagement, SIGNAL(generateKeypair()),
            this, SLOT(showGenerateKeypair())
        );

        connect(
            keyManagement, SIGNAL(changeKeyPassword(const QString &)),
            this, SLOT(showChangeKeyPassword(const QString &))
        );

        connect(
            this, SIGNAL(keyPasswordChanged(const QString &)),
            keyManagement, SLOT(readKeys())
        );

        connect(
            this, SIGNAL(keypairGenerated(const QString &)),
            keyManagement, SLOT(readKeys())
        );
    }

    keyManagement->readKeys();
    showDialog(keyManagement);
}

void DialogManager::showGenerateKeypair()
{
    if (!generateKeypair)
    {
        generateKeypair  = new GenerateKeypair(parentWidget(), monotoneHandle);

        connect(
            generateKeypair, SIGNAL(keypairGenerated(const QString &)),
            this, SIGNAL(keypairGenerated(const QString &))
        );
    }

    showDialog(generateKeypair);
}

void DialogManager::showChangeKeyPassword(const QString & keyHash)
{
    if (!changeKeyPassword)
    {
        changeKeyPassword = new ChangeKeyPassword(parentWidget());

        connect(
            changeKeyPassword, SIGNAL(keyPasswordChanged(const QString &)),
            this, SIGNAL(keyPasswordChanged(const QString &))
        );
    }

    changeKeyPassword->init(keyHash);
    showDialog(changeKeyPassword);
}

void DialogManager::showSourcePicker(const MonotoneHandle::Type & type)
{
    if (sourcePicker)
    {
        delete sourcePicker;
    }

    if (type == MonotoneHandle::server_handle)
    {
        sourcePicker = new ConnectServerDialog(parentWidget());
    }
    else
    {
        sourcePicker = new FileDialog(parentWidget(), 0, type);
    }

    connect(
        sourcePicker, SIGNAL(sourceSelected(const MonotoneHandlePtr &)),
        this, SIGNAL(sourceSelected(const MonotoneHandlePtr &))
    );

    showDialog(sourcePicker);
}

void DialogManager::showDialog(QDialog * dlg)
{
    if (!openDialogs.contains(dlg))
    {
        connect(
            dlg, SIGNAL(finished(int)),
            this, SLOT(dialogFinished())
        );
        openDialogs.push(dlg);
    }

    dlg->raise();
    dlg->activateWindow();
    dlg->show();
}

void DialogManager::dialogFinished()
{
    QDialog * dlg = qobject_cast<QDialog *>(sender());
    I(openDialogs.contains(dlg));

    disconnect(
        dlg, SIGNAL(finished(int)),
        this, SLOT(dialogFinished())
    );

    // this is the simple case, the closed dialog is on top of the stack
    if (openDialogs.top() == dlg)
    {
        openDialogs.pop();
    }
    else
    {
        // Now apparently the user accepted / rejected a window from which he
        // navigated to other windows. Since all of these are not modal, but
        // show dependent information, we close them one after another.
        // Since this triggers this slot again, we need to ensure before
        // that this code path is aborted.
        // FIXME: not sure this is really what users want, but we're cleaning
        // up here for them... in case a user works really in parallel, this
        // could have weird side effects, like closing dialogs which are not
        // supposed to be closed, just because they're created after the main
        // dialog...
        while (true)
        {
            QDialog * other = openDialogs.pop();
            disconnect(
                other, SIGNAL(finished(int)),
                this, SLOT(dialogFinished())
            );
            other->reject();
            if (other == dlg) break;
        }
    }

    // is there anything to raise?
    if (!openDialogs.isEmpty())
    {
        QDialog * other = openDialogs.top();
        other->raise();
        other->activateWindow();
        return;
    }

    emit allDialogsClosed();
}

