/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "FileDialog.h"
#include "MonotoneUtil.h"

#include <QUrl>
#include <QGridLayout>
#include <QToolButton>
#include <QDialogButtonBox>
#include <QPushButton>

FileDialog::FileDialog(QWidget * parent, Qt::WindowFlags flags, MonotoneHandle::Type type)
  : QFileDialog(parent, flags), expectedType(type)
{
    setOptions(DontUseNativeDialog);

    if (expectedType == MonotoneHandle::workspace_handle)
    {
        setWindowTitle(tr("Open a workspace"));
        setLabelText(QFileDialog::FileName, QObject::tr("Workspace path:"));
        setFileMode(QFileDialog::Directory);
        setOptions(options() | ShowDirsOnly);
        setFilter(QDir::Drives | QDir::AllDirs | QDir::NoDotAndDotDot);
    }
    else if (expectedType == MonotoneHandle::database_handle)
    {
        setWindowTitle(tr("Open a database"));
        setLabelText(QFileDialog::FileName, QObject::tr("Database file:"));
        setNameFilter("*.db *.mtn");

        QList<QUrl> urls = sidebarUrls();
        QStringList defaultDatabaseLocations =
            MonotoneUtil::getDefaultDatabaseLocations(MonotoneHandle::create());
        foreach (const QString & location, defaultDatabaseLocations)
        {
            QUrl url = QUrl::fromLocalFile(location);
            if (!urls.contains(url))
                urls.push_back(url);
        }
        setSidebarUrls(urls);
    }
    else
        I(false);

    // FIXME: all this, of course, highly depends on the Qt 4.x layout
    // of the file dialog and will probably break in any other version
    QGridLayout * currentLayout = qobject_cast<QGridLayout*>(layout());

    // hide the filter selection and label
    currentLayout->itemAtPosition(3, 0)->widget()->hide();
    currentLayout->itemAtPosition(3, 1)->widget()->hide();

    // fix the button box layout and orientation
    QDialogButtonBox * box =
        qobject_cast<QDialogButtonBox*>(currentLayout->itemAtPosition(2,2)->widget());
    box->setOrientation(Qt::Horizontal);
    box->setLayoutDirection(Qt::RightToLeft);
    // remove and re-add without vertical span
    int buttonPos = currentLayout->indexOf(box);
    currentLayout->takeAt(buttonPos);
    currentLayout->addWidget(box, 2, 2);

    // add the validation message label
    validationMessage = new SizeableLabel(this);
    validationMessage->setStyleSheet(
         "SizeableLabel { qproperty-relativeSize: 0.85; color: red; }"
    );
    currentLayout->addWidget(validationMessage, 3, 1);
    validationMessage->hide();

    // remove the ugly border of the tool buttons in the upper area
    QLayout * btnLayout = currentLayout->itemAtPosition(0, 1)->layout();
    for (int i=1; i<=6; i++)
    {
        QToolButton * btn = qobject_cast<QToolButton*>(btnLayout->itemAt(i)->widget());
        btn->setStyleSheet("QToolButton { border: none }");
        btn->setAutoRaise(true);
    }

    setLayout(currentLayout);

    connect(
        this, SIGNAL(currentChanged(const QString &)),
        this, SLOT(slotCurrentChanged(const QString &))
    );
}

void FileDialog::slotCurrentChanged(const QString & path)
{
    QGridLayout * currentLayout = qobject_cast<QGridLayout*>(layout());
    QDialogButtonBox * box =
        qobject_cast<QDialogButtonBox*>(currentLayout->itemAtPosition(2,2)->widget());

    try
    {
        lastValidatedHandle.clear();
        lastValidatedHandle = MonotoneHandle::create(path);

        if (lastValidatedHandle->getType() != expectedType)
        {
            lastValidatedHandle.clear();
            throw GuitoneException("wrong or empty source picked");
        }
    }
    catch (GuitoneException & e)
    {
        validationMessage->setText(e.what());
        validationMessage->show();
        box->button(QDialogButtonBox::Open)->setEnabled(false);
        return;
    }

    validationMessage->hide();
    box->button(QDialogButtonBox::Open)->setEnabled(true);
}

void FileDialog::accept()
{
    if (lastValidatedHandle)
    {
        emit sourceSelected(lastValidatedHandle);
        done(0);
    }
}

