/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "FileDiff.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QHeaderView>
#include <QMessageBox>
#include <QScrollBar>
#include <QTextCodec>

FileDiff::FileDiff(QWidget * parent)
  : Dialog(parent), diffModel(0), fileModel(0), fileProxyModel(0)
{
    setupUi(this);
    Dialog::init();

    // setup correct header widths
    diffView->header()->setDefaultSectionSize(40);
    diffView->header()->setResizeMode(0, QHeaderView::Interactive);
    diffView->header()->setResizeMode(1, QHeaderView::Interactive);
    diffView->header()->setResizeMode(2, QHeaderView::ResizeToContents);

    QScrollBar * vertBar = diffView->verticalScrollBar();
    connect(
        vertBar, SIGNAL(valueChanged(int)),
        this, SLOT(changeCurrentIndex(int))
    );

    connect(
        firstRevision, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );
    connect(
        secondRevision, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );
    connect(
        bothRevisions, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );

    connect(
        scrollToNext, SIGNAL(clicked()),
        this, SLOT(getNextGroup())
    );

    connect(
        scrollToPrev, SIGNAL(clicked()),
        this, SLOT(getPrevGroup())
    );

    fileModel = new GetFile(this);
    fileProxyModel = new GetFileProxyModel(this);
    fileProxyModel->setSourceModel(fileModel);

    diffModel = new ContentDiff(this);

    diffView->setModel(fileProxyModel);
    diffStatusView->setModel(fileProxyModel);

    connect(
        fileModel, SIGNAL(fileRead()),
        this, SLOT(fileRead())
    );

    connect(
        diffModel, SIGNAL(diffRead()),
        this, SLOT(diffRead())
    );

    ByteArrayList codecNamesAndAliases = QTextCodec::availableCodecs();
    QSet<QByteArray> codecNameSet;

    // eliminate aliases
    foreach (const QByteArray & codec, codecNamesAndAliases)
    {
        codecNameSet.insert(QTextCodec::codecForName(codec)->name());
    }

    ByteArrayList codecNames = codecNameSet.toList();
    qSort(codecNames);

    foreach (const QByteArray & codec, codecNames)
    {
        //! FIXME: better ordering, i.e. popular ones on top
        textEncoding->addItem(QString(codec));
    }

    connect(
        textEncoding, SIGNAL(currentIndexChanged(const QString &)),
        this, SLOT(switchOutputEncoding(const QString &))
    );
}

void FileDiff::init(const MonotoneHandlePtr & handle, const QString & file,
                    const QString & base, const QString & target)
{
    fileName = file;
    monotoneHandle = handle;
    QString encoding;

    QPair<QString, QString> attrval;
    bool success = MonotoneUtil::getFileAttribute(
        handle, base, file, "guitone:file-encoding", attrval
    );

    if (!success || attrval.second == "dropped")
    {
        L(QString("could not query file encoding of %1, using default").arg(file));
        encoding = "UTF-8";
    }
    else
    {
        encoding = attrval.first;
    }

    textEncoding->blockSignals(true);
    textEncoding->setCurrentIndex(textEncoding->findText(encoding));
    textEncoding->blockSignals(false);

    setWindowTitle(tr("Loading file differences"));

    diffModel->readDiff(monotoneHandle, fileName, base, target, encoding);
}

FileDiff::~FileDiff()
{
    if (fileModel)      delete fileModel;
    if (fileProxyModel) delete fileProxyModel;
    if (diffModel)      delete diffModel;
}

void FileDiff::diffRead()
{
    Diff * diff = diffModel->getDiff(fileName);
    if (diff == 0)
    {
        QMessageBox::information(
            this,
            tr("File has not changed"),
            tr("The file has not been changed between these revisions."),
            QMessageBox::Ok
        );
        reject();
        return;
    }

    if (diff->is_binary)
    {
        QMessageBox::information(
            this,
            tr("File is binary"),
            tr("The file is binary and cannot be diffed."),
            QMessageBox::Ok
        );
        reject();
        return;
    }

    QString base = diffModel->getBase();
    if (base.isEmpty())
    {
        I(monotoneHandle->getType() == MonotoneHandle::workspace_handle);
        base = MonotoneUtil::getBaseWorkspaceRevision(monotoneHandle);
    }

    fileModel->readFileByName(monotoneHandle, fileName,
                              base, textEncoding->currentText());
}

void FileDiff::fileRead()
{
    setWindowTitle(tr("File differences of \"%1\"").arg(fileName));

    QString left = tr("workspace parent"), right = tr("workspace revision");
    QString base = diffModel->getBase(), target = diffModel->getTarget();

    if (!base.isEmpty())
    {
        left = base.left(12).append("...");
    }

    if (!target.isEmpty())
    {
        right = target.left(12).append("...");
    }

    firstRevision->setText(tr("only left (%1)").arg(left));
    secondRevision->setText(tr("only right (%1)").arg(right));

    applyDiff();
}

void FileDiff::applyDiff()
{
    Diff * diff = diffModel->getDiff(fileName);
    fileModel->applyDiff(diff);
    diffStatusView->update();
}

void FileDiff::versionToggled()
{
    if (firstRevision->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Left);
        diffView->header()->showSection(0);
        diffView->header()->hideSection(1);
    }
    else if (secondRevision->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Right);
        diffView->header()->showSection(1);
        diffView->header()->hideSection(0);
    }
    else if (bothRevisions->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Both);
        diffView->header()->showSection(0);
        diffView->header()->showSection(1);
    }

    diffStatusView->update();
}

void FileDiff::scrollToGroup(bool forward)
{
    if (!currentIndex.isValid()) return;
    QModelIndex proxyIndex = fileProxyModel->mapFromSource(currentIndex);

    // recursion: if the queried group is not valid in the view because it
    // has been filtered out, get the next group and try again
    if (!proxyIndex.isValid())
    {
        if (forward)
        {
            getNextGroup();
            return;
        }
        getPrevGroup();
        return;
    }

    diffView->scrollTo(proxyIndex, QAbstractItemView::PositionAtTop);
}

void FileDiff::getPrevGroup()
{
    currentIndex = fileModel->getPrevGroup(currentIndex);
    scrollToGroup(false);
}

void FileDiff::getNextGroup()
{
    currentIndex = fileModel->getNextGroup(currentIndex);
    scrollToGroup(true);
}

void FileDiff::changeCurrentIndex(int pos)
{
    if (!fileProxyModel) return;

    QModelIndex newCurrent = fileProxyModel->mapToSource(
         fileProxyModel->index(pos, 0, QModelIndex())
    );

    if (newCurrent.isValid())
    {
        currentIndex = newCurrent;
    }
}

void FileDiff::switchOutputEncoding(const QString & encoding)
{
    fileModel->switchOutputEncoding(encoding);
    diffModel->switchOutputEncoding(encoding);

    // the model data are converted instantly, so we don't have
    // to wait for signals this time...
    applyDiff();

    if (monotoneHandle->getType() == MonotoneHandle::workspace_handle)
    {
        MonotoneTaskPtr task(new MonotoneTask(
            QStringList() << "set_attribute"
                          << fileName
                          << "guitone:file-encoding"
                          << encoding
        ));

        MonotoneUtil::runSynchronousTask(monotoneHandle, task);

        if (task->getReturnCode() != 0)
        {
            W(QString("couldn't set guitone:file-encoding on %1: %2")
                .arg(fileName).arg(task->getDecodedOutput()));
        }
    }
}

