/***************************************************************************
 *   Copyright (C) 2006 by Ingo Maindorfer                                 *
 *   ingo@liquidcooling.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GenerateKeypair.h"
#include "MonotoneUtil.h"
#include "Settings.h"
#include "BasicIOParser.h"

#include <QMessageBox>
#include <QStringList>

GenerateKeypair::GenerateKeypair(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();
}

GenerateKeypair::~GenerateKeypair() {}

void GenerateKeypair::accept()
{
    if (lineKeyName->text().size() == 0 ||
        lineKeyPasswd->text().size() == 0 ||
        lineKeyPasswd2->text().size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Missing information"),
            tr("Please fill out all fields."),
            QMessageBox::Ok
        );
        return;
    }

    if (lineKeyPasswd->text().compare(lineKeyPasswd2->text()) != 0)
    {
        QMessageBox::information(
            this,
            tr("Password mismatch"),
            tr("The entered passwords did not match."),
            QMessageBox::Ok
        );
        return;
    }

    QStringList options;
    if (allowDuplicateKey->isChecked())
    {
        options << "force-duplicate-key" << "1";
    }

    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "generate_key"
                      << lineKeyName->text()
                      << lineKeyPasswd->text(),
        options
    ));
    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    if (!task->isFinished())
    {
        C(QString("key generation was aborted: %1").arg(task->getDecodedOutput()));
        return;
    }

    if (task->getReturnCode() > 0)
    {
        QMessageBox::critical(
             this,
             tr("Error creating keypair"),
             tr("There was an error creating the keypair:\n%1")
                .arg(task->getLast(MonotoneTask::Error)),
             QMessageBox::Ok, 0, 0
        );
        return;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1);
    Stanza st = stanzas.at(0);
    I(st.size() > 1);
    I(st.at(1).sym == "hash" && st.at(1).vals.size() == 1);

    QString keyHash = st.at(1).vals.at(0);

    if (rememberPassword->isChecked())
    {
        Settings::addItemToMap(
           "KeyPasswords",
           keyHash,
           lineKeyPasswd->text()
        );

        QMessageBox::information(
             this,
             tr("Password saved"),
             tr("Your key password has been saved. Please note that you "
                "need to re-open the database or workspace you want to "
                "use this key with."),
             QMessageBox::Ok
        );
    }

    emit keypairGenerated(keyHash);
    done(QDialog::Accepted);
}

