/***************************************************************************
 *   Copyright (C) 2006 by Ingo Maindorfer                                 *
 *   ingo@liquidcooling.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "KeyManagement.h"
#include "GenerateKeypair.h"
#include "MonotoneProcess.h"
#include "Settings.h"

#include <QMessageBox>
#include <QApplication>
#include <QClipboard>

KeyManagement::KeyManagement(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();

    model = new Keys(this, monotoneHandle);
    proxyModel = new QSortFilterProxyModel(this);
    proxyModel->setSourceModel(model);
    proxyModel->setFilterCaseSensitivity(Qt::CaseInsensitive);
    proxyModel->setFilterKeyColumn(-1);

    keyList->setModel(proxyModel);
    keyList->setSortingEnabled(true);

    connect(
        generateKey, SIGNAL(clicked()),
        this, SIGNAL(generateKeypair())
    );

    connect(
        keyList, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuEvent(const QModelIndexList &, const QPoint &))
    );

    connect(
        filterField, SIGNAL(textChanged(const QString &)),
        proxyModel, SLOT(setFilterWildcard(const QString &))
    );

    actCopyHash = new QAction(tr("Copy key hash to clipboard"), this);
    connect(
        actCopyHash, SIGNAL(triggered()),
        this, SLOT(copyKeyHashToClipboard())
    );

    actCopyGivenName = new QAction(tr("Copy given name to clipboard"), this);
    connect(
        actCopyGivenName, SIGNAL(triggered()),
        this, SLOT(copyGivenNameToClipboard())
    );

    actCopyLocalName = new QAction(tr("Copy local name to clipboard"), this);
    connect(
        actCopyLocalName, SIGNAL(triggered()),
        this, SLOT(copyLocalNameToClipboard())
    );

    actCopyPublicKey = new QAction(tr("Copy public key data to clipboard"), this);
    connect(
        actCopyPublicKey, SIGNAL(triggered()),
        this, SLOT(copyPubkeyDataToClipboard())
    );

    actForgetPassword = new QAction(tr("Forget remembered password"), this);
    connect(
        actForgetPassword, SIGNAL(triggered()),
        this, SLOT(forgetPassword())
    );

    actChangePassword = new QAction(tr("Change password"), this);
    connect(
        actChangePassword, SIGNAL(triggered()),
        this, SLOT(changePassword())
    );

    actDrop = new QAction(tr("Drop key"), this);
    connect(
        actDrop, SIGNAL(triggered()),
        this, SLOT(dropKey())
    );
}

KeyManagement::~KeyManagement()
{
    delete proxyModel;
    delete model;

    delete actCopyHash;
    delete actCopyGivenName;
    delete actCopyLocalName;
    delete actCopyPublicKey;
    delete actForgetPassword;
    delete actChangePassword;
    delete actDrop;
}

void KeyManagement::readKeys()
{
    model->readKeys();
}

void KeyManagement::contextMenuEvent(const QModelIndexList & indexList, const QPoint & p)
{
    Q_UNUSED(indexList);

    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QMenu popupMenu;

    popupMenu.addAction(actCopyHash);
    popupMenu.addAction(actCopyGivenName);
    popupMenu.addAction(actCopyLocalName);

    popupMenu.addSeparator();
    popupMenu.addAction(actCopyPublicKey);

    if (key->isPrivate())
    {
        popupMenu.addSeparator();
        popupMenu.addAction(actChangePassword);
        if (!Settings::getItemFromMap("KeyPasswords", key->hash).isNull())
        {
            popupMenu.addAction(actForgetPassword);
        }
    }

    popupMenu.addSeparator();
    popupMenu.addAction(actDrop);

    popupMenu.exec(p);
}

Key * KeyManagement::getKeyFromSelection() const
{
    QItemSelectionModel * selection = keyList->selectionModel();
    QList<QModelIndex> indexList = selection->selectedIndexes();

    if (indexList.size() == 0)
        return 0;

    QModelIndex srcIndex = proxyModel->mapToSource(indexList.at(0));
    return static_cast<Key *>(srcIndex.internalPointer());
}

void KeyManagement::copyKeyHashToClipboard()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QClipboard * clipboard = QApplication::clipboard();
    clipboard->setText(key->hash);
}

void KeyManagement::copyGivenNameToClipboard()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QClipboard * clipboard = QApplication::clipboard();
    clipboard->setText(key->given_name);
}

void KeyManagement::copyLocalNameToClipboard()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QClipboard * clipboard = QApplication::clipboard();
    clipboard->setText(key->local_name);
}

// TODO: should be converted to automate pubkey, if available
void KeyManagement::copyPubkeyDataToClipboard()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QClipboard * clipboard = QApplication::clipboard();

    QStringList args;
    QString output;
    args << "pubkey" << key->hash;

    if (!MonotoneProcess::singleRun(args, monotoneHandle, output))
    {
        QMessageBox::critical(
            this,
            tr("Couldn't query public key data"),
            tr("Failed to query public key data for key \"%1\".")
                .arg(key->hash),
            QMessageBox::Ok
         );
         return;
    }

    clipboard->setText(output);
}

void KeyManagement::forgetPassword()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    Settings::removeItemFromMap("KeyPasswords", key->hash);
    model->readKeys();

    QMessageBox::information(
         this,
         tr("Password removed"),
         tr("Your key password has been removed. Please note that "
            "currently opened workspaces and / or databases might "
            "still cache the password."),
         QMessageBox::Ok
    );
}

void KeyManagement::changePassword()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;
    emit changeKeyPassword(key->hash);
}

void KeyManagement::dropKey()
{
    Key * key = getKeyFromSelection();
    if (key == 0) return;

    QMessageBox::StandardButton btn = QMessageBox::question(
        this,
        tr("Drop key?"),
        tr("Do you really want to drop the key %1 (%2)?")
            .arg(key->given_name)
            .arg(key->hash),
        QMessageBox::Yes | QMessageBox::No
    );

    if (btn == QMessageBox::No) return;

    // ask one more time for private keys
    if (key->private_locations > 0)
    {
        QMessageBox::StandardButton btn = QMessageBox::question(
            this,
            tr("Drop private key?"),
            tr("WARNING: If you do this and you have no backup of your keypair, "
               "any particular trust setup (f.e. in a monotone server) "
               "will no longer recognize you.\n\n"
               "Do you really want to continue?"),
            QMessageBox::Yes | QMessageBox::No
         );

         if (btn == QMessageBox::No) return;
    }

    QStringList args;
    args << "dropkey" << key->hash;

    QString output;
    if (!MonotoneProcess::singleRun(args, monotoneHandle, output))
    {
        QMessageBox::critical(
            this,
            tr("Couldn't drop key"),
            tr("Failed to drop key \"%1\": %2")
                .arg(key->hash)
                .arg(output),
            QMessageBox::Ok
         );
         return;
    }

    Settings::removeItemFromMap("KeyPasswords", key->hash);

    model->readKeys();
}

