/***************************************************************************
 *   Copyright (C) 2009 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Netsync.h"
#include "Settings.h"
#include "MonotoneUtil.h"

#include <QMessageBox>

Netsync::Netsync(QWidget * parent, const MonotoneHandlePtr & handle)
    : Dialog(parent), monotoneHandle(handle), running(false)
{
    setupUi(this);
    Dialog::init();

    connect(
        startStopButton, SIGNAL(clicked()),
        this, SLOT(startStopClicked())
    );

    connect(
        this, SIGNAL(dialogHidden()),
        this, SLOT(stop())
    );
}

Netsync::~Netsync() {}

void Netsync::init()
{
    progressWidget->hide();

    QMap<QString, QString> defaults =
        MonotoneUtil::getDatabaseVariables(monotoneHandle, "database");

    QList<QString> knownServers =
        (MonotoneUtil::getDatabaseVariables(monotoneHandle, "known-servers")).keys();

    includePattern->setText(defaults.contains("default-include-pattern") ?
                            defaults["default-include-pattern"] : "*");

    excludePattern->setText(defaults.contains("default-exclude-pattern") ?
                            defaults["default-exclude-pattern"] : "");

    host->clear();
    host->addItems(knownServers);

    if (defaults.contains("default-server"))
    {
        if (!knownServers.contains(defaults["default-server"]))
        {
            host->addItem(defaults["default-server"]);
        }
        host->setCurrentIndex(host->findText(defaults["default-server"]));
    }

    QMap<QString, QString> privateKeys =
        MonotoneUtil::getPrivateKeyMap(monotoneHandle);

    keys->clear();
    keys->addItem(tr("<anonymous>"), QString());

    for (QMap<QString, QString>::const_iterator it = privateKeys.constBegin();
         it != privateKeys.constEnd(); it++)
    {
        keys->addItem(it.value(), it.key());
    }
}

void Netsync::start(Action act,
                    const QString & host,
                    const QString & key,
                    const QString & include,
                    const QString & exclude)
{
    QMap<int, QString> actions;
    actions[Pull] = "pull";
    actions[Push] = "push";
    actions[Sync] = "sync";

    QStringList cmd;
    cmd << actions[act] << host << include;

    QStringList opts;
    if (!exclude.isEmpty())
    {
        opts << "exclude" << exclude;
    }

    if (!key.isEmpty())
    {
        opts << "key" << key;
    }

    MonotoneTaskPtr task(new MonotoneTask(cmd, opts));
    AutomateCommand::enqueueTask(monotoneHandle, task);

    progressText->setStyleSheet("QLabel { color: palette(text); }");
    progressText->setText(tr("Connecting to %1...").arg(host));
    startStopButton->setText(tr("Stop"));
    settingsGroup->setEnabled(false);
    actionsGroup->setEnabled(false);
    overallProgress->setMaximum(0);
    progressWidget->show();
    running = true;
}

void Netsync::stop()
{
    if (running)
        AutomateCommand::abortThreads();
}

void Netsync::startStopClicked()
{
    if (running)
    {
        stop();
        return;
    }

    Action act;
    QStringList cmd;
    if (actionPull->isChecked())
        act = Pull;
    else if (actionPush->isChecked())
        act = Push;
    else if (actionSync->isChecked())
        act = Sync;
    else
        I(false);

    QString include("*");
    if (!includePattern->text().isEmpty())
    {
        include = includePattern->text();
    }

    QString exclude;
    if (!excludePattern->text().isEmpty())
    {
        exclude = excludePattern->text();
    }

    QString key = keys->itemData(keys->currentIndex()).toString();

    start(act, host->currentText(), key, include, exclude);
}

void Netsync::processTaskResult(const MonotoneTaskPtr & task)
{
    running = false;
    startStopButton->setText(tr("Start"));
    settingsGroup->setEnabled(true);
    actionsGroup->setEnabled(true);

    overallProgress->setMaximum(1);
    overallProgress->setValue(1);

    if (task->getReturnCode() != 0)
    {
        progressText->setStyleSheet("QLabel { color: red; }");

        QRegExp re1("failed to decrypt old private RSA key");
        if (task->outOfBandMessagesMatch(MonotoneTask::Error, re1))
        {
            progressText->setText(
                tr("Could not decode private key, please check your settings.")
            );
            return;
        }

        QRegExp re2("network error: (.+)");
        if (task->outOfBandMessagesMatch(MonotoneTask::Error, re2))
        {
            progressText->setText(
                tr("An network error has occurred: %1").arg(re2.cap(1))
            );
            return;
        }

        QRegExp re3("denied '(.+)' write permission for '(.+)'");
        if (task->outOfBandMessagesMatch(MonotoneTask::Warning, re3))
        {
            progressText->setText(
                tr("Key '%1' has no write permission for '%2'.")
                    .arg(re3.cap(1)).arg(re3.cap(2))
            );
            return;
        }

        QRegExp re4("remote public key hash '(.+)' is unknown");
        if (task->outOfBandMessagesMatch(MonotoneTask::Warning, re4))
        {
            progressText->setText(
                tr("Key '%1' is unknown to server.").arg(re4.cap(1))
            );
            return;
        }

        QRegExp re5("denied '(.+)' read permission for '(.+)' .+ because of branch '(.+)'");
        if (task->outOfBandMessagesMatch(MonotoneTask::Warning, re5))
        {
            progressText->setText(
                tr("Key '%1' has no read permission for '%2' because of branch '%3'.")
                  .arg(re5.cap(1)).arg(re5.cap(2)).arg(re5.cap(3))
            );
            return;
        }

        progressText->setText(
            tr("An unknown error occurred, please check the logs for details.")
        );
        return;
    }

    progressText->setStyleSheet("QLabel { color: green; }");
    progressText->setText(tr("Successful exchange with %1")
                              .arg(QString(task->getArguments().at(1))));
}

void Netsync::processTaskAbortion(const MonotoneTaskPtr & task)
{
    Q_UNUSED(task);

    progressWidget->hide();
    startStopButton->setText(tr("Start"));
    settingsGroup->setEnabled(true);
    actionsGroup->setEnabled(true);
    running = false;
}

void Netsync::tickerUpdate(const TickerMap & tickers)
{
    // we "compress" the values of all found tickers into one progress bar
    // this also means that we cannot display an actual progress unless
    // we know the total values of all tickers
    bool knowTotalValue = true;
    int progress = 0, total = 0;
    QStringList progressTexts;
    foreach (char ident, tickers.keys())
    {
        if (tickers[ident].complete)
            continue;

        // we never get total values for the transferred bytes, so ignore
        // these completly for the progress bar calculation
        if (ident == '<' || ident == '>')
        {
            progressTexts.push_back(tickers[ident].toString(true));
        }
        else
        {
            if (tickers[ident].total == 0)
                knowTotalValue = false;

            progress += tickers[ident].progress;
            total    += tickers[ident].total;
            progressTexts.push_back(tickers[ident].toString(false));
        }
    }

    if (knowTotalValue)
    {
        overallProgress->setMaximum(total);
        overallProgress->setValue(progress);
    }
    else
    if (overallProgress->maximum() != 0)
    {
        overallProgress->setMaximum(0);
    }

    progressText->setText(progressTexts.join(", "));

    progressWidget->show();
}

