/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "OpenPrompt.h"
#include "Settings.h"
#include "MonotoneUtil.h"

#include <QTimer>
#include <QMessageBox>

class GuitoneStandalone;

OpenPrompt::OpenPrompt(QWidget * parent) : Dialog(parent)
{
    setupUi(this);
    Dialog::init();

    connect(
        btnOpenDatabase, SIGNAL(clicked()),
        this, SLOT(openDatabase())
    );

    connect(
        btnOpenWorkspace, SIGNAL(clicked()),
        this, SLOT(openWorkspace())
    );

    connect(
        btnNewConnection, SIGNAL(clicked()),
        this, SLOT(newServerConnection())
    );

    connect(
        btnExit, SIGNAL(clicked()),
        this, SIGNAL(quitApp())
    );

    connect(
        this, SIGNAL(rejected()),
        this, SIGNAL(quitApp())
    );

    connect(
        recentWorkspaces, SIGNAL(currentIndexChanged(int)),
        this, SLOT(recentWorkspacesChanged(int))
    );

    connect(
        recentDatabases, SIGNAL(currentIndexChanged(int)),
        this, SLOT(recentDatabasesChanged(int))
    );

    connect(
        recentServers, SIGNAL(currentIndexChanged(int)),
        this, SLOT(recentServersChanged(int))
    );

    dialogManager = new DialogManager(this);

    connect(
        btnCreateDatabase, SIGNAL(clicked()),
        dialogManager, SLOT(showCreateDatabase())
    );

    connect(
        dialogManager, SIGNAL(sourceSelected(const MonotoneHandlePtr &)),
        this, SIGNAL(loadFromHandle(const MonotoneHandlePtr &))
    );

    connect(
        btnAbout, SIGNAL(clicked()),
        dialogManager, SLOT(showAbout())
    );

    connect(
        btnPreferences, SIGNAL(clicked()),
        dialogManager, SLOT(showPreferences())
    );

    connect(
        btnKeys, SIGNAL(clicked()),
        dialogManager, SLOT(showKeyManagement())
    );

    connect(
        qApp, SIGNAL(updateRecentLists()),
        this, SLOT(updateRecentLists())
    );

    updateRecentLists();

    QTimer::singleShot(0, this, SLOT(checkForPrivateKeys()));
}

OpenPrompt::~OpenPrompt()
{
    delete dialogManager;
}

void OpenPrompt::openDatabase()
{
    dialogManager->showSourcePicker(MonotoneHandle::database_handle);
}

void OpenPrompt::openWorkspace()
{
    dialogManager->showSourcePicker(MonotoneHandle::workspace_handle);
}

void OpenPrompt::newServerConnection()
{
    dialogManager->showSourcePicker(MonotoneHandle::server_handle);
}

void OpenPrompt::recentWorkspacesChanged(int idx)
{
    // skip the first dummy item
    if (idx < 1) return;
    setEnabled(false);
    emit loadFromString(recentWorkspaces->itemText(idx));
}

void OpenPrompt::recentDatabasesChanged(int idx)
{
    // skip the first dummy item
    if (idx < 1) return;
    setEnabled(false);
    emit loadFromString(recentDatabases->itemText(idx));
}

void OpenPrompt::recentServersChanged(int idx)
{
    // skip the first dummy item
    if (idx < 1) return;
    setEnabled(false);
    emit loadFromString(recentServers->itemData(idx).toString());
}

void OpenPrompt::loadFromStringFailed(const QString & source, const QString & errmsg)
{
    QMessageBox::critical(
        NULL,
        tr("Failed to load source"),
        tr("The source '%1' could not be loaded:\n%2")
            .arg(source).arg(errmsg),
        QMessageBox::Ok
    );

    setEnabled(true);
    raise();
}

void OpenPrompt::updateRecentLists()
{
    recentWorkspaces->blockSignals(true);
    QStringList workspaces = Settings::getItemList("RecentWorkspaceList");

    if (workspaces.size() > 0)
    {
        recentWorkspaces->addItems(workspaces);
        recentWorkspaces->setEnabled(true);
    }
    else
    {
        recentWorkspaces->setEnabled(false);
    }
    recentWorkspaces->blockSignals(false);

    recentDatabases->blockSignals(true);
    QStringList databases = Settings::getItemList("RecentDatabaseList");

    if (databases.size() > 0)
    {
        recentDatabases->addItems(databases);
        recentDatabases->setEnabled(true);
    }
    else
    {
        recentDatabases->setEnabled(false);
    }
    recentDatabases->blockSignals(false);

    recentServers->blockSignals(true);
    QStringList servers = Settings::getItemList("RecentServerList");

    if (servers.size() > 0)
    {
        int c = 0;
        QRegExp rx("^mtn://(?:(\\w*)@)?([\\w\\-]+(?:\\.[\\w\\-]+)*(?::\\d+)?)$");
        foreach (const QString & server, servers)
        {
            QString label = server;
            if (rx.indexIn(label) != -1)
            {
                QString key(tr("<anonymous>"));
                if (!rx.cap(1).isEmpty())
                    key = rx.cap(1).left(8) + "...";
                label = tr("%1 (%2)").arg(rx.cap(2)).arg(key);
            }
            recentServers->insertItem(++c, label, server);
        }
        recentServers->setEnabled(true);
    }
    else
    {
        recentServers->setEnabled(false);
    }
    recentServers->blockSignals(false);
}

void OpenPrompt::checkForPrivateKeys()
{
    if (!Settings::getBool("PromptWhenPrivateKeyIsMissing"))
    {
        return;
    }

    QMap<QString, QString> privateKeys =
        MonotoneUtil::getPrivateKeyMap(MonotoneHandle::create());

    if (privateKeys.size() == 0)
    {
        QMessageBox msgBox(this);
        msgBox.setWindowFlags(Qt::Sheet);
        msgBox.setIcon(QMessageBox::Question);
        msgBox.setText("No private key found");
        msgBox.setInformativeText(
          "You do not have a monotone private key yet. A private key "
          "is needed to make commits and authenticate yourself towards "
          "other nodes.\nDo you want to create a private key now?");
        QPushButton * yesBtn   = msgBox.addButton(tr("Yes"),
                                                     QMessageBox::ActionRole);
        QPushButton * laterBtn = msgBox.addButton(tr("No, later"),
                                                     QMessageBox::ActionRole);
        QPushButton * neverBtn = msgBox.addButton(tr("No, never"),
                                                     QMessageBox::ActionRole);

        msgBox.exec();

        if (msgBox.clickedButton() == yesBtn)
        {
            dialogManager->showGenerateKeypair();
        }
        else
        if (msgBox.clickedButton() == neverBtn)
        {
            Settings::setBool("PromptWhenPrivateKeyIsMissing", false);
        }
        else
        if (msgBox.clickedButton() == laterBtn)
        {
            return;
        }
    }
}

