/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Preferences.h"
#include "MonotoneManager.h"
#include "Settings.h"
#include "DebugLog.h"
#include "GuitoneCore.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QHeaderView>

Preferences::Preferences(QWidget * parent) : Dialog(parent)
{
    setupUi(this);
    Dialog::init();

    // init the values before connecting the signals, otherwise
    // the user might be notified of something which we did
    // programmatically
    init();

    connect(
        selectMtnExecutable, SIGNAL(clicked()),
        this, SLOT(openFileBrowser())
    );

    connect(
        relaxedVersionCheck, SIGNAL(toggled(bool)),
        this, SLOT(warnAboutRelaxedVersionChecking(bool))
    );

    connect(
        resetSettings, SIGNAL(clicked()),
        this, SLOT(resetAllSettings())
    );
}

Preferences::~Preferences() {}

void Preferences::init()
{
    mtnExecutablePath->setText(Settings::getString("MtnExePath"));

    enableConsoleLog->setCheckState(
        Settings::getBool("ConsoleLogEnabled") ?
        Qt::Checked : Qt::Unchecked
    );

    enableFileLog->setCheckState(
        Settings::getBool("FileLogEnabled") ?
        Qt::Checked : Qt::Unchecked
    );

    enableFileLog->setText(tr("enable file logging (%1)").arg(DebugLog::logFilePath()));

    logLevel->clear();
    logLevel->addItem(tr("Very low (only fatal)"), 1);
    logLevel->addItem(tr("Low (critical)"), 2);
    logLevel->addItem(tr("Medium (warnings)"), 3);
    logLevel->addItem(tr("High (info messages)"), 4);

    int lvl = Settings::getLogLevel();
#ifndef QT_NO_DEBUG
    logLevel->addItem(tr("Very high (debug messages)"), 5);
#else
    lvl = lvl > 4 ? 4 : lvl;
#endif

    logLevel->setCurrentIndex(lvl - 1);

    checkForUpdates->setCheckState(
        Settings::getBool("CheckForUpdates") ?
        Qt::Checked : Qt::Unchecked
    );

    readWorkspaceIncrementally->setCheckState(
        Settings::getBool("ReadWorkspaceIncrementally") ?
        Qt::Checked : Qt::Unchecked
    );

    disableChangelogAutoCompletion->setCheckState(
        Settings::getBool("DisableChangelogAutoCompletion") ?
        Qt::Checked : Qt::Unchecked
    );

    fixUnwantedReverseDiffs->setCheckState(
        Settings::getBool("FixUnwantedReverseDiffs") ?
        Qt::Checked : Qt::Unchecked
    );

    saveEncodingAsFileAttribute->setCheckState(
        Settings::getBool("SaveEncodingAsFileAttribute") ?
        Qt::Checked : Qt::Unchecked
    );

    colorPickerOldestAnnotatedRevision->setSelectedColor(
        QColor(Settings::getString("AnnotationColorOldestRevision"))
    );

    colorPickerNewestAnnotatedRevision->setSelectedColor(
        QColor(Settings::getString("AnnotationColorNewestRevision"))
    );

    colorPickerHighlightAnnotatedRevision->setSelectedColor(
        QColor(Settings::getString("AnnotationColorHighlight"))
    );

    colorPickerAddedLines->setSelectedColor(
        QColor(Settings::getString("DiffColorAddedLines"))
    );

    colorPickerRemovedLines->setSelectedColor(
        QColor(Settings::getString("DiffColorRemovedLines"))
    );

    colorPickerSuspendedBranches->setSelectedColor(
        QColor(Settings::getString("ChangesetBrowserSuspendedBranches"))
    );

    colorPickerMergeRevisions->setSelectedColor(
        QColor(Settings::getString("ChangesetBrowserMergeRevisions"))
    );

    colorPickerOutOfBranchRevisions->setSelectedColor(
        QColor(Settings::getString("ChangesetBrowserOutOfBranchRevisions"))
    );

    relaxedVersionCheck->setCheckState(
        Settings::getBool("RelaxedVersionCheck") ?
        Qt::Checked : Qt::Unchecked
    );

    askFileOpen->setCheckState(
        Settings::getBool("AskFileOpen") ?
        Qt::Checked : Qt::Unchecked
    );

    askFileOpenExtensions->setText(
        Settings::getString("AskFileOpenExt")
    );

    readWorkspaceIncrementally->setCheckState(
        Settings::getBool("ReadWorkspaceIncrementally") ?
        Qt::Checked : Qt::Unchecked
    );

    readUserRcFiles->setCheckState(
        Settings::getBool("ReadUserRcFiles") ?
        Qt::Checked : Qt::Unchecked
    );
}

void Preferences::accept()
{
    Settings::setBool("ConsoleLogEnabled", enableConsoleLog->isChecked());
    Settings::setBool("FileLogEnabled", enableFileLog->isChecked());
    Settings::setLogLevel(logLevel->itemData(logLevel->currentIndex()).toInt());

    QString oldPath = Settings::getString("MtnExePath");
    QString newPath = mtnExecutablePath->text();
    bool checkUpperBound = relaxedVersionCheck->isChecked();

    QString installedVersion;

    // check if the new executable satisfies our version needs
    // we'll first check the program's plain version and if this
    // does not succeed we check for the interface version
    if (!APP->manager()->setMonotoneBinaryPath(newPath, checkUpperBound, installedVersion))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("The path to the monotone binary is either invalid "
               "or points to a version of monotone with which guitone can't "
               "work with. Guitone requires monotone with an interface version "
               "between '%1' and '%2'. Your installation has the interface "
               "version '%3'.")
                   .arg(MonotoneManager::MinInterfaceVersion)
                   .arg(MonotoneManager::MaxInterfaceVersion)
                   .arg(installedVersion),
            QMessageBox::Ok, 0, 0
        );
        return;
    }

    bool requireRestart = false;
    bool requireReload  = false;

    if (oldPath != newPath)
    {
        requireRestart = true;
        Settings::setString("MtnExePath", newPath);
    }

    Settings::setBool("CheckForUpdates", checkForUpdates->isChecked());
    Settings::setBool("SaveEncodingAsFileAttribute", saveEncodingAsFileAttribute->isChecked());
    Settings::setBool("FixUnwantedReverseDiffs", fixUnwantedReverseDiffs->isChecked());
    Settings::setBool("RelaxedVersionCheck", relaxedVersionCheck->isChecked());
    Settings::setBool("AskFileOpen", askFileOpen->isChecked());
    Settings::setString("AskFileOpenExt", askFileOpenExtensions->text());

    Settings::setString("AnnotationColorOldestRevision",
                        colorPickerOldestAnnotatedRevision->getSelectedColor().name());
    Settings::setString("AnnotationColorNewestRevision",
                        colorPickerNewestAnnotatedRevision->getSelectedColor().name());
    Settings::setString("AnnotationColorHighlight",
                        colorPickerHighlightAnnotatedRevision->getSelectedColor().name());

    Settings::setString("DiffColorAddedLines",
                        colorPickerAddedLines->getSelectedColor().name());
    Settings::setString("DiffColorRemovedLines",
                        colorPickerRemovedLines->getSelectedColor().name());

    Settings::setString("ChangesetBrowserSuspendedBranches",
                        colorPickerSuspendedBranches->getSelectedColor().name());
    Settings::setString("ChangesetBrowserMergeRevisions",
                        colorPickerMergeRevisions->getSelectedColor().name());
    Settings::setString("ChangesetBrowserOutOfBranchRevisions",
                        colorPickerOutOfBranchRevisions->getSelectedColor().name());

    if (Settings::getBool("ReadWorkspaceIncrementally") !=
        readWorkspaceIncrementally->isChecked())
    {
        requireReload = true;
        Settings::setBool("ReadWorkspaceIncrementally",
                          readWorkspaceIncrementally->isChecked());
    }

    if (Settings::getBool("ReadUserRcFiles") !=
        readUserRcFiles->isChecked())
    {
        requireReload = true;
        Settings::setBool("ReadUserRcFiles",
                          readUserRcFiles->isChecked());
    }

    if (requireRestart)
    {
        QMessageBox::information(
             this,
             tr("Notice"),
             tr("You need to restart guitone to use the new settings."),
             QMessageBox::Ok
        );
    }
    else
    if (requireReload)
    {
        QMessageBox::information(
             this,
             tr("Notice"),
             tr("You need to reopen your currently open databases and / or "
                "workspaces to use the new settings."),
             QMessageBox::Ok
        );
    }

    done(QDialog::Accepted);
}

void Preferences::openFileBrowser()
{
    QString file = QFileDialog::getOpenFileName(
         this,
         tr("Choose the monotone executable"),
         "",
         tr("Binaries (mtn mtn.exe)")
    );

    if (!file.isEmpty())
    {
        mtnExecutablePath->setText(file);
    }
}

void Preferences::warnAboutRelaxedVersionChecking(bool enabled)
{
    if (!enabled) return;

    QMessageBox::warning(
         this,
         tr("Warning"),
         tr("Activating this may let guitone work with newer, yet untested "
            "versions of monotone, in which certain functionalities may "
            "have changed or be no longer available.\n\n"
            "Use this at your own risk!"),
         QMessageBox::Ok
    );
}

void Preferences::resetAllSettings()
{
    if (QMessageBox::question(
          this,
          tr("Warning"),
          tr("Do you really want to reset all settings to their defaults?"),
          QMessageBox::Yes | QMessageBox::No) ==
        QMessageBox::Yes)
    {
        Settings::clear();
        Settings::sync();
        init();
    }
}

