/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "RevisionManifest.h"
#include "Platform.h"
#include "MonotoneUtil.h"
#include "FileExporter.h"
#include "Settings.h"

#include <QHeaderView>
#include <QDir>
#include <QMessageBox>
#include <QMenu>
#include <QFileDialog>

RevisionManifest::RevisionManifest(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();

    manifestModel = new Manifest(this, monotoneHandle);
    manifestView->setModel(manifestModel);

    connect(
        saveSelected, SIGNAL(clicked()),
        this, SLOT(saveSelectedFiles())
    );

    connect(
        saveAll, SIGNAL(clicked()),
        this, SLOT(saveAllFiles())
    );

    connect(
        manifestView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(openFile(const QModelIndex &))
    );

    connect(
        manifestView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

RevisionManifest::~RevisionManifest()
{
    delete manifestModel;
}

void RevisionManifest::readManifest(const QString & rev)
{
    revision = rev;
    setWindowTitle(tr("Files in %1").arg(revision));
    manifestModel->readManifest(revision);
}

void RevisionManifest::openFile(const QModelIndex & index)
{
    if (!index.isValid()) return;

    ManifestEntry * entry = static_cast<ManifestEntry *>(index.internalPointer());

    // ignore silently
    if (entry->is_directory) return;

    MonotoneTaskPtr task(new MonotoneTask(
        QStringList() << "get_file" << entry->hash
    ));
    MonotoneUtil::runSynchronousTask(monotoneHandle, task);

    if (!task->isFinished()) F("task aborted");

    if (task->getReturnCode() > 0)
    {
        C(QString("Cannot execute get_file: %1").arg(task->getDecodedOutput()));
        return;
    }

    //
    // FIXME: temporary files are currently saved into /tmp/guitone-<user>/<name>.ext
    // on unices and somewhere in %WINDIR%\temp\guitone-<user> or similar under Win32.
    // We need to check that
    //  a) user A can't read the temp files of user B (unsure in Win32...)
    //  b) what we do if two different instances of guitone run for the same
    //     user which both want the same filename
    //
    QString tempPath = QDir::tempPath();
    tempPath.append(QDir::separator()).append("guitone-").append(Platform::getUsername());
    QFileInfo path(tempPath);
    if (!path.isDir() && !QDir().mkpath(tempPath))
    {
        C("Could not create temporary guitone directory");
        return;
    }
    tempPath.append(QDir::separator()).append(entry->filename());

    QFile file(tempPath);
    if (file.exists()) file.remove();

    if (!file.open(QIODevice::WriteOnly))
    {
        C("Can't open temporary file for writing");
        return;
    }

    file.write(task->getOutput());
    file.close();

    if (Settings::getBool("AskFileOpen"))
    {
        QFileInfo fileInfo(tempPath);
        QString suffix = fileInfo.completeSuffix();
        QStringList harmfulSuffixes =
            Settings::getString("AskFileOpenExt").split(",", QString::SkipEmptyParts);

        if (harmfulSuffixes.contains(suffix))
        {
            QMessageBox::StandardButton btn = QMessageBox::question(
                NULL,
                QObject::tr("Attention"),
                QObject::tr("The file '%1' you're trying to open is included "
                            "in the list of precarious file formats - do you "
                            "really want to continue?").arg(tempPath),
                QMessageBox::Yes | QMessageBox::No
            );

            if (btn == QMessageBox::No)
                return;
        }
    }

    try
    {
        Platform::openFile(tempPath);

        // if the action was successful, block execution (i.e. deleting of the
        // temporary file) until the user tells us to continue
        QMessageBox::information(
            this,
            tr("Information"),
            tr("Please close this message to remove the temporary file.")
        );
    }
    catch (GuitoneException & e)
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Could not open '%1': %2").arg(tempPath).arg(e.what()),
            QMessageBox::Ok, 0, 0
        );
    }

    file.remove();
}

void RevisionManifest::contextMenuRequested(const QModelIndexList & indexList,
                                            const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QAction * actOpenFile = 0;
    QAction * actFileHistory = 0;
    QAction * actAnnotate = 0;

    if (indexList.size() == 1 &&
        !static_cast<ManifestEntry *>(indexList.at(0).internalPointer())->is_directory)
    {
        actOpenFile = menu.addAction(tr("Open file"));
        actFileHistory = menu.addAction(tr("History"));
        actAnnotate = menu.addAction(tr("Annotate"));
        actOpenFile->setFont(activeFont);
        menu.addSeparator();

        // FIXME: change this as soon as there is an automate annotate
        if (monotoneHandle->getType() == MonotoneHandle::server_handle)
        {
            actAnnotate->setEnabled(false);
        }
    }

    QAction * actSaveSelected = menu.addAction(tr("Save selected"));
    QAction * actSaveAll = menu.addAction(tr("Save all"));

    QAction * act = menu.exec(pos);
    if (act == actOpenFile)
    {
        openFile(indexList.at(0));
    }
    else if (act == actFileHistory)
    {
        ManifestEntry * entry = static_cast<ManifestEntry *>(indexList.at(0).internalPointer());
        emit fileHistory(entry->path, revision);
    }
    else if (act == actAnnotate)
    {
        ManifestEntry * entry = static_cast<ManifestEntry *>(indexList.at(0).internalPointer());
        emit annotateFile(revision, entry->path);
    }
    else if (act == actSaveSelected)
    {
        saveSelectedFiles();
    }
    else if (act == actSaveAll)
    {
        saveAllFiles();
    }
}

void RevisionManifest::saveAllFiles()
{
    QModelIndexList list;
    int row = 0;
    while (true)
    {
        QModelIndex index(manifestModel->index(row++, 0, QModelIndex()));
        if (!index.isValid()) break;
        list.append(index);
    }

    QString exportDir =
        QFileDialog::getExistingDirectory(0, tr("Select your export directory..."));
    if (!exportDir.isEmpty())
    {
        FileExporter exporter(monotoneHandle, revision, QDir(exportDir));

        FileEntryList entries = extractSelectedEntries(list);
        if (!exporter.run(entries))
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("The file export was aborted - please look in the log for more details.")
            );
            return;
        }
    }
}

void RevisionManifest::saveSelectedFiles()
{
    QModelIndexList indexes = manifestView->selectionModel()->selectedIndexes();
    FileEntryList entries = extractSelectedEntries(indexes);

    if (entries.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Information"),
            tr("Please select one or more files you want to save locally.")
        );
        return;
    }

    QString exportDir =
        QFileDialog::getExistingDirectory(0, tr("Select your export directory..."));
    if (!exportDir.isEmpty())
    {
        FileExporter exporter(monotoneHandle, revision, QDir(exportDir));
        if (!exporter.run(entries))
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("The file export was aborted - please look in the log for more details.")
            );
            return;
        }
    }
}

//! \fixme this should be moved into the FileExporter class
/*!
    \sa FileExporter
*/
FileEntryList RevisionManifest::extractSelectedEntries(const QModelIndexList & indexList)
{
    FileEntryList entries;
    for (int i=0, j=indexList.size(); i<j; i++)
    {
        ManifestEntry * entry =
            static_cast<ManifestEntry*>(indexList.at(i).internalPointer());

        if (entry->is_directory)
        {
            entries.append(FileEntry(entry->path, true));
            entries += addEntriesRecursive(entry);
        }
        else
        {
            entries.append(FileEntry(entry->path, false));
        }
    }

    return entries;
}

//! \fixme this should be moved into the FileExporter class
/*!
    \sa FileExporter
*/
FileEntryList RevisionManifest::addEntriesRecursive(ManifestEntry * entry)
{
    FileEntryList entries;
    for (int i=0, j=entry->childCount(); i<j; i++)
    {
        if (entry->children.at(i)->is_directory)
        {
            entries.append(FileEntry(entry->children.at(i)->path, true));
            entries += addEntriesRecursive(entry->children.at(i));
        }
        else
        {
            entries.append(FileEntry(entry->children.at(i)->path, false));
        }
    }
    return entries;
}

