/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "SelectRevision.h"
#include "Settings.h"
#include "MonotoneUtil.h"

#include <QMessageBox>
#include <QHeaderView>
#include <QMenu>

SelectRevision::SelectRevision(QWidget * parent, const MonotoneHandlePtr & database)
    : Dialog(parent), selectorCompleter(0), monotoneHandle(database)
{
    setupUi(this);
    Dialog::init();

    splitter->restoreState();

    // add some items to the selector box
    selectorBox->addItem(tr("Revision"), "i");
    selectorBox->addItem(tr("Branch"), "b");
    selectorBox->addItem(tr("Tag"), "t");
    selectorBox->addItem(tr("Author"), "a");
    selectorBox->addItem(tr("Date"), "d");
    selectorBox->addItem(tr("Custom"), "c");
    selectorBox->addItem(tr("Combined"), "");

    // create the models
    selectorModel       = new Select(this, monotoneHandle);
    sortModel           = new Toposort(this, monotoneHandle);
    certsModel          = new Certs(this, monotoneHandle);
    branchesModel       = new Branches(this, monotoneHandle);
    tagsModel           = new Tags(this, monotoneHandle);
    keysModel           = new Keys(this, monotoneHandle);

    // toposort revisions
    sortModel->setSourceModel(selectorModel);

    connect(
        sortModel, SIGNAL(sortingFinished()),
        this, SLOT(trySelectFirstRow())
    );

    // set initial completer
    setCompleter(0);

    // assign the models to the views
    revisionList->setModel(sortModel);

    certList->setKeyMap(MonotoneUtil::getPublicKeyMap(monotoneHandle));

    connect(
        certsModel, SIGNAL(certsRead(const CertList &)),
        certList, SLOT(setCertList(const CertList &))
    );

    // enable sorting of revisions
    revisionList->setSortingEnabled(true);
    revisionList->sortByColumn(0, Qt::AscendingOrder);

    // connect its item clicked signal with a local slot
    connect(
        revisionList, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(enableDisableOkButton(const QModelIndex &))
    );

    // trigger the search when clicking on "Find"
    connect(
        searchRevsBtn, SIGNAL(clicked()),
        this, SLOT(triggerRevisionSearch())
    );

    // display the certs of a selected revision on click
    connect(
        revisionList, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(readCerts(const QModelIndex &))
    );

    connect(
        revisionList, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(enableDiffButtons(const QModelIndex &))
    );

    // show the manifest on double-click and right-click (context menu)
    connect(
        revisionList, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(revListDoubleClicked(const QModelIndex &))
    );
    connect(
        revisionList, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuRequested(const QModelIndexList &, const QPoint &))
    );

    // handle selector errors
    connect(
        selectorModel, SIGNAL(invalidSelection(QString)),
        this, SLOT(handleInvalidSelection(QString))
    );

    // connect to the completer for valid types
    connect(
        selectorBox, SIGNAL(currentIndexChanged(int)),
        this, SLOT(setCompleter(int))
    );

    connect(
        selectFirst, SIGNAL(clicked()),
        this, SLOT(setFirstRevision())
    );

    connect(
        selectSecond, SIGNAL(clicked()),
        this, SLOT(setSecondRevision())
    );

    connect(
        showDiff, SIGNAL(clicked()),
        this, SLOT(openDiffDialog())
    );

    // disable the OK button unless we have a valid selected revision
    buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
}

SelectRevision::~SelectRevision()
{
    splitter->saveState();

    delete selectorModel;
    delete sortModel;
    delete certsModel;
    delete branchesModel;
    delete tagsModel;
    delete keysModel;
}

void SelectRevision::triggerRevisionSearch()
{
    selectFirst->setEnabled(false);
    selectSecond->setEnabled(false);
    showDiff->setEnabled(false);
    enableDisableOkButton(QModelIndex());

    QString selector = selectorBox->itemData(selectorBox->currentIndex()).toString();

    if (selector == "i")
    {
        Settings::addItemToList("RecentRevisionSelector", selectorValue->text(), 10);
    }

    if (selector == "d")
    {
        Settings::addItemToList("RecentDateSelector", selectorValue->text(), 10);
    }

    if (selector == "c")
    {
        Settings::addItemToList("RecentCustomSelector", selectorValue->text(), 10);
    }

    if (selector == "")
    {
        Settings::addItemToList("RecentCombinedSelector", selectorValue->text(), 10);
    }

    // combined selectors have no common prefix
    if (selector != "")
    {
        selector.append(":");
    }

    selector.append(selectorValue->text());
    selectorModel->readSelection(selector);

    // remove any previously set data in this model and update the view
    certsModel->clearData();
}

void SelectRevision::handleInvalidSelection(QString error)
{
    QMessageBox::warning(
        this,
        tr("Invalid selection"),
        tr("The given selection could not be processed: %1").arg(error),
        QMessageBox::Ok
    );
}

void SelectRevision::enableDisableOkButton(const QModelIndex & index)
{
    if (!index.isValid())
    {
        buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
        return;
    }

    QString rev(index.data().toString());
    if (rev.size() == 0)
    {
        buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
        return;
    }

    buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void SelectRevision::setCompleter(int index)
{
    QString selector = selectorBox->itemData(index).toString();

    if (selectorCompleter) delete selectorCompleter;

    if (selector == "t")
    {
        tagsModel->readTags(QString());
        selectorCompleter = new QCompleter(this);
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorCompleter->setModel(tagsModel);
        selectorCompleter->setCompletionRole(Qt::DisplayRole);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    if (selector == "a")
    {
        keysModel->readKeys();
        selectorCompleter = new QCompleter(this);
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorCompleter->setModel(keysModel);
        selectorCompleter->setCompletionRole(Qt::DisplayRole);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    if (selector == "b")
    {
        branchesModel->readBranches();
        selectorCompleter = new QCompleter(this);
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorCompleter->setModel(branchesModel);
        selectorCompleter->setCompletionRole(Qt::DisplayRole);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    if (selector == "i")
    {
        selectorCompleter = new QCompleter(
             Settings::getItemList("RecentRevisionSelector"),
             this
        );
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    if (selector == "d")
    {
        selectorCompleter = new QCompleter(
             Settings::getItemList("RecentDateSelector"),
             this
        );
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    if (selector == "c")
    {
        selectorCompleter = new QCompleter(
             Settings::getItemList("RecentCustomSelector"),
             this
        );
        selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
        selectorValue->setCompleter(selectorCompleter);
        return;
    }

    selectorCompleter = new QCompleter(
         Settings::getItemList("RecentCombinedSelector"),
         this
    );
    selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
    selectorValue->setCompleter(selectorCompleter);
}

void SelectRevision::setFirstRevision()
{
    I(currentRevision.size() > 0);
    if (currentRevision == secondRevision) return;
    firstRevision = currentRevision;
    if (secondRevision.size() > 0) showDiff->setEnabled(true);
    selectFirst->setText(tr("First: %1...").arg(firstRevision.left(12)));
}

void SelectRevision::setSecondRevision()
{
    I(currentRevision.size() > 0);
    if (currentRevision == firstRevision) return;
    secondRevision = currentRevision;
    if (firstRevision.size() > 0) showDiff->setEnabled(true);
    selectSecond->setText(tr("Second: %1...").arg(secondRevision.left(12)));
}

void SelectRevision::enableDiffButtons(const QModelIndex & index)
{
    if (!index.isValid()) return;
    currentRevision = index.data().toString();
    selectFirst->setEnabled(true);
    selectSecond->setEnabled(true);
}

void SelectRevision::openDiffDialog()
{
    I(firstRevision.size() > 0 && secondRevision.size() > 0);
    emit revisionDiff (QString(), firstRevision, secondRevision);
}

void SelectRevision::readCerts(const QModelIndex & index)
{
    if (!index.isValid()) return;

    QString rev(index.data().toString());
    certsModel->readCerts(rev);
}

void SelectRevision::revListDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;
    emit revisionManifest(index.data().toString());
}

void SelectRevision::contextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;
    QString rev(indexList.at(0).data().toString());

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QAction * actOpenManifest = menu.addAction(tr("View files in this revision"));
    actOpenManifest->setFont(activeFont);
    QAction * actSetFirstRevision = menu.addAction(tr("Select as first revision"));
    QAction * actSetSecondRevision = menu.addAction(tr("Select as second revision"));

    QAction * act = menu.exec(pos);
    if (act == actOpenManifest)
    {
        emit revisionManifest(rev);
    }
    else if (act == actSetFirstRevision)
    {
        setFirstRevision();
    }
    else if (act == actSetSecondRevision)
    {
        setSecondRevision();
    }
}

void SelectRevision::queryRevisions(const QString & selector)
{
    selectorBox->setCurrentIndex(selectorBox->findData(""));
    selectorValue->setText(selector);
    triggerRevisionSearch();
}

void SelectRevision::accept()
{
    // since we gray out the ok button we _should_
    // have a current revision by now
    I(!currentRevision.isEmpty());
    emit revisionSelected(currentRevision);
    done(0);
}

void SelectRevision::trySelectFirstRow()
{
    if (sortModel->rowCount() > 0)
    {
        QModelIndex idx = sortModel->index(0, 0);
        currentRevision = idx.data().toString();
        revisionList->setCurrentIndex(idx);
        readCerts(idx);
        enableDisableOkButton(idx);
    }
}

