/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "UnaccountedRenames.h"
#include "IconProvider.h"

#include <QFont>
#include <QHeaderView>
#include <QMessageBox>

UnaccountedRenames::UnaccountedRenames(QWidget * parent, const MonotoneHandlePtr & handle)
  : Dialog(parent), monotoneHandle(handle), inventory(0)
{
    setupUi(this);
    Dialog::init();

    inventory = new Inventory(this);
    inventory->setMonotoneHandle(monotoneHandle);

    connect(
        inventory, SIGNAL(inventoryRead(const QString &)),
        this, SLOT(findUnaccountedRenames())
    );

    connect(
        renameTree, SIGNAL(itemClicked(QTreeWidgetItem *, int)),
        this, SLOT(itemClicked(QTreeWidgetItem *, int))
    );

    connect(
        performRenames, SIGNAL(clicked()),
        this, SLOT(doPerformRenames())
    );

    renameTree->setHeaderLabels(QStringList()
        << tr("Missing paths and possible rename targets")
        << tr("File ID (if applicable)")
    );

    renameTree->header()->resizeSection(0, 300);
}

UnaccountedRenames::~UnaccountedRenames()
{
    delete inventory;
}

void UnaccountedRenames::checkForUnaccountedRenames()
{
    connect(
        inventory, SIGNAL(inventoryRead(const QString &)),
        this, SLOT(findUnaccountedRenames())
    );

    inventory->read(QString(), true);
}

void UnaccountedRenames::findUnaccountedRenames()
{
    const QMap<FileEntry, FileEntryList> & renames =
        inventory->findUnaccountedRenames();

    if (renames.size() == 0)
    {
        QString msg = tr("No unaccounted renames found for this workspace.");
        if (isVisible())
            msg = tr("No more unaccounted renames found for this workspace.");

        QMessageBox::information(
            this,
            tr("Nothing found"),
            msg,
            QMessageBox::Ok
        );

        disconnect(
            inventory, SIGNAL(inventoryRead(const QString &)),
            this, SLOT(findUnaccountedRenames())
        );

        done(0);
        return;
    }

    renameTree->clear();
    performRenames->setEnabled(true);
    renameTree->setEnabled(true);

    QFont boldFont;
    boldFont.setBold(true);
    IconProvider * iconProvider = IconProvider::singleton();

    QMapIterator<FileEntry, FileEntryList> i(renames);
    while (i.hasNext())
    {
        i.next();
        FileEntry sen = i.key();
        FileEntryList list = i.value();
        QTreeWidgetItem * source = new QTreeWidgetItem(renameTree);
        source->setFlags(Qt::ItemIsEnabled);
        source->setText(0, sen.path);
        source->setFont(0, boldFont);
        source->setIcon(0, sen.is_dir ?
                        iconProvider->getPlainFolderIcon() :
                        iconProvider->getPlainFileIcon());
        source->setText(1, sen.fileid.isEmpty() ? tr("n/a") : sen.fileid);
        source->setFont(1, boldFont);

        foreach (const FileEntry & en, list)
        {
            QTreeWidgetItem * target = new QTreeWidgetItem(source);
            target->setFlags(Qt::ItemIsEnabled | Qt::ItemIsUserCheckable);
            target->setCheckState(0, Qt::Unchecked);
            target->setText(0, en.path);
            target->setToolTip(0, en.fileid.isEmpty() || sen.fileid != en.fileid ?
                               tr("matched by name") :
                               tr("matched by content"));

            target->setText(1, en.fileid.isEmpty() ? tr("n/a") : en.fileid);
        }
    }

    disconnect(
        inventory, SIGNAL(inventoryRead(const QString &)),
        this, SLOT(findUnaccountedRenames())
    );

    show();
}

void UnaccountedRenames::itemClicked(QTreeWidgetItem * item, int column)
{
    if (item == 0 || column == -1) return;
    QTreeWidgetItem * parent = item->parent();
    // top level item clicked
    if (!parent) return;

    Qt::CheckState state = item->checkState(0);

    for (int i=0, j=parent->childCount(); i<j; i++)
    {
        parent->child(i)->setCheckState(0, Qt::Unchecked);
    }

    item->setCheckState(0, state);
}

void UnaccountedRenames::doPerformRenames()
{
    QMap<QString, QString> renames;

    QTreeWidgetItem * rootItem = renameTree->invisibleRootItem();
    for (int i=0; i < rootItem->childCount(); i++)
    {
        QTreeWidgetItem * target = 0;
        QTreeWidgetItem * source = rootItem->child(i);

        for (int j=0; j < source->childCount(); j++)
        {
            if (source->child(j)->checkState(0) == Qt::Checked)
            {
                target = source->child(j);
                break;
            }
        }

        if (target == 0)
            continue;

        renames.insert(source->text(0), target->text(0));
    }

    if (renames.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Nothing selected"),
            tr("Please select a rename target for one or more missing items."),
            QMessageBox::Ok
        );
        return;
    }

    performRenames->setEnabled(false);
    renameTree->setEnabled(false);

    QMapIterator<QString, QString> it(renames);
    while (it.hasNext())
    {
        it.next();
        emit renamePath(it.key(), it.value());
    }

    checkForUnaccountedRenames();
}

