/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "UpdateWorkspace.h"
#include "SelectRevision.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QMessageBox>
#include <QTimer>

UpdateWorkspace::UpdateWorkspace(QWidget * parent, const MonotoneHandlePtr & handle)
    : Dialog(parent), monotoneHandle(handle)
{
    setupUi(this);
    Dialog::init();

    branchesModel = new Branches(this, handle);
    branchList->setModel(branchesModel);

    tagsModel = new Tags(this, handle);
    tagList->setModel(tagsModel);

    selectorCompleter = new QCompleter(
         Settings::getItemList("RecentUpdateSelectors"),
         this
    );

    selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
    selectedRevision->setCompleter(selectorCompleter);

    connect(
        selectRevision, SIGNAL(clicked()),
        this, SLOT(slotSelectRevision())
    );

    connect(
        branchList, SIGNAL(activated(int)),
        this, SLOT(selectBranchOption())
    );

    connect(
        tagList, SIGNAL(activated(int)),
        this, SLOT(selectTagOption())
    );

    connect(
        selectedRevision, SIGNAL(textChanged(const QString &)),
        this, SLOT(selectSelectorOption())
    );

    mtnProcess = new MonotoneProcess();
    mtnProcess->setMonotoneHandle(handle);

    connect(
        mtnProcess, SIGNAL(output(const QString &)),
        this, SLOT(processOutput(const QString &))
    );

    connect(
        mtnProcess, SIGNAL(error(const QString &)),
        this, SLOT(processError(const QString &))
    );

    connect(
        mtnProcess, SIGNAL(finished(int, QProcess::ExitStatus)),
        this, SLOT(processFinished(int, QProcess::ExitStatus))
    );
}

UpdateWorkspace::~UpdateWorkspace()
{
    delete branchesModel;
    delete tagsModel;
    delete selectorCompleter;
    delete mtnProcess;
}

void UpdateWorkspace::init()
{
    tagsModel->readTags(QString());
    branchesModel->readBranches();

    updateToLatest->setText(
        tr("the latest revision (branch: %1)")
            .arg(MonotoneUtil::getBranchName(monotoneHandle))
    );

    updateSelection->setEnabled(true);
    buttonBox->setEnabled(true);
    updateProgress->hide();
}

void UpdateWorkspace::accept()
{
    QString selector;

    if (updateToLatest->isChecked())
    {
        selector = "h:" + MonotoneUtil::getBranchName(monotoneHandle);
    }

    if (updateToBranch->isChecked())
    {
        QString selection(branchList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No branch selected"),
                tr("Please select a branch from the list of branches."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "h:" + selection;
    }

    if (updateToTag->isChecked())
    {
        QString selection(tagList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No tag selected"),
                tr("Please select a tag from the list of tags (if any are available)."),
                QMessageBox::Ok
            );
            return;
        }
        selector = "t:" + selection;
    }

    if (updateToRevision->isChecked())
    {
        QString selection = selectedRevision->text();
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No selector given"),
                tr("Please enter a monotone selector or choose a revision "
                   "by clicking on 'select revision'."),
                QMessageBox::Ok
            );
            return;
        }
        selector = selection;
        Settings::addItemToList("RecentUpdateSelectors", selection, 10);
    }

    QStringList revList =
        MonotoneUtil::resolveSelector(monotoneHandle, selector);

    if (revList.size() == 0)
    {
        QMessageBox::warning(
            this,
            tr("No revisions matched"),
            tr("Your selection didn't match any revisions or was invalid."),
            QMessageBox::Ok
        );
        return;
    }

    if (revList.size() > 1)
    {
        QMessageBox::information(
            this,
            tr("Multiple revisions matched"),
            tr("Your selection matched multiple revisions. Please choose one."),
            QMessageBox::Ok
        );

        emit selectRevisionFromSelector(selector);
        return;
    }

    selectedRevisionId = revList.at(0);
    QString baseRevisionId =
        MonotoneUtil::getBaseWorkspaceRevision(monotoneHandle);

    QStringList commonAncestors = MonotoneUtil::getCommonAncestors(
        monotoneHandle,
        QStringList() << selectedRevisionId << baseRevisionId
    );

    if (commonAncestors.size() == 0)
    {
         QMessageBox::StandardButton btn = QMessageBox::question(
            this,
            tr("No common ancestors found"),
            tr("The selected revision shares no common ancestor with the "
               "current revision of the workspace. Do you really want to "
               "continue?"),
            QMessageBox::Yes | QMessageBox::No
         );

         if (btn == QMessageBox::No) return;
    }

    updateSelection->setEnabled(false);
    buttonBox->setEnabled(false);
    output->clear();
    updateProgress->show();

    emit beforeWorkspaceUpdate();

    QStringList params;
    params  << "update"
            << "-r" << selectedRevisionId
            << "--move-conflicting-paths";
    mtnProcess->start(params);
}

void UpdateWorkspace::processOutput(const QString & out)
{
    QListWidgetItem * item = 0;
    foreach (QString line, out.split("\n", QString::SkipEmptyParts))
    {
        item = new QListWidgetItem(line, output);
        output->addItem(item);
    }
    if (item) output->scrollToItem(item);
}

void UpdateWorkspace::processError(const QString & error)
{
    QListWidgetItem * item = new QListWidgetItem(
        tr("Update failed: %1").arg(error),
        output
    );
    item->setForeground(QBrush(QColor("red")));
    output->addItem(item);
    output->scrollToItem(item);

    updateSelection->setEnabled(true);
    buttonBox->setEnabled(true);
}

void UpdateWorkspace::processFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    Q_UNUSED(exitStatus);
    if (exitCode == 0)
    {
        QListWidgetItem * item = new QListWidgetItem(
            tr("Update successfully completed"), output
        );
        item->setForeground(QBrush(QColor("green")));
        output->addItem(item);
        output->scrollToItem(item);

        emit workspaceUpdated(selectedRevisionId);
        // give the user some time to read the successful message
        QTimer::singleShot(1000, this, SLOT(close()));
        return;
    }
    processError(tr("monotone exited with code %1").arg(exitCode));
}

void UpdateWorkspace::slotSelectRevision()
{
    emit selectRevisionFromSelector(QString());
}

void UpdateWorkspace::setSelectedRevision(const QString & revision)
{
    selectedRevision->setText(revision);
}

void UpdateWorkspace::selectBranchOption()
{
    updateToBranch->setChecked(true);
}

void UpdateWorkspace::selectTagOption()
{
    updateToTag->setChecked(true);
}

void UpdateWorkspace::selectSelectorOption()
{
    updateToRevision->setChecked(true);
}

void UpdateWorkspace::closeEvent(QCloseEvent * event)
{
    if (mtnProcess->state() == QProcess::Running)
    {
        event->ignore();
        return;
    }
    event->accept();
}

