/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DatabaseWindow.h"
#include "DatabaseMenuBar.h"
#include "DatabaseDialogManager.h"
#include "GuitoneCore.h"
#include "Settings.h"
#include "MonotoneUtil.h"

#include <QMessageBox>
#include <QHBoxLayout>
#include <QTimer>

DatabaseWindow::DatabaseWindow() : MainWindow(),
    loadedDbLabel(0), keyManagement(0), revisionBrowser(0),
    checkoutRevision(0), databaseVariables(0)
{
    setObjectName("DatabaseWindow");
}

DatabaseWindow::~DatabaseWindow()
{
    if (loadedDbLabel)      delete loadedDbLabel;
    if (keyManagement)      delete keyManagement;
    if (revisionBrowser)    delete revisionBrowser;
    if (checkoutRevision)   delete checkoutRevision;
    if (databaseVariables)
    {
        removeDockWidget(databaseVariables);
        delete databaseVariables;
    }
}

void DatabaseWindow::init()
{
    if (!menuBar)
    {
        menuBar = new DatabaseMenuBar(this);
    }

    if (!dialogManager)
    {
        dialogManager = new DatabaseDialogManager(this);
    }

    connect(
        dialogManager, SIGNAL(revisionCheckedOut(const MonotoneHandlePtr &)),
        APP, SLOT(loadFromHandle(const MonotoneHandlePtr &))
    );

    connect(
        dialogManager, SIGNAL(newProjectCreated(const MonotoneHandlePtr &)),
        APP, SLOT(loadFromHandle(const MonotoneHandlePtr &))
    );

    connect(
        menuBar, SIGNAL(showNetsync()),
        dialogManager, SLOT(showNetsync())
    );

    connect(
        menuBar, SIGNAL(showChangesetBrowser()),
        dialogManager, SLOT(showChangesetBrowser())
    );

    connect(
        menuBar, SIGNAL(showCheckoutRevision()),
        dialogManager, SLOT(showCheckoutRevision())
    );

    connect(
        menuBar, SIGNAL(showKeyManagement()),
        dialogManager, SLOT(showKeyManagement())
    );

    connect(
        menuBar, SIGNAL(showNewProjectSetup()),
        dialogManager, SLOT(showNewProjectSetup())
    );

    databaseVariables = new DatabaseVariables(this);

    dynamic_cast<DatabaseMenuBar *>(menuBar)
        ->addDockWidgetAction(databaseVariables->toggleViewAction());

    MainWindow::init();
}

void DatabaseWindow::setup()
{
    //
    // UI stuff
    //
    QWidget * centralWidget = new QWidget(this);

    QVBoxLayout * vboxLayout = new QVBoxLayout(centralWidget);
    QGroupBox *loadedDbGroup = new QGroupBox(tr("Loaded database"), centralWidget);
    vboxLayout->addWidget(loadedDbGroup);

    QVBoxLayout * vboxLayout1 = new QVBoxLayout(loadedDbGroup);
    loadedDbLabel = new QLabel(tr("nothing loaded"), this);
    vboxLayout1->addWidget(loadedDbLabel);

    QGroupBox * actionGroup = new QGroupBox(tr("Actions"), centralWidget);
    QVBoxLayout * vboxLayout2 = new QVBoxLayout(actionGroup);
    vboxLayout->addWidget(actionGroup);

    QHBoxLayout * hboxLayout = new QHBoxLayout();

    keyManagement = new QPushButton(tr("Key management"), actionGroup);
    hboxLayout->addWidget(keyManagement);
    revisionBrowser = new QPushButton(tr("Revision browser"), actionGroup);
    hboxLayout->addWidget(revisionBrowser);
    checkoutRevision = new QPushButton(tr("Checkout revision"), actionGroup);
    hboxLayout->addWidget(checkoutRevision);

    vboxLayout2->addLayout(hboxLayout);

    connect(
        keyManagement, SIGNAL(clicked()),
        dialogManager, SLOT(showKeyManagement())
    );

    connect(
        revisionBrowser, SIGNAL(clicked()),
        dialogManager, SLOT(showChangesetBrowser())
    );

    connect(
        checkoutRevision, SIGNAL(clicked()),
        dialogManager, SLOT(showCheckoutRevision())
    );

    setCentralWidget(centralWidget);
}

void DatabaseWindow::load(const MonotoneHandlePtr & handle)
{
    I(handle->getType() == MonotoneHandle::database_handle);
    APP->manager()->getThread(handle);
    monotoneHandle = handle;

    QString dbpath = handle->getData();
    loadedDbLabel->setText(dbpath);

    QFileInfo fi(dbpath);
    setWindowTitle(
        tr("%1 - database mode - guitone").arg(fi.fileName())
    );

    dialogManager->init(monotoneHandle);
    databaseVariables->setMonotoneHandle(monotoneHandle);

    QTimer::singleShot(0, this, SLOT(checkForEmptyDatabase()));
}

void DatabaseWindow::checkForEmptyDatabase()
{
    bool ok;
    if (MonotoneUtil::getRoots(monotoneHandle, ok).size() == 0 && ok)
    {
        QMessageBox msgBox(this);
        msgBox.setWindowFlags(Qt::Sheet);
        msgBox.setIcon(QMessageBox::Question);
        msgBox.setText("This database is empty");
        msgBox.setInformativeText("Do you want to pull in new changes "
                                  "or setup a new project?");

        QPushButton * netsyncBtn = msgBox.addButton(tr("Pull data"),
                                                    QMessageBox::ActionRole);
        QPushButton * setupBtn   = msgBox.addButton(tr("Setup project"),
                                                    QMessageBox::ActionRole);
        msgBox.addButton(QMessageBox::Abort);

        msgBox.exec();

        if (msgBox.clickedButton() == netsyncBtn)
        {
            qobject_cast<DatabaseDialogManager *>(dialogManager)->showNetsync();
        }
        else
        if (msgBox.clickedButton() == setupBtn)
        {
            qobject_cast<DatabaseDialogManager *>(dialogManager)->showNewProjectSetup();
        }
    }
}

