/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "WorkspaceWindow.h"
#include "GuitoneCore.h"
#include "Settings.h"
#include "Platform.h"
#include "MonotoneUtil.h"
#include "InventoryItem.h"
#include "WorkspaceMenuBar.h"
#include "WorkspaceDialogManager.h"

#include <QMessageBox>

WorkspaceWindow::WorkspaceWindow() : DatabaseWindow(), mainSplitter(0),
    listSplitter(0), treeView(0), listView(0), attrView(0), statusBar(0),
    toolBar(0), searchInput(0), iconHelp(0), invModel(0), attrModel(0),
    proxyModelFolderTree(0), proxyModelFileList(0), invWatcher(0)
{
    setObjectName("WorkspaceWindow");
    setDockOptions(QMainWindow::ForceTabbedDocks);
    setUnifiedTitleAndToolBarOnMac(true);
}

WorkspaceWindow::~WorkspaceWindow()
{
    listSplitter->saveState();
    mainSplitter->saveState();

    // order is important here - at first the views inside the splitters,
    // then the splitters themselves!
    if (treeView) delete treeView;
    if (listView) delete listView;
    if (attrView) delete attrView;
    if (listSplitter) delete listSplitter;
    if (mainSplitter) delete mainSplitter;
    if (statusBar) delete statusBar;
    if (searchInput) delete searchInput;
    if (toolBar) delete toolBar;
    if (iconHelp)
    {
        removeDockWidget(iconHelp);
        delete iconHelp;
    }

    // do the same with the models
    if (proxyModelFolderTree) delete proxyModelFolderTree;
    if (proxyModelFileList) delete proxyModelFileList;
    if (invModel) delete invModel;
    if (attrModel) delete attrModel;
    if (invWatcher) delete invWatcher;
}

void WorkspaceWindow::init()
{
    if (!menuBar)
    {
        menuBar = new WorkspaceMenuBar(this);
    }

    if (!dialogManager)
    {
        dialogManager = new WorkspaceDialogManager(this);
    }

    connect(
        menuBar, SIGNAL(showUpdateWorkspace()),
        dialogManager, SLOT(showUpdateWorkspace())
    );

    connect(
        menuBar, SIGNAL(showCommitRevision()),
        dialogManager, SLOT(showCommitRevision())
    );

    connect(
        menuBar, SIGNAL(checkForUnaccountedRenames()),
        dialogManager, SLOT(checkForUnaccountedRenames())
    );

    DatabaseWindow::init();
}

void WorkspaceWindow::setup()
{
    statusBar = new QStatusBar(this);
    setStatusBar(statusBar);

    toolBar = new QToolBar(this);

    QWidget * spacer = new QWidget(toolBar);
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    toolBar->addWidget(spacer);

    searchInput = new SearchInput(toolBar);
    searchInput->setMaximumWidth(150);
    toolBar->addWidget(searchInput);

    addToolBar(toolBar);

    mainSplitter = new Splitter(this, "MainSplitter");
    mainSplitter->setOrientation(Qt::Horizontal);

    treeView = new InventoryView(mainSplitter, "FolderTree");
    treeView->setEditTriggers(QAbstractItemView::EditKeyPressed);
    treeView->setSelectionMode(QAbstractItemView::ExtendedSelection);

    listSplitter = new Splitter(mainSplitter, "ListSplitter");
    listSplitter->setOrientation(Qt::Vertical);

    listView = new InventoryView(listSplitter, "FileList");
    listView->setEditTriggers(QAbstractItemView::EditKeyPressed);
    listView->setSelectionMode(QAbstractItemView::ExtendedSelection);

    attrView = new AttributesView(listSplitter, "AttributesView");
    attrView->setAlternatingRowColors(true);

    connect(
        treeView, SIGNAL(selectedRows(const QModelIndexList &)),
        this, SLOT(maybeReadNodeInfo(const QModelIndexList &))
    );
    connect(
        listView, SIGNAL(selectedRows(const QModelIndexList &)),
        this, SLOT(maybeReadNodeInfo(const QModelIndexList &))
    );

    connect(
        treeView, SIGNAL(directoryChanged(const QModelIndex &)),
        listView, SLOT(changeDirectory(const QModelIndex &))
    );
    connect(
        listView, SIGNAL(directoryChanged(const QModelIndex &)),
        treeView, SLOT(changeDirectory(const QModelIndex &))
    );

    connect(
        listView, SIGNAL(diffFile(const QString &)),
        dialogManager, SLOT(showFileDiff(const QString &))
    );

    connect(
        listView, SIGNAL(diffRevision(const QString &, const QString &, const QString &)),
        dialogManager, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
    );

    connect(
        listView, SIGNAL(fileHistory(const QString &)),
        dialogManager, SLOT(showFileHistory(const QString &))
    );

    connect(
        listView, SIGNAL(annotateFile(const QString &)),
        dialogManager, SLOT(showAnnotation(const QString &))
    );

    connect(
        listView, SIGNAL(commitRevision(const QStringList &)),
        dialogManager, SLOT(showCommitRevision(const QStringList &))
    );

    connect(
        listView, SIGNAL(openFile(const QString &)),
        this, SLOT(openFile(const QString &))
    );

    connect(
        listView, SIGNAL(deleteLocalPath(const QString &)),
        this, SLOT(deleteLocalPath(const QString &))
    );

    connect(
        treeView, SIGNAL(deleteLocalPath(const QString &)),
        this, SLOT(deleteLocalPath(const QString &))
    );

    // the tree view shows only directories, therefor file diff / file history
    // doesn't make sense there
    connect(
        treeView, SIGNAL(diffRevision(const QString &, const QString &, const QString &)),
        dialogManager, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
    );

    connect(
        treeView, SIGNAL(commitRevision(const QStringList &)),
        dialogManager, SLOT(showCommitRevision(const QStringList &))
    );

    connect(
        treeView, SIGNAL(openFile(const QString &)),
        this, SLOT(openFile(const QString &))
    );

    connect(
        menuBar, SIGNAL(expandTree(bool)),
        treeView, SLOT(expandAllNodes(bool))
    );

    listSplitter->addWidget(listView);
    listSplitter->addWidget(attrView);

    mainSplitter->addWidget(treeView);
    mainSplitter->addWidget(listSplitter);

    setCentralWidget(mainSplitter);

    mainSplitter->restoreState();
    listSplitter->restoreState();

    // models
    invModel = new InventoryModel(this);

    connect(
        invModel, SIGNAL(workspaceCommandError(const QString &)),
        this, SLOT(workspaceCommandError(const QString &))
    );

    invWatcher = new InventoryWatcher(this);
    attrModel = new GetAttributes(this);

    proxyModelFolderTree = new InventoryProxyModel(this, true);
    proxyModelFileList = new InventoryProxyModel(this, false);

    connect(
        searchInput, SIGNAL(textChanged(const QString &)),
        proxyModelFolderTree, SLOT(filterItemsByName(const QString &))
    );

    connect(
        searchInput, SIGNAL(textChanged(const QString &)),
        proxyModelFileList, SLOT(filterItemsByName(const QString &))
    );

    proxyModelFolderTree->setSourceModel(invModel);
    proxyModelFileList->setSourceModel(invModel);

    connect(
        invModel, SIGNAL(pathsInserted(const QStringList &)),
        invWatcher, SLOT(watchPaths(const QStringList &))
    );

    connect(
        invModel, SIGNAL(pathsRemoved(const QStringList &)),
        invWatcher, SLOT(unwatchPaths(const QStringList &))
    );

    connect(
        invWatcher, SIGNAL(changedPath(const QString &)),
        invModel, SLOT(refresh(const QString &))
    );

    connect(
        dialogManager, SIGNAL(beforeWorkspaceUpdate()),
        invWatcher, SLOT(clearAllWatches())
    );

    connect(
        invModel, SIGNAL(invalidWorkspaceFormat()),
        this, SLOT(invalidWorkspaceFormat())
    );

    connect(
        dialogManager, SIGNAL(revisionCommitted(const QString &)),
        invModel, SLOT(refresh())
    );

    connect(
        dialogManager, SIGNAL(renamePath(const QString &, const QString &)),
        invModel, SLOT(renamePath(const QString &, const QString &))
    );

    connect(
        menuBar, SIGNAL(reloadWorkspace()),
        invModel, SLOT(refresh())
    );

    connect(
        menuBar, SIGNAL(hideIgnoredFiles(bool)),
        proxyModelFileList, SLOT(setHideIgnoredFiles(bool))
    );

    connect(
        menuBar, SIGNAL(hideIgnoredFiles(bool)),
        proxyModelFolderTree, SLOT(setHideIgnoredFiles(bool))
    );

    connect(
        menuBar, SIGNAL(filterFiles(int)),
        proxyModelFileList, SLOT(setViewOption(int))
    );

    connect(
        menuBar, SIGNAL(filterFiles(int)),
        proxyModelFolderTree, SLOT(setViewOption(int))
    );

    treeView->setModel(proxyModelFolderTree);
    treeView->setType(InventoryView::FolderTree);

    listView->setModel(proxyModelFileList);
    listView->setType(InventoryView::FileList);

    connect(
        treeView, SIGNAL(refreshPath(const QString &)),
        invModel, SLOT(refresh(const QString &))
    );

    connect(
        listView, SIGNAL(refreshPath(const QString &)),
        invModel, SLOT(refresh(const QString &))
    );

    connect(
        treeView, SIGNAL(addPaths(const QStringList &)),
        invModel, SLOT(addPaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(addPaths(const QStringList &)),
        invModel, SLOT(addPaths(const QStringList &))
    );

    connect(
        treeView, SIGNAL(dropPaths(const QStringList &)),
        invModel, SLOT(dropPaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(dropPaths(const QStringList &)),
        invModel, SLOT(dropPaths(const QStringList &))
    );

    connect(
        treeView, SIGNAL(revertPaths(const QStringList &)),
        invModel, SLOT(revertPaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(revertPaths(const QStringList &)),
        invModel, SLOT(revertPaths(const QStringList &))
    );

    connect(
        treeView, SIGNAL(ignorePaths(const QStringList &)),
        invModel, SLOT(ignorePaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(ignorePaths(const QStringList &)),
        invModel, SLOT(ignorePaths(const QStringList &))
    );

    connect(
        treeView, SIGNAL(unignorePaths(const QStringList &)),
        invModel, SLOT(unignorePaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(unignorePaths(const QStringList &)),
        invModel, SLOT(unignorePaths(const QStringList &))
    );

    connect(
        listView, SIGNAL(newFile(const QString &)),
        invModel, SLOT(newFile(const QString &))
    );

    connect(
        listView, SIGNAL(newDirectory(const QString &)),
        invModel, SLOT(newDirectory(const QString &))
    );

    attrView->setModel(attrModel);

    iconHelp = new IconHelp(this);
    iconHelp->toggleViewAction()->setShortcut(QKeySequence(tr("Shift+F1")));
    dynamic_cast<DatabaseMenuBar *>(menuBar)
        ->addDockWidgetAction(iconHelp->toggleViewAction());

    nodeInfo = new NodeInfo(this);
    nodeInfo->toggleViewAction()->setShortcut(QKeySequence(tr("Ctrl+I")));
    dynamic_cast<DatabaseMenuBar *>(menuBar)
        ->addDockWidgetAction(nodeInfo->toggleViewAction());

    addDockWidget(Qt::LeftDockWidgetArea, iconHelp);
    addDockWidget(Qt::LeftDockWidgetArea, nodeInfo);

    connect(
        nodeInfo, SIGNAL(showSelectRevision(const QString &)),
        dialogManager, SLOT(showSelectRevision(const QString &))
    );

    connect(
        dialogManager, SIGNAL(workspaceUpdated(const QString &)),
        this, SLOT(workspaceUpdated(const QString &))
    );
}

void WorkspaceWindow::load(const MonotoneHandlePtr & handle)
{
    I(handle->getType() == MonotoneHandle::workspace_handle);

    APP->manager()->getThread(handle);

    setWindowTitle(
        tr("%1 - workspace mode - guitone").
        arg(MonotoneUtil::getBranchNameShort(handle))
    );

    dialogManager->init(handle);

    invModel->setMonotoneHandle(handle);
    invWatcher->setMonotoneHandle(handle);
    attrModel->setMonotoneHandle(handle);
    nodeInfo->setMonotoneHandle(handle);
    databaseVariables->setMonotoneHandle(handle);

    // read the inventory
    invModel->refresh();
}

void WorkspaceWindow::openFile(const QString & filePath)
{
    QString completePath = monotoneHandle->getData() + "/" + filePath;
    QFileInfo fileInfo(completePath);

    if (!fileInfo.exists())
    {
        W(QString("cannot open %1, file does not exist").arg(completePath));
        return;
    }

    if (Settings::getBool("AskFileOpen"))
    {
        QString suffix = fileInfo.completeSuffix();
        QStringList harmfulSuffixes =
            Settings::getString("AskFileOpenExt").split(",", QString::SkipEmptyParts);

        if (!fileInfo.isDir() && (fileInfo.isExecutable() || harmfulSuffixes.contains(suffix)))
        {
            QMessageBox::StandardButton btn = QMessageBox::question(
                NULL,
                QObject::tr("Attention"),
                QObject::tr("The file '%1' you're trying to open is either "
                            "executable or included in the list of "
                            "precarious file formats - do you really "
                            "want to continue?").arg(completePath),
                QMessageBox::Yes | QMessageBox::No
            );

            if (btn == QMessageBox::No)
                return;
        }
    }

    try
    {
        Platform::openFile(completePath);
    }
    catch (GuitoneException & e)
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Could not open '%1': %2").arg(completePath).arg(e.what()),
            QMessageBox::Ok, 0, 0
        );
    }
}

void WorkspaceWindow::deleteLocalPath(const QString & path)
{
    QString completePath = monotoneHandle->getData() + "/" + path;

    QMessageBox::StandardButton btn = QMessageBox::question(
        this,
        tr("Delete confirmation"),
        tr("Do you really want to delete '%1'? "
           "This action cannot be undone.").arg(completePath),
        QMessageBox::Yes | QMessageBox::No
    );

    if (btn == QMessageBox::No) return;

    try
    {
        Platform::removeRecursively(completePath);
    }
    catch (GuitoneException & e)
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Could not remove '%1': %2").arg(completePath).arg(e.what()),
            QMessageBox::Ok, 0, 0
        );
    }
}

void WorkspaceWindow::maybeReadNodeInfo(const QModelIndexList & indexes)
{
    attrModel->revert();
    nodeInfo->resetInfo();

    if (indexes.size() != 1)
    {
        D("multiple or no items selected");
        return;
    }

    QModelIndex sourceIndex = static_cast<const QSortFilterProxyModel *>
        (indexes.at(0).model())->mapToSource(indexes.at(0));
    ModelItem * item = static_cast<ModelItem *>
        (sourceIndex.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    if (!invitem)
    {
        D("selected item is not inventory item");
        return;
    }

    nodeInfo->setItem(invitem);
    if (nodeInfo->isVisible())
        nodeInfo->readAndSetInfo();

    if (invitem->isOldNode() || !invitem->isTracked())
    {
        D("selected item is old node or not tracked");
        return;
    }

    attrModel->readAttributes(invitem->getPath());
}

void WorkspaceWindow::invalidWorkspaceFormat()
{
    QMessageBox::critical(
        this,
        tr("Error"),
        tr("Unable to read the workspace. Please check the log for details."),
        QMessageBox::Ok, 0, 0
    );
    close();
}

void WorkspaceWindow::workspaceUpdated(const QString & rev)
{
    Q_UNUSED(rev);

    // monotone tends to write back old data to _MTN/options, so we
    // need to ensure that the new branch set by mtn update before
    // doesn't get overwritten by any next workspace command issued
    // over stdio
    APP->manager()->stopThreads(monotoneHandle);

    setWindowTitle(
        tr("%1 - workspace mode - guitone").
        arg(MonotoneUtil::getBranchNameShort(monotoneHandle))
    );

    invModel->refresh();
}

void WorkspaceWindow::workspaceCommandError(const QString & error)
{
    QMessageBox::critical(
        this,
        tr("Error"),
        tr("Unable to execute workspace action: %1").arg(error),
        QMessageBox::Ok, 0, 0
    );
}

