/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DatabaseVariables.h"
#include "AddEditDatabaseVariable.h"

#include <QMessageBox>

DatabaseVariables::DatabaseVariables(QWidget * parent)
    : QDockWidget(parent)
{
    setVisible(false);

    QWidget * parentWidget = new QWidget(this);
    setupUi(parentWidget);
    setWidget(parentWidget);

    databaseVariables = new GetDatabaseVariables(this);
    variableTree->setModel(databaseVariables);

    connect(
        databaseVariables, SIGNAL(variablesRead()),
        this, SLOT(variablesRead())
    );

    connect(
        updateVariables, SIGNAL(clicked()),
        databaseVariables, SLOT(readVariables())
    );

    connect(
        addVariable, SIGNAL(clicked()),
        this, SLOT(openVariableEditor())
    );

    connect(
        removeVariable, SIGNAL(clicked()),
        this, SLOT(removeCurrentVariable())
    );

    connect(
        variableTree, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(enableDisableDropButton(const QModelIndex &))
    );

    connect(
        variableTree, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(openVariableEditor(const QModelIndex &))
    );

    setWindowTitle(tr("Database variables"));
    setFloating(true);
    setAllowedAreas(Qt::BottomDockWidgetArea);
}

DatabaseVariables::~DatabaseVariables()
{
    delete databaseVariables;
}

void DatabaseVariables::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    databaseVariables->setMonotoneHandle(handle);
    // FIXME: this is hacky, but we can't properly determine when to
    // read the variables other than here or on every view change
    // (see QDockWidget::visibilityChanged(bool) for that purpose), but
    // the latter is not working properly as well because the widget
    // might be made visible / displayed before the database/workspace
    // path is set and we have no other signal floating around than
    // this one here to determine when we actually _can_ start to read
    // them
    databaseVariables->readVariables();
}

/**
 * This is an ugly hack because Qt does not figure out column spanning
 * itself when columnCount() returns different values for this tree view
 */
void DatabaseVariables::variablesRead()
{
    QModelIndex parent;

    int count = databaseVariables->rowCount(parent);
    for (int row=0; row<count; row++)
    {
        variableTree->setFirstColumnSpanned(row, parent, true);
    }

    variableTree->expandAll();
}

void DatabaseVariables::openVariableEditor(const QModelIndex & index)
{
    if (!index.isValid() || !index.parent().isValid()) return;

    QString domain, key, value;

    domain = index.parent().data().toString();

    if (index.column() == 0)
    {
        key = index.data().toString();
        value = index.sibling(index.row(), 1).data().toString();
    }
    else
    if (index.column() == 1)
    {
        value = index.data().toString();
        key = index.sibling(index.row(), 0).data().toString();
    }
    else
        I(false);

    openVariableEditor(domain, key, value);
}

void DatabaseVariables::openVariableEditor(const QString & domain,
                                           const QString & key,
                                           const QString & value)
{
    AddEditDatabaseVariable dlg(this, domain, key, value);

    if (dlg.exec() == QDialog::Accepted)
    {
        if (!databaseVariables->setVariable(dlg.getDomain(), dlg.getKey(), dlg.getValue()))
        {
            C("could not store new database variable");
        }
        databaseVariables->readVariables();
    }
}

void DatabaseVariables::removeCurrentVariable()
{
    QModelIndex index = variableTree->currentIndex();

    if (!index.isValid() || !index.parent().isValid()) return;

    QString domain, key;

    domain = index.parent().data().toString();

    if (index.column() == 0)
    {
        key = index.data().toString();
    }
    else
    if (index.column() == 1)
    {
        key = index.sibling(index.row(), 0).data().toString();
    }
    else
        I(false);

    QMessageBox msgBox(this);
    msgBox.setIcon(QMessageBox::Question);
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::Yes);
    msgBox.setWindowTitle(tr("Attention"));
    msgBox.setText(tr("Are you sure you want to remove the "
                      "variable '%1' from the domain '%2'?").arg(key).arg(domain));

    if (msgBox.exec() == QMessageBox::Yes)
    {
        if (!databaseVariables->dropVariable(domain, key))
        {
            C("could not remove database variable");
        }
        databaseVariables->readVariables();
    }
}

void DatabaseVariables::enableDisableDropButton(const QModelIndex & index)
{
    // only enable the children, i.e. single entries of a particular domain
    removeVariable->setEnabled(index.isValid() && index.parent().isValid());
}

