/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "NodeInfo.h"
#include "MonotoneUtil.h"
#include "CertList.h"

#include <QFileInfo>
#include <QDateTime>

NodeInfo::NodeInfo(QWidget * parent)
 : QDockWidget(parent), inventoryItem(0)
{
    setVisible(false);

    QWidget * parentWidget = new QWidget(this);
    setupUi(parentWidget);
    setWidget(parentWidget);

    connect(
        fileBirthRev, SIGNAL(linkActivated(const QString &)),
        this, SIGNAL(showSelectRevision(const QString &))
    );

    connect(
        lastChangeRevision, SIGNAL(linkActivated(const QString &)),
        this, SIGNAL(showSelectRevision(const QString &))
    );
    resetInfo();

    setWindowTitle(tr("Node information"));
    setFloating(true);
    setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
}

NodeInfo::~NodeInfo() {}

void NodeInfo::setMonotoneHandle(const MonotoneHandlePtr & handle)
{
    monotoneHandle = handle;
}

void NodeInfo::setItem(const InventoryItem * item)
{
    inventoryItem = item;
}

void NodeInfo::readAndSetInfo()
{
    if (inventoryItem == 0) return;

    I(!monotoneHandle.isNull() &&
      monotoneHandle->getType() == MonotoneHandle::workspace_handle);

    QString strLastChangeRevision("-"),
            strLastChangeAuthor("-"),
            strLastChangeDate("-"),
            strLastChangelog("");

    if (inventoryItem->isTracked() &&
        inventoryItem->hasNotStatus(InventoryItem::Added) &&
        !inventoryItem->isDirectory())
    {
        // if this is a renamed node, it can't be found under its new
        // new name in the base revision of the workspace
        QString path = inventoryItem->getPath();
        if (inventoryItem->hasStatus(InventoryItem::RenameTarget))
        {
            path = inventoryItem->getRenameSource();
        }

        QStringList revs = MonotoneUtil::getPreviousContentMarks(
            monotoneHandle, path,
            MonotoneUtil::getBaseWorkspaceRevision(monotoneHandle)
        );

        I(revs.size() > 0);

        // FIXME: for now we only handle the first returned content
        // mark; its not clear if merges would really pop up here
        if (revs.size() > 1)
        {
            W(QString("multiple content marks returned, only using the first"));
        }

        strLastChangeRevision =  QObject::tr("<a href=\"%1\">%2...</a>")
                            .arg(revs.at(0))
                            .arg(revs.at(0).left(12));

        CertList certs = MonotoneUtil::getCerts(monotoneHandle, revs.at(0));

        strLastChangeAuthor   = certs.findByNameValues("author").join(", ");

        QList<Cert> changelogCerts = certs.findByName("changelog");
        QStringList changelogs;
        foreach (const Cert & cert, changelogCerts)
        {
            changelogs.push_back(cert.htmlValue());
        }
        strLastChangelog = changelogs.join("\n<hr>\n");

        QList<Cert> dateCerts = certs.findByName("date");
        QStringList dates;
        foreach (const Cert & cert, dateCerts)
        {
            dates.push_back(
                cert.dateTimeValue().toString(Qt::DefaultLocaleShortDate)
            );
        }
        strLastChangeDate = dates.join(", ");
    }

    lastChangeRevision->setText(strLastChangeRevision);
    lastChangeAuthor->setText(strLastChangeAuthor);
    lastChangeDate->setText(strLastChangeDate);
    lastChangeLog->setHtml(strLastChangelog);

    QString strFileSize("-"),
            strFileLastModified("-"),
            strFileOwner("-"),
            strFileGroup("-"),
            strFilePermissions("-"),
            strFileSymlinkTarget("-");

    if (inventoryItem->getFSType() != InventoryItem::None)
    {
        QFileInfo fileInfo(
            monotoneHandle->getData() + "/" + inventoryItem->getPath()
        );
        I(fileInfo.exists());

        strFileSize = QString("%1 KB").arg(fileInfo.size() / 1024);
        // FIXME: use Qt::DefaultLocaleShortDate if support for Qt 4.3 is dropped
        strFileLastModified = fileInfo.lastModified().toString(Qt::LocaleDate);

#ifdef Q_OS_UNIX
        strFileOwner = QString("%1 (UID %2)").arg(fileInfo.owner()).arg(fileInfo.ownerId());
        strFileGroup = QString("%1 (GID %2)").arg(fileInfo.group()).arg(fileInfo.groupId());
        QFile::Permissions perms = fileInfo.permissions();
        strFilePermissions = QString("%1%2%3 %4%5%6 %7%8%9")
            .arg((perms & QFile::ReadUser)   == QFile::ReadUser ? "r" : "-")
            .arg((perms & QFile::WriteUser)  == QFile::WriteUser ? "w" : "-")
            .arg((perms & QFile::ExeUser)    == QFile::ExeUser ? "x" : "-")
            .arg((perms & QFile::ReadGroup)  == QFile::ReadGroup ? "r" : "-")
            .arg((perms & QFile::WriteGroup) == QFile::WriteGroup ? "w" : "-")
            .arg((perms & QFile::ExeGroup)   == QFile::ExeGroup ? "x" : "-")
            .arg((perms & QFile::ReadOther)  == QFile::ReadOther ? "r" : "-")
            .arg((perms & QFile::WriteOther) == QFile::WriteOther ? "w" : "-")
            .arg((perms & QFile::ExeOther)   == QFile::ExeOther ? "x" : "-");
#endif

        if (fileInfo.isSymLink())
            strFileSymlinkTarget =  fileInfo.symLinkTarget();
    }

    fileSize->setText(strFileSize);
    fileLastModified->setText(strFileLastModified);
    fileOwner->setText(strFileOwner);
    fileGroup->setText(strFileGroup);
    filePermissions->setText(strFilePermissions);
    fileSymlinkTarget->setText(strFileSymlinkTarget);

    filePath->setText(inventoryItem->getPath());
    fileStatus->setText(inventoryItem->getStatusString());

    QString renameSource = inventoryItem->getRenameSource();
    if (renameSource.isEmpty()) renameSource = "-";
    fileOldPath->setText(renameSource);

    QString renameTarget = inventoryItem->getRenameTarget();
    if (renameTarget.isEmpty()) renameTarget = "-";
    fileNewPath->setText(renameTarget);

    QString birthRev = inventoryItem->getBirthRevision();
    if (!birthRev.isEmpty())
        fileBirthRev->setText(tr("<a href=\"%1\">%2...</a>").
                              arg(birthRev).
                              arg(birthRev.left(12)));
    else
        fileBirthRev->setText("-");

    fileBaseType->setText(InventoryItem::translateFileType(inventoryItem->getOldType()));
    fileNewType->setText(InventoryItem::translateFileType(inventoryItem->getNewType()));
    fileDiskType->setText(InventoryItem::translateFileType(inventoryItem->getFSType()));
}

void NodeInfo::resetInfo()
{
    inventoryItem = 0;

    lastChangeRevision->setText("-");
    lastChangeAuthor->setText("-");
    lastChangeDate->setText("-");
    lastChangeLog->setPlainText("");

    filePath->setText("-");
    fileStatus->setText("-");
    fileOldPath->setText("-");
    fileNewPath->setText("-");
    fileBirthRev->setText("-");
    fileBaseType->setText("-");
    fileNewType->setText("-");
    fileDiskType->setText("-");

    fileLastModified->setText("-");
    fileSize->setText("-");
    fileOwner->setText("-");
    fileGroup->setText("-");
    filePermissions->setText("-");
    fileSymlinkTarget->setText("-");
}

void NodeInfo::showEvent(QShowEvent * ev)
{
    Q_UNUSED(ev);
    readAndSetInfo();
}

