/***************************************************************************
 *   Copyright (C) 2010 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CertListBrowser.h"
#include "MonotoneUtil.h"

CertListBrowser::CertListBrowser(QWidget * parent)
  : QTextBrowser(parent)
{
    nameI18NMap.insert("author",     tr("Author"));
    nameI18NMap.insert("date",       tr("Date"));
    nameI18NMap.insert("branch",     tr("Branch"));
    nameI18NMap.insert("tag",        tr("Tag"));
    nameI18NMap.insert("changelog",  tr("Changelog"));
    nameI18NMap.insert("comment",    tr("Comment"));
    nameI18NMap.insert("testresult", tr("Test result"));
    nameI18NMap.insert("suspend",    tr("Suspended branch"));

    // the order is important here!
    certFormat.append(qMakePair(QString("date"),       table));
    certFormat.append(qMakePair(QString("author"),     table));
    certFormat.append(qMakePair(QString("branch"),     table));
    certFormat.append(qMakePair(QString("tag"),        table));
    certFormat.append(qMakePair(QString("suspend"),    table));
    certFormat.append(qMakePair(QString("testresult"), table));
    certFormat.append(qMakePair(QString("changelog"),  definition));
    certFormat.append(qMakePair(QString("comment"),    definition));

    certsPerTableRow = 2;
}

CertListBrowser::~CertListBrowser() {}

void CertListBrowser::setKeyMap(const QMap<QString, QString> & map)
{
    keyMap = map;
}

void CertListBrowser::setCertList(const CertList & list)
{
    QString html =
        "<style type=\"text/css\">"
            "h1 { font-size: medium; } "
            "dl { margin-top: 10px; margin-left: 10px; } "
            "dl dt { margin-left: 3px; } "
            "ul {margin-left: -40px; } "
            "table { margin-top: 10px; margin-left: 10px; } "
            "table th { font-weight: normal; } "
            "table td { padding-left: 10px; padding-right: 30px; } "
            "table th, table td { vertical-align: middle; } "
            "table th, dl dt { font-size: small; font-weight: bold; } "
        "</style>"
    ;

    QStringList signers = list.signers();

    foreach (const QString & signer, signers)
    {
        QString certHtml, signerHtml;

        if (keyMap.contains(signer))
            signerHtml = tr("<h1>Signer: %1</h1>").arg(keyMap.value(signer));
        else
            signerHtml = tr("<h1>Signer: %1</h1>").arg(signer);

        QMap<QString, QStringList> formattedCertValues;

        foreach (const Cert & cert, list.findBySigner(signer))
        {
            QString value;
            if (cert.name == "date")
            {
                value = cert.dateTimeValue().toString(
                    Qt::DefaultLocaleShortDate
                );
            }
            else if (cert.name == "changelog" ||
                     cert.name == "comment")
            {
                value = cert.htmlValue();
            }
            else
            {
                value = cert.value;
                value.replace("<", "&lt;");
            }

            formattedCertValues[cert.name].push_back(value);
        }

        format_type current = none;
        int certCounter = 0;

        for (int i=0; i<certFormat.size(); ++i)
        {
            QString certName = certFormat.at(i).first;
            format_type next = certFormat.at(i).second;

            if (!formattedCertValues.contains(certName))
                continue;

            if (next != current || current == none)
            {
                if (current == table)
                    certHtml += "</tr></table>";
                else if (current == definition)
                    certHtml += "</dl>";

                if (next == table)
                    certHtml += "<table><tr>";
                else if (next == definition)
                    certHtml += "<dl>";

                current = next;
                certCounter = 0;
            }

            QString format;
            if (current == table)
            {
                if (certCounter % certsPerTableRow == 0 && certCounter != 0)
                    certHtml += "</tr><tr>";
                format = QString("<th align=\"right\">%1</th><td>%2</td>");
            }
            else
            {
                format = QString("<dt>%1</dt><dd>%2</dd>");
            }

            certHtml += format.arg(nameI18NMap.value(certName))
                              .arg(formattedCertValues.value(certName).join(", "));

            formattedCertValues.remove(certName);

            certCounter++;
        }

        if (formattedCertValues.size() > 0)
        {
            if (current != table)
            {
                if (current == definition)
                    certHtml += "</dl>";

                certHtml += "<table><tr>";
                current = table;
                certCounter = 0;
            }

            QMapIterator<QString, QStringList> it(formattedCertValues);
            while (it.hasNext())
            {
                it.next();
                if (certCounter % certsPerTableRow == 0 && certCounter != 0)
                    certHtml += "</tr><tr>";
                certHtml += QString("<th>%1</th><td>%2</td>")
                    .arg(it.key())
                    .arg(formattedCertValues.value(it.key()).join(", "));
            }
        }

        if (current == table)
        {
            certHtml += "</tr></table>";
        }
        else if (current == definition)
        {
            certHtml += "</dl>";
        }

        html += signerHtml + certHtml;
    }

    setHtml(html);
}

