/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "ChangeLogEdit.h"
#include "vocab.h"

#include <QStringListModel>
#include <QScrollBar>

ChangeLogEdit::ChangeLogEdit(QWidget * parent)
  : QTextEdit(parent), completionEnabled(false)
{
    completer = new QCompleter();
    completer->setWidget(this);
    completer->setCompletionMode(QCompleter::PopupCompletion);
    completer->setCaseSensitivity(Qt::CaseInsensitive);
    completer->setModelSorting(QCompleter::CaseInsensitivelySortedModel);
    completer->setWrapAround(false);

    connect(
        completer, SIGNAL(activated(const QString &)),
        this, SLOT(insertCompletion(const QString &))
    );
}

ChangeLogEdit::~ChangeLogEdit()
{
    delete completer;
}

void ChangeLogEdit::setCompleterStrings(const QStringList & strings)
{
    if (completer->model())
    {
        delete completer->model();
    }

    if (strings.size() > 0)
    {
        completer->setModel(new QStringListModel(strings, completer));
        completionEnabled = true;
    }
}

void ChangeLogEdit::insertCompletion(const QString & completion)
{
    QTextCursor tc = textCursor();
    int extra = completion.length() - completer->completionPrefix().length();

    tc.movePosition(QTextCursor::Left);
    tc.movePosition(QTextCursor::EndOfWord);
    tc.insertText(completion.right(extra));

    setTextCursor(tc);
}

QString ChangeLogEdit::textUnderCursor() const
{
    QTextCursor tc = textCursor();
    tc.select(QTextCursor::WordUnderCursor);
    return tc.selectedText();
}

void ChangeLogEdit::keyPressEvent(QKeyEvent * e)
{
    if (!completionEnabled)
    {
        QTextEdit::keyPressEvent(e);
        return;
    }

    if (completer->popup()->isVisible())
    {
        // The following keys are forwarded by the completer to the widget
        switch (e->key())
        {
            case Qt::Key_Enter:
            case Qt::Key_Return:
            case Qt::Key_Escape:
            case Qt::Key_Tab:
            case Qt::Key_Backtab:
                // let the completer do default behavior
                e->ignore();
                return;
            default:
                break;
        }
    }

    bool isShortcut = (e->modifiers() & Qt::ControlModifier) &&
                       e->key() == Qt::Key_E;

    // dont process the shortcut when we have a completer
    if (!isShortcut)
        QTextEdit::keyPressEvent(e);

    const bool ctrlOrShift = e->modifiers() & (Qt::ControlModifier | Qt::ShiftModifier);

    static QString eow("~!@#$%^&*()_+{}|:\"<>?,./;'[]\\-="); // end of word
    bool hasModifier = (e->modifiers() != Qt::NoModifier) && !ctrlOrShift;

    QString completionPrefix = textUnderCursor();

    if (!isShortcut && (
         hasModifier ||
         e->text().isEmpty() ||
         completionPrefix.length() < 3 ||
         eow.contains(e->text().right(1))
        ))
    {
        completer->popup()->hide();
        return;
    }

    if (completionPrefix != completer->completionPrefix()) {
        completer->setCompletionPrefix(completionPrefix);
        completer->popup()->setCurrentIndex(completer->completionModel()->index(0, 0));
    }

    QRect cr = cursorRect();
    cr.setWidth(completer->popup()->sizeHintForColumn(0) +
                completer->popup()->verticalScrollBar()->sizeHint().width());

    // pop it up!
    completer->complete(cr);
}
