/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "InventoryView.h"
#include "Settings.h"
#include "InventoryItem.h"
#include "vocab.h"

#include <QMenu>
#include <QHeaderView>
#include <QSortFilterProxyModel>
#include <QMessageBox>
#include <QTimer>

InventoryView::InventoryView(QWidget * parent, const QString & objectName)
  : TreeView(parent, objectName), invViewDelegate(parent)
{
    setItemDelegate(&invViewDelegate);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setDragEnabled(true);
    setAcceptDrops(true);
    setDropIndicatorShown(true);
    setSortingEnabled(true);
    sortByColumn(0, Qt::AscendingOrder);

    header()->setClickable(true);

    createAndConnectContextActions();

    connect(
        this, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(slotContextMenuRequested(const QModelIndexList &, const QPoint &))
    );

    // this is a hack: if we connect the model's layoutChanged() signal
    // directly with a slot in the view, actions like setting the root
    // index of an item or expanding it won't be processed, probably
    // because QAbstractItemView also listens to the event and processes
    // it after our slot is done. So we're doing our work within the next
    // event loop round
    timer = new QTimer();
    timer->setSingleShot(true);
    timer->setInterval(0);

    connect(
        timer, SIGNAL(timeout()),
        this, SLOT(delayedLayoutChanged())
    );
}

QModelIndex InventoryView::sourceIndex(const QModelIndex & proxy, bool useIndexModel) const
{
    if (useIndexModel)
        return static_cast<const QSortFilterProxyModel *>(proxy.model())->mapToSource(proxy);
    else
        return static_cast<const QSortFilterProxyModel *>(model())->mapToSource(proxy);
}

QModelIndex InventoryView::proxyIndex(const QModelIndex & source) const
{
    return static_cast<const QSortFilterProxyModel *>(model())->mapFromSource(source);
}

const ModelItem * InventoryView::modelItem(const QModelIndex & source)
{
    // ensure we got a source model item here
    I(qobject_cast<const QAbstractProxyModel *>(source.model()) == 0);
    return static_cast<const ModelItem *>(source.internalPointer());
}

void InventoryView::setType(Type t)
{
    // ensure that a model is already set, otherwise the app crashes
    I(model());

    type = t;

    if (type == FileList)
    {
        disconnect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );

        connect(
            this, SIGNAL(doubleClicked(const QModelIndex &)),
            this, SLOT(itemDoubleClicked(const QModelIndex &))
        );

        setRootIsDecorated(false);
        setItemsExpandable(false);
    }
    else
    {
        disconnect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemDoubleClicked(const QModelIndex &))
        );

        connect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );
        setRootIsDecorated(true);
        setItemsExpandable(true);
    }
}

void InventoryView::createAndConnectContextActions()
{
    actChdir = new QAction(tr("Go into"), this);
    actChdir->setStatusTip(tr("Go into the directory"));
    connect(actChdir, SIGNAL(triggered()), this, SLOT(slotChdir()));

    actOpen = new QAction(tr("Open"), this);
    actOpen->setStatusTip(tr("Open in default program"));
    connect(actOpen, SIGNAL(triggered()), this, SLOT(slotOpen()));

    actAdd = new QAction(tr("Add"), this);
    actAdd->setStatusTip(tr("Add to workspace"));
    connect(actAdd, SIGNAL(triggered()), this, SLOT(slotAdd()));

    actRemove = new QAction(tr("Remove"), this);
    actRemove->setStatusTip(tr("Remove from workspace"));
    connect(actRemove, SIGNAL(triggered()), this, SLOT(slotRemove()));

    actCommit = new QAction(tr("Commit"), this);
    actCommit->setStatusTip(tr("Commit"));
    connect(actCommit, SIGNAL(triggered()), this, SLOT(slotCommit()));

    actIgnore = new QAction(tr("Ignore"), this);
    actIgnore->setStatusTip(tr("Ignore file"));
    connect(actIgnore, SIGNAL(triggered()), this, SLOT(slotIgnore()));

    actUnignore = new QAction(tr("Unignore"), this);
    actUnignore->setStatusTip(tr("Unignore file"));
    connect(actUnignore, SIGNAL(triggered()), this, SLOT(slotUnignore()));

    actRevert = new QAction(tr("Revert"), this);
    actRevert->setStatusTip(tr("Revert uncommitted changes"));
    connect(actRevert, SIGNAL(triggered()), this, SLOT(slotRevert()));

    actFileDiff = new QAction(tr("Diff"), this);
    actFileDiff->setStatusTip(tr("Diff against base revision"));
    connect(actFileDiff, SIGNAL(triggered()), this, SLOT(slotFileDiff()));

    actFileHistory = new QAction(tr("History"), this);
    actFileHistory->setStatusTip(tr("Display the history of this file"));
    connect(actFileHistory, SIGNAL(triggered()), this, SLOT(slotFileHistory()));

    actRevisionDiff = new QAction(tr("Diff all"), this);
    actRevisionDiff->setStatusTip(tr("Show all differences"));
    connect(actRevisionDiff, SIGNAL(triggered()), this, SLOT(slotRevisionDiff()));

    actAnnotate = new QAction(tr("Annotate"), this);
    actAnnotate->setStatusTip(tr("Annotates the contents of this file"));
    connect(actAnnotate, SIGNAL(triggered()), this, SLOT(slotAnnotate()));

    actRename = new QAction(tr("Rename"), this);
    actRename->setStatusTip(tr("Rename file"));
    connect(actRename, SIGNAL(triggered()), this, SLOT(slotRename()));

    actRefresh = new QAction(tr("Refresh"), this);
    actRefresh->setStatusTip(tr("Re-reads the item and refeshes the view"));
    connect(actRefresh, SIGNAL(triggered()), this, SLOT(slotRefresh()));

    actNewFile = new QAction(tr("New file"), this);
    actNewFile->setStatusTip(tr("Creates a new versioned file"));
    connect(actNewFile, SIGNAL(triggered()), this, SLOT(slotNewFile()));

    actNewDirectory = new QAction(tr("New directory"), this);
    actNewDirectory->setStatusTip(tr("Creates a new versioned directory"));
    connect(actNewDirectory, SIGNAL(triggered()), this, SLOT(slotNewDirectory()));

    actDelete = new QAction(tr("Delete"), this);
    actDelete->setStatusTip(tr("Deletes the file from the filesystem"));
    connect(actDelete, SIGNAL(triggered()), this, SLOT(slotDelete()));

    actAddMultiple = new QAction(tr("Add %1 items"), this);
    actAddMultiple->setStatusTip(tr("Add multiple items"));
    connect(actAddMultiple, SIGNAL(triggered()), this, SLOT(slotAdd()));

    actRemoveMultiple = new QAction(tr("Remove %1 items"), this);
    actRemoveMultiple->setStatusTip(tr("Remove multiple items"));
    connect(actRemoveMultiple, SIGNAL(triggered()), this, SLOT(slotRemove()));

    actCommitMultiple = new QAction(tr("Commit %1 items"), this);
    actCommitMultiple->setStatusTip(tr("Commit multiple items"));
    connect(actCommitMultiple, SIGNAL(triggered()), this, SLOT(slotCommit()));

    actIgnoreMultiple = new QAction(tr("Ignore %1 items"), this);
    actIgnoreMultiple->setStatusTip(tr("Ignore multiple items"));
    connect(actIgnoreMultiple, SIGNAL(triggered()), this, SLOT(slotIgnore()));

    actUnignoreMultiple = new QAction(tr("Unignore %1 items"), this);
    actUnignoreMultiple->setStatusTip(tr("Unignore multiple items"));
    connect(actUnignoreMultiple, SIGNAL(triggered()), this, SLOT(slotUnignore()));

    actRevertMultiple = new QAction(tr("Revert %1 items"), this);
    actRevertMultiple->setStatusTip(tr("Revert multiple items"));
    connect(actRevertMultiple, SIGNAL(triggered()), this, SLOT(slotRevert()));
}

InventoryView::~InventoryView()
{
    delete actChdir;
    delete actOpen;
    delete actAdd;
    delete actRemove;
    delete actCommit;
    delete actIgnore;
    delete actUnignore;
    delete actRevert;
    delete actFileDiff;
    delete actFileHistory;
    delete actRevisionDiff;
    delete actAnnotate;
    delete actRename;
    delete actRefresh;
    delete actNewFile;
    delete actNewDirectory;
    delete actDelete;
    delete actAddMultiple;
    delete actRemoveMultiple;
    delete actCommitMultiple;
    delete actIgnoreMultiple;
    delete actUnignoreMultiple;
    delete actRevertMultiple;
}

// ensure that we only accept sortfilterproxymodels
void InventoryView::setModel(QAbstractItemModel * model)
{
    Q_UNUSED(model);
    I(false);
}

void InventoryView::setModel(QSortFilterProxyModel * newModel)
{
    QAbstractItemModel * oldModel = model();
    if (oldModel)
    {
        disconnect(
            oldModel, SIGNAL(rowsInserted(const QModelIndex &, int, int)),
            this, SLOT(rowsInserted(QModelIndex &, int, int))
        );

        disconnect(
            oldModel, SIGNAL(layoutChanged()),
            timer, SLOT(start())
        );
    }

    TreeView::setModel(newModel);

    connect(
        newModel, SIGNAL(rowsInserted(const QModelIndex &, int, int)),
        this, SLOT(rowsInserted(const QModelIndex &, int, int))
    );

    connect(
        newModel, SIGNAL(layoutChanged()),
        timer, SLOT(start())
    );
}

// adds elements to the popup menu based on the selected items
// since not all actions may apply on all items
void InventoryView::slotContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    //
    // no selected items - just show the refresh entry
    //
    if (indexList.size() == 0)
    {
        QMenu menu(this);

        menu.addAction(actRefresh);
        QFont font;
        font.setBold(true);
        actRefresh->setFont(font);

        if (type == FileList)
        {
            const ModelItem * item = modelItem(sourceIndex(rootIndex()));
            const InventoryItem * invItem = qobject_cast<const InventoryItem *>(item);
            if (invItem && invItem->isNewNode())
            {
                menu.addSeparator();
                menu.addAction(actNewFile);
                menu.addAction(actNewDirectory);
            }
        }

        menu.exec(pos);

        actRefresh->setFont(QFont());
        return;
    }

    //
    // single item selected
    //
    if (indexList.size() == 1)
    {
        const ModelItem * item = modelItem(sourceIndex(indexList[0]));
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>
            (item);

        // only display the popup for real workspace items
        if (!invitem)
            return;

        QMenu menu(this);

        if (invitem->isDirectory())
        {
            menu.addAction(actChdir);
        }

        if (invitem->hasStatus(InventoryItem::Unknown) ||
            invitem->hasStatus(InventoryItem::Ignored) || (
               invitem->isNewNode() &&
               invitem->hasNotStatus(InventoryItem::Missing) &&
               invitem->hasNotStatus(InventoryItem::Invalid)
           ))
        {
            menu.addAction(actOpen);
            menu.addSeparator();
        }

        if (invitem->hasStatus(InventoryItem::Unknown))
        {
            menu.addAction(actAdd);
            menu.addAction(actIgnore);
            menu.addSeparator();
            menu.addAction(actDelete);
        }

        if (invitem->hasStatus(InventoryItem::Ignored))
        {
            menu.addAction(actUnignore);
            menu.addSeparator();
            menu.addAction(actDelete);
        }

        if (invitem->hasChangedRecursive())
        {
            menu.addAction(actCommit);

            if (invitem->isDirectory() &&
                invitem->getFSType() == InventoryItem::Directory)
            {
                menu.addAction(actRevisionDiff);
            }
        }

        // FIXME: we currently exclude the workspace root for this.
        // While its reasonable for the remove action, the root could
        // actually be renamed by pivot_root - decide whether we want
        // to implement that here or offer a workspace menu and dialog
        // for this rather rare use case
        if (invitem->hasNotStatus(InventoryItem::Ignored) &&
            invitem->hasNotStatus(InventoryItem::Unknown) &&
            !invitem->getPath().isEmpty())
        {
            if (invitem->isNewNode())
            {
                menu.addAction(actRemove);
                menu.addAction(actRename);
                menu.addSeparator();
            }

            // as long as it is not unchanged, it can be reverted and diffed
            if (invitem->hasChanged())
            {
                menu.addAction(actRevert);
                menu.addSeparator();

                if (invitem->hasStatus(InventoryItem::ContentsChanged) &&
                    invitem->hasNotStatus(InventoryItem::Added))
                {
                    menu.addAction(actFileDiff);
                }
            }

            if (invitem->getFSType() == InventoryItem::File &&
                invitem->isNewNode() &&
                invitem->hasNotStatus(InventoryItem::Added))
            {
                menu.addAction(actFileHistory);
                menu.addAction(actAnnotate);
            }
        }

        //
        // mark the default action with bold text
        //
        DefaultAction act = getDefaultAction(indexList[0]);
        QAction * defaultAction = 0;

        switch (act)
        {
            case Chdir:
                defaultAction = actChdir;
                break;
            case FileDiff:
                defaultAction = actFileDiff;
                break;
            case Commit:
                defaultAction = actCommit;
                break;
            case Open:
                defaultAction = actOpen;
                break;
            default:
                break;
        }

        if (defaultAction)
        {
            QFont font;
            font.setBold(true);
            defaultAction->setFont(font);
        }

        //
        // add the refresh action
        //
        menu.addSeparator();
        menu.addAction(actRefresh);

        menu.exec(pos);

        if (defaultAction)
        {
            defaultAction->setFont(QFont());
        }

        return;
    }

    //
    // multiple items selected
    //
    const ModelItem * item;
    const InventoryItem * invitem;

    QStringList actions;
    actions << "add" << "drop" << "commit"
            << "ignore" << "unignore" << "revert";

    QMap<QString, QAction *> actionMapping;
    actionMapping.insert("add", actAddMultiple);
    actionMapping.insert("drop", actRemoveMultiple);
    actionMapping.insert("commit", actCommitMultiple);
    actionMapping.insert("revert", actRevertMultiple);
    actionMapping.insert("ignore", actIgnoreMultiple);
    actionMapping.insert("unignore", actUnignoreMultiple);

    QMap<QString, int> actionCounter;

    foreach (const QString & action, actions)
    {
        actionCounter.insert(action, 0);
    }

    foreach (const QModelIndex & index, indexList)
    {
        item = modelItem(sourceIndex(index));
        invitem = dynamic_cast<const InventoryItem *>(item);

        // skip actions on anything other than InventoryItems
        if (!invitem)
            continue;

        if (!invitem->isTracked())
        {
            actionCounter["add"]++;

            if (invitem->hasStatus(InventoryItem::Unknown))
            {
                actionCounter["ignore"]++;
            }
            else
            {
                actionCounter["unignore"]++;
            }
            continue;
        }

        if (invitem->hasChangedRecursive())
        {
            actionCounter["commit"]++;
            actionCounter["revert"]++;
        }

        actionCounter["drop"]++;
    }

    QMapIterator<QString, int> i(actionCounter);
    int totalItems = indexList.size();
    bool commonActions = false;

    QMenu menu(this);

    while (i.hasNext())
    {
        i.next();
        if (i.value() < totalItems) continue;
        commonActions = true;
        QAction * act = actionMapping[i.key()];
        act->setText(act->text().arg(totalItems));
        menu.addAction(act);
    }

    if (!commonActions)
    {
        QAction * act = new QAction(
            tr("No common action for selected items"), this
        );
        menu.addAction(act);
    }

    menu.exec(pos);
}

void InventoryView::slotChdir()
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());
    I(list.size() > 0);

    emit directoryChanged(list[0]);
    changeDirectory(list[0]);
}

void InventoryView::changeDirectory(const QModelIndex & proxy)
{
    // this seems to be a bit backward, but is here for a reason: the index
    // we're getting here might be created from another proxy model and all
    // we now is that this other proxy model and the proxy model we're using
    // ourselves have the same original model. So we need to translate the
    // proxy index at first _from_ the correct proxy model index to the
    // "unique" source index and later translate the source index back to
    // the proxy index which is used by our own model...
    QModelIndex source = sourceIndex(proxy, true);
    QModelIndex ourProxyIndex = proxyIndex(source);

    if (source.isValid())
        lastActiveIndex = source;

    if (type == FileList)
    {
        // if we don't do this, right-clicking in the empty space will
        // popup actions for a hidden item if the root index changed
        clearSelection();

        const ModelItem * item = modelItem(source);
        const PseudoItem * psitem = dynamic_cast<const PseudoItem *>(item);

        if (psitem && psitem->isCdUp())
        {
            lastActiveIndex = source.parent().parent();
            // list the contents of the parent of the parent directory
            setRootIndex(ourProxyIndex.parent().parent());
            return;
        }

        setRootIndex(ourProxyIndex);
        return;
    }

    // expand the selection if needed
    expand(ourProxyIndex);

    // select the item
    selectionModel()->setCurrentIndex(
        ourProxyIndex,
        QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows
    );
}

void InventoryView::slotOpen()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem =
        dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    emit openFile(invitem->getPath());
}

void InventoryView::slotAdd()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;
    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        if (!invitem || invitem->isTracked())
        {
            continue;
        }

        // rename the "root" directory to ".", since empty restrictions
        // a la "" will make mtn bail out
        QString path = invitem->getPath();
        if (path.isEmpty())
        {
            path = ".";
        }
        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be added"),
            tr("None of the selected items can be added."),
            QMessageBox::Ok
        );
        return;
    }

    emit addPaths(paths);
}

void InventoryView::slotRemove()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;
    int pathsWithChanges = 0;

    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        if (!invitem)
            continue;

        QString path = invitem->getPath();
        if (path.isEmpty())
        {
            QMessageBox::information(
                this,
                tr("Item can't be dropped"),
                tr("The workspace root cannot be dropped."),
                QMessageBox::Ok
            );
            return;
        }

        int status = invitem->getStatusRecursive();

        if ((status & InventoryItem::Unknown) == InventoryItem::Unknown ||
            (status & InventoryItem::Ignored) == InventoryItem::Ignored)
        {
            QMessageBox::information(
                this,
                tr("Item can't be dropped"),
                tr("The item '%1' contains unversioned files and "
                   "therefor cannot be dropped.").arg(path),
                QMessageBox::Ok
            );
            return;
        }
        if ((status & InventoryItem::ContentsChanged) == InventoryItem::ContentsChanged ||
            (status & InventoryItem::AttributesChanged) == InventoryItem::AttributesChanged)
        {
            pathsWithChanges++;
        }

        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be dropped"),
            tr("None of the selected items can be dropped."),
            QMessageBox::Ok
        );
        return;
    }

    if (pathsWithChanges > 0)
    {
        QMessageBox::StandardButton btn = QMessageBox::question(
            this,
            tr("Items have changes"),
            tr("%n out of %1 item(s) have changes which cannot "
               "be reverted after the removal. Are you sure you "
               "want to continue?", "", pathsWithChanges).arg(paths.size()),
            QMessageBox::Yes | QMessageBox::No
        );

        if (btn == QMessageBox::No) return;
    }

    emit dropPaths(paths);
}

void InventoryView::slotCommit()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;
    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        // not an inventory item
        if (!invitem) continue;

        // check if the node has committable changes
        if (!invitem->hasChangedRecursive()) continue;

        // rename the "root" directory to ".", since empty restrictions
        // a la "" will make mtn bail out
        QString path = invitem->getPath();
        if (path.isEmpty())
        {
            path = ".";
        }
        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be committed"),
            tr("None of the selected items have changes or can be committed."),
            QMessageBox::Ok
        );
        return;
    }

    emit commitRevision(paths);
}

void InventoryView::slotRefresh()
{
    QSet<const ModelItem *> items = getSelectedItems();

    // for now we can only handle one refresh at a time
    if (items.size() > 1) return;

    const ModelItem * item;

    // special hack: if there are no selected items, refresh the
    // current root item
    if (items.size() == 0)
    {
        QModelIndex sourceRoot = sourceIndex(rootIndex());

        // even more special case - select the first child of this (pseudo) root
        if (!sourceRoot.isValid())
        {
            QModelIndex firstItem = model()->index(0, 0, QModelIndex());
            sourceRoot = sourceIndex(firstItem);
        }

        // still not valid, i.e. filtered out?
        if (!sourceRoot.isValid())
        {
            return;
        }

        item = modelItem(sourceRoot);
    }
    else
    {
        item = *(items.begin());
    }

    if (!item) return;

    const InventoryItem * invItem = dynamic_cast<const InventoryItem *>(item);

    if (!invItem) return;

    emit refreshPath(invItem->getPath());
    return;
}

void InventoryView::slotRevert()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;

    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        if (!invitem)
            continue;

        if (!invitem->hasChangedRecursive())
            continue;

        QString path = invitem->getPath();
        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be reverted"),
            tr("None of the selected items can be reverted."),
            QMessageBox::Ok
        );
        return;
    }

    QMessageBox::StandardButton btn = QMessageBox::question(
        this,
        tr("Revert items"),
        tr("Are you sure you want to revert the selected items? "
           "This action cannot be undone."),
        QMessageBox::Yes | QMessageBox::No
    );

    if (btn == QMessageBox::No) return;

    emit revertPaths(paths);
}

void InventoryView::slotRename()
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list;
    foreach (QModelIndex index, selectionModel->selectedIndexes())
    {
        if (index.column() != 0) continue;
        list.append(index);
    }

    if (list.size() != 1) return;
    edit(list.at(0));
}

//
// TODO: there is no mtn API for the next two commands available
// right now so we need to manually edit the WD/.mtn-ignore file
// Q: If this file is under version control, should a change
// be committed in the background?
//
void InventoryView::slotIgnore()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;

    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        if (!invitem)
            continue;

        if (invitem->hasNotStatus(InventoryItem::Unknown))
            continue;

        QString path = invitem->getPath();
        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be ignored"),
            tr("None of the selected items can be ignored."),
            QMessageBox::Ok
        );
        return;
    }

    emit ignorePaths(paths);
}

void InventoryView::slotUnignore()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    QStringList paths;

    foreach (const ModelItem * item, items)
    {
        const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);

        if (!invitem)
            continue;

        if (invitem->hasNotStatus(InventoryItem::Ignored))
            continue;

        QString path = invitem->getPath();
        paths.push_back(path);
    }

    if (paths.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Items can't be unignored"),
            tr("None of the selected items can be unignored."),
            QMessageBox::Ok
        );
        return;
    }

    emit unignorePaths(paths);
}

void InventoryView::slotFileDiff()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    if (!invitem->hasStatus(InventoryItem::ContentsChanged))
    {
        L("File has no content changes");
        return;
    }

    emit diffFile(invitem->getPath());
}

void InventoryView::slotFileHistory()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    QString path = invitem->getPath();

    // determine the original path, if needed
    if (invitem->hasStatus(InventoryItem::RenameSource))
    {
        path = invitem->getRenameTarget();
    }
    if (invitem->hasStatus(InventoryItem::RenameTarget))
    {
        path = invitem->getRenameSource();
    }

    emit fileHistory(path);
}

void InventoryView::slotRevisionDiff()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    emit diffRevision(invitem->getPath(), QString(), QString());
}

void InventoryView::slotAnnotate()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    emit annotateFile(invitem->getPath());
}

QString InventoryView::getNewPath(const QModelIndex & parent, const QString & prefix)
{
    I(!prefix.isEmpty());
    I(parent.isValid());

    const ModelItem * modelParent = modelItem(sourceIndex(parent, true));
    const InventoryItem * invParent = qobject_cast<const InventoryItem *>(modelParent);
    if (!invParent) return QString();

    QString newFilename;
    int c = 0;
    while (true)
    {
        newFilename = ++c == 1 ? prefix : prefix + " " + QString::number(c);

        bool exists = false;
        foreach (ModelItem * child, invParent->getChildren())
        {
            InventoryItem * invChild = qobject_cast<InventoryItem *>(child);
            if (!invChild) continue;
            if (invChild->getFilename() == newFilename)
            {
                exists = true;
                break;
            }
        }

        if (!exists)
            break;
    }

    QString baseDirectory = invParent->getPath();
    if (baseDirectory.isEmpty())
        return newFilename;

    return baseDirectory + "/" + newFilename;
}

void InventoryView::slotNewFile()
{
    if (type != FileList) return;
    newPath = getNewPath(rootIndex(), tr("New file"));
    if (!newPath.isEmpty())
        emit newFile(newPath);
}

void InventoryView::slotNewDirectory()
{
    if (type != FileList) return;
    newPath = getNewPath(rootIndex(), tr("New directory"));
    if (!newPath.isEmpty())
        emit newDirectory(newPath);
}

void InventoryView::slotDelete()
{
    QSet<const ModelItem *> items = getSelectedItems();
    if (items.size() == 0) return;

    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(*items.begin());

    if (!invitem)
        return;

    emit deleteLocalPath(invitem->getPath());
}

QSet<const ModelItem *> InventoryView::getSelectedItems() const
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());

    QSet<const ModelItem *> items;

    if (list.size() == 0)
    {
        return items;
    }

    foreach (const QModelIndex & index, list)
    {
        items.insert(modelItem(sourceIndex(index)));
    }
    return items;
}

void InventoryView::itemClicked(const QModelIndex & index)
{
    DefaultAction act = getDefaultAction(index);
    if (act == Chdir)
    {
        emit directoryChanged(index);
        changeDirectory(index);
    }
}

void InventoryView::itemDoubleClicked(const QModelIndex & index)
{
    DefaultAction act = getDefaultAction(index);
    switch (act)
    {
        case Chdir:
            emit directoryChanged(index);
            changeDirectory(index);
            break;
        case FileDiff:
            slotFileDiff();
            break;
        case Commit:
            slotCommit();
            break;
        case Open:
            slotOpen();
            break;
        default:
            break;
    }
}

InventoryView::DefaultAction InventoryView::getDefaultAction(const QModelIndex & index) const
{
    if (!index.isValid())
        return None;

    QModelIndex source = sourceIndex(index);
    I(source.isValid());

    const ModelItem * item = modelItem(source);
    const InventoryItem * invitem = dynamic_cast<const InventoryItem *>(item);
    const PseudoItem * psitem = dynamic_cast<const PseudoItem *>(item);

    if ((invitem && invitem->isDirectory()) ||
        (psitem && psitem->isCdUp()))
    {
        return Chdir;
    }

    I(invitem);

    if (invitem->hasStatus(InventoryItem::ContentsChanged) &&
        invitem->hasNotStatus(InventoryItem::Added))
    {
        return FileDiff;
    }

    if (invitem->hasChangedRecursive())
    {
        return Commit;
    }

    if (invitem->hasStatus(InventoryItem::Unknown) ||
        invitem->hasStatus(InventoryItem::Ignored) || (
           invitem->isNewNode() &&
           invitem->hasNotStatus(InventoryItem::Missing) &&
           invitem->hasNotStatus(InventoryItem::Invalid)
       ))
    {
        return Open;
    }

    L(QString("no default action for item %1 applicable")
        .arg(invitem->getLabel()));
    return None;
}

void InventoryView::rowsInserted(const QModelIndex & parent, int start, int end)
{
    Q_UNUSED(start);
    Q_UNUSED(end);

    if (parent.isValid() && !newPath.isEmpty())
    {
        for (int i=start; i<=end; i++)
        {
            QModelIndex proxyIndex = parent.model()->index(i, 0, parent);
            const ModelItem * item = modelItem(sourceIndex(proxyIndex, true));
            const InventoryItem * invItem = qobject_cast<const InventoryItem *>(item);

            if (!invItem || invItem->getPath() != newPath)
                continue;

            selectionModel()->setCurrentIndex(
                proxyIndex,
                QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows
            );

            edit(proxyIndex);

            newPath = QString();
        }
    }
}

void InventoryView::delayedLayoutChanged()
{
    QModelIndex index = model()->index(0, 0, QModelIndex());
    // if we have a previous, still valid active source index, use
    // that to restore the state somewhat which existed before the layout
    // of the view changed completely
    if (lastActiveIndex.isValid())
    {
        index = proxyIndex(lastActiveIndex);
    }

    if (index.isValid())
    {
        setStyleSheet(QString());

        QVariant dat = index.data();

        if (type == FileList)
        {
            setRootIndex(index);
            return;
        }

        expand(index);
        return;
    }

    setStyleSheet("InventoryView { "
                  "background-color: white;"
                  "background-image: url(:/images/no_matches.png);"
                  "background-attachment: fixed;"
                  "background-position: center;"
                  "background-repeat: no-repeat;}"
    );
}

