/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "TreeView.h"
#include "Settings.h"
#include "vocab.h"

#include <QHeaderView>
#include <QItemSelectionModel>

TreeView::TreeView(QWidget * parent) : QTreeView(parent)
{
    init();
}

TreeView::TreeView(QWidget * parent, const QString & objName) : QTreeView(parent)
{
    setObjectName(objName);
    init();
}

TreeView::~TreeView()
{
    saveHeaderViewState();
}

void TreeView::init()
{
    header()->setResizeMode(QHeaderView::Interactive);
    stateLoaded = false;
}

void TreeView::setModel(QAbstractItemModel * model)
{
    QItemSelectionModel * oldSelectionModel = selectionModel();
    if (oldSelectionModel != 0)
    {
        disconnect(
           oldSelectionModel, SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
           this, SLOT(itemSelectionChanged(const QItemSelection &, const QItemSelection &))
        );
    }

    // each time a new model is set, a new selection model is internally
    // created by Qt, so we need to re-connect to this new model to get
    // notified about selection changes
    QTreeView::setModel(model);
    QItemSelectionModel * newSelectionModel = selectionModel();

    connect(
       newSelectionModel, SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
       this, SLOT(itemSelectionChanged(const QItemSelection &, const QItemSelection &))
    );

    Settings::restoreHeaderViewState(header(), QString(objectName()).append("_header"));
    stateLoaded = true;
}

void TreeView::saveHeaderViewState()
{
    if (!stateLoaded) return;
    Settings::saveHeaderViewState(header(), QString(objectName()).append("_header"));
}

void TreeView::expandAllNodes(bool expand)
{
    if (expand)
        expandAll();
    else
        collapseAll();
}

void TreeView::contextMenuEvent(QContextMenuEvent* ev)
{
    QItemSelectionModel * selection = selectionModel();

    // make sure the element under the cursor is selected as well
    QModelIndex currentItem = indexAt(ev->pos());
    selection->select(currentItem,
        QItemSelectionModel::Select |
        QItemSelectionModel::Rows
    );
    setSelectionModel(selection);

    // remove multiple cols of the same row, since we're selected
    // complete rows we're not interested in these
    QModelIndexList currentSelection = selection->selectedIndexes();
    for (int i=0; i < currentSelection.size(); i++)
    {
        if (currentSelection.at(i).column() > 0)
        {
            currentSelection.removeAt(i);
            i--;
        }
    }

    QPoint pos = ev->pos();

    QHeaderView * headerView = header();
    if (headerView)
    {
        if (headerView->orientation() == Qt::Horizontal)
        {
            pos += QPoint(0, headerView->height());
        }
        else
        {
            pos += QPoint(headerView->width(), 0);
        }
    }

    emit contextMenuRequested(currentSelection, mapToGlobal(pos));
}

void TreeView::itemSelectionChanged(const QItemSelection & selected, const QItemSelection & deselected)
{
    // we're not interested in these
    Q_UNUSED(deselected);

    QModelIndexList selectedIndexes = selected.indexes();

    for (int i=0; i<selectedIndexes.size(); )
    {
        // ...and neither are we interested in these in a row-based view
        if (selectedIndexes.at(i).column() > 0)
        {
            selectedIndexes.removeAt(i);
            continue;
        }
        i++;
    }

    emit selectedRows(selectedIndexes);
}

