#ifndef VOCAB_H
#define VOCAB_H

//
// global macros and defines
//

// if you use APP, you also have to include GuitoneCore.h
class GuitoneCore;
#define APP reinterpret_cast<GuitoneCore *>(qApp)

#include "DebugLog.h"
#ifdef QT_NO_DEBUG
#define D(msg) void(msg)
#else
#define D(msg) DebugLog::debug(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#endif

#define L(msg) DebugLog::info(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#define W(msg) DebugLog::warn(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#define C(msg) DebugLog::critical(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))

// for abort()
#include <stdlib.h>
#define F(msg) \
    do {  DebugLog::fatal(QString("%1:%2:%3: %4") \
          .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg))); \
          abort(); \
    } while (0);

#define I(expr) \
    do { if (!(expr)) { F(QString("invariant \"%1\" violated").arg(#expr)) } } while (0);

//
// type definitions
//

#include <QMap>
#include <QPair>
#include <QList>
#include <QStringList>

// used for BasicIOParser and BasicIOWriter
struct StanzaEntry
{
    QString sym;
    QStringList vals;

    StanzaEntry() {}
    StanzaEntry(QString s, QString v) : sym(s) { vals.push_back(v); }
    StanzaEntry(QString s, QStringList v) : sym(s), vals(v) {}
};

typedef QList <StanzaEntry> Stanza;
typedef QList <Stanza> StanzaList;

// used for manifest entries, if the bool var is true, the entry is a directory
struct FileEntry
{
    QString path;
    bool is_dir;
    QString fileid;
    QMap<QString, QString> attrs;

    FileEntry() {}
    FileEntry(QString p, bool d) : path(p), is_dir(d) {}
    FileEntry(QString p, bool d, QString f) : path(p), is_dir(d), fileid(f) {}

    inline bool operator<(const FileEntry & other) const
    {
        return path < other.path;
    }
};

typedef QList<FileEntry> FileEntryList;

typedef QList<QByteArray> ByteArrayList;

struct Ticker
{
    QByteArray name;
    int total;
    int progress;
    bool complete;

    Ticker(const QByteArray & n = QByteArray())
        : name(n), total(0), progress(0), complete(false) {}

    /**
     * Updates this ticker's values with the values from another one
     * (of the same type)
     */
    void update(const Ticker & other)
    {
        if (name.isEmpty()) name = other.name;
        if (total < other.total) total = other.total;
        if (progress < other.progress) progress = other.progress;
        if (!complete && other.complete) complete = true;
    }

    QString toString(bool shortened = true) const
    {
        if (complete)
            return QObject::tr("%1: completed")
                .arg(QString(name));

        QString prog;
        if (shortened && progress > (1024 * 1024))
            prog = QString("%1M").arg((double)progress / (1024 * 1024), 0, 'f', 1);
        else if (shortened && progress > 1024)
            prog = QString("%1K").arg((double)progress / 1024, 0, 'f', 1);
        else
            prog = QString::number(progress);

        if (total == 0)
            return QObject::tr("%1: %2")
                .arg(QString(name)).arg(prog);

        QString tot;
        if (shortened && total > (1024 * 1024))
            tot = QString("%1M").arg((double)total / (1024 * 1024), 0, 'f', 1);
        else if (shortened && total > 1024)
            tot = QString("%1K").arg((double)total / 1024, 0, 'f', 1);
        else
            tot = QString::number(total);

        return QObject::tr("%1: %2 / %3")
            .arg(QString(name)).arg(prog).arg(tot);
    }
};

typedef QMap<char, Ticker> TickerMap;

typedef QPair<char, QByteArray> MessagePair;

#include <QSharedPointer>

class MonotoneThread;
typedef QSharedPointer<MonotoneThread> MonotoneThreadPtr;

class MonotoneTask;
typedef QSharedPointer<MonotoneTask>   MonotoneTaskPtr;

class MonotoneHandle;
typedef QSharedPointer<MonotoneHandle> MonotoneHandlePtr;

// these are used almost everywhere, so include them here
#include "MonotoneHandle.h"
#include "MonotoneTask.h"

#include <exception>
class GuitoneException : public std::exception
{
   QString message;
public:
   GuitoneException(const QString & msg) : message(msg) {}
   virtual ~GuitoneException() throw() {}
   const char * what() const throw() { return message.toStdString().c_str(); }
};

#endif

